/*
 * domain_exec.c: helper for executing commands in guest
 *
 * Copyright (C) 2022 Virtuozzo, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include "domain_exec.h"

#include "virlog.h"
#include "virobject.h"
#include "viralloc.h"
#include "virtime.h"
#include "domain_conf.h"

#define VIR_FROM_THIS VIR_FROM_DOMAIN

VIR_LOG_INIT("util.exec");

typedef struct _virDomainExecuteStreamReader virDomainExecuteStreamReader;
struct _virDomainExecuteStreamReader {
    virObjectLockable parent;

    virCond *cond;
    virStreamPtr stream;
    char *buf;
    size_t length;
    size_t offset;
    bool err;
    bool eof;
};

static virClass *virDomainExecuteStreamReaderClass;

static void
virDomainExecuteStreamReaderDispose(void *obj)
{
    virDomainExecuteStreamReader *rd = obj;

    if (rd->stream) {
        virObjectUnref(rd->stream);
        rd->stream = NULL;
    }

    VIR_FREE(rd->buf);
}

static int
virDomainExecuteStreamReaderOnceInit(void)
{
    if (!VIR_CLASS_NEW(virDomainExecuteStreamReader, virClassForObjectLockable()))
        return -1;

    return 0;
}

VIR_ONCE_GLOBAL_INIT(virDomainExecuteStreamReader);

static virDomainExecuteStreamReader *
virDomainExecuteStreamReaderNew(virConnectPtr conn,
                                virCond *cond)
{
    virDomainExecuteStreamReader *ret;

    if (virDomainExecuteStreamReaderInitialize() < 0)
        return NULL;

    if (!(ret = virObjectNew(virDomainExecuteStreamReaderClass)))
        return NULL;

    ret->cond = cond;
    ret->length = 4096;
    ret->buf = g_new0(char, ret->length);

    if (!(ret->stream = virStreamNew(conn, VIR_STREAM_NONBLOCK))) {
        virObjectUnref(ret);
        return NULL;
    }

    return ret;
}


static void
virDomainExecuteEvent(virStreamPtr stream,
                      int events,
                      void *opaque)
{
    int ret;
    virDomainExecuteStreamReader *rd = opaque;

    virObjectLock(rd);
    if (!rd->cond) {
        virObjectUnlock(rd);
        return;
    }

    if (events & VIR_EVENT_HANDLE_ERROR) {
        VIR_ERROR(_("Error for reading stream"));
        goto error;
    }

    if (!(events & (VIR_EVENT_HANDLE_READABLE | VIR_EVENT_HANDLE_HANGUP))) {
        VIR_ERROR(_("Received unknown stream event %d"), events);
        goto error;
    }

    ret = virStreamRecv(stream, rd->buf + rd->offset, rd->length - rd->offset);
    if (ret == -1) {
        VIR_ERROR(_("Error for reading stream"));
        goto error;
    }

    if (ret == 0) {
        virStreamEventRemoveCallback(stream);
        virStreamFinish(stream);
        rd->eof = true;
        virCondSignal(rd->cond);
    } else if (ret > 0) {
        rd->offset += ret;
        if (rd->offset == rd->length) {
            rd->length += 1024;
            VIR_REALLOC_N(rd->buf, rd->length);
        }
    }

    virObjectUnlock(rd);
    return;

 error:
    virStreamEventRemoveCallback(stream);
    virStreamAbort(stream);
    rd->err = true;
    virCondSignal(rd->cond);

    virObjectUnlock(rd);
}

/*
 * Returns:
 *   -2 - command execution error
 *   -1 - command was terminated by signal
 *  >=0 - command exit code from guest
 */
int
virDomainExecuteCommand(virConnectPtr conn,
                        virDomainObj *obj,
                        const char *cmd,
                        const char **args,
                        unsigned int nargs,
                        char **output,
                        virDomainExecuteCommandCallbacks *callbacks,
                        void *opaque)
{
    int ret = -2;
    int pid = -1; /* exempt from syntax-check */
    virCond cond;
    virDomainExecuteStreamReader *errRd = NULL;
    virDomainExecuteStreamReader *outRd = NULL;
    virTimeBackOffVar timebackoff;
    virDomainCommandXStatus status;
    bool timeout = false;
    virError *err;

    if (virCondInit(&cond) < 0)
        goto cleanup;

    if (!(errRd = virDomainExecuteStreamReaderNew(conn, &cond)))
        goto cleanup;

    if (!(outRd = virDomainExecuteStreamReaderNew(conn, &cond)))
        goto cleanup;

    pid = callbacks->spawn(conn, obj, cmd,
                           (const char **)args, nargs,
                           NULL, 0,
                           NULL, outRd->stream, errRd->stream,
                           0, opaque);
    if (pid < 0)
        goto cleanup;

    virObjectRef(errRd);
    if (virStreamEventAddCallback(errRd->stream,
                                  VIR_STREAM_EVENT_HANGUP |
                                  VIR_STREAM_EVENT_ERROR |
                                  VIR_STREAM_EVENT_READABLE,
                                  virDomainExecuteEvent, errRd,
                                  virObjectFreeCallback) < 0) {
        virObjectUnref(errRd);
        goto cleanup;
    }

    virObjectRef(outRd);
    if (virStreamEventAddCallback(outRd->stream,
                                  VIR_STREAM_EVENT_HANGUP |
                                  VIR_STREAM_EVENT_ERROR |
                                  VIR_STREAM_EVENT_READABLE,
                                  virDomainExecuteEvent, outRd,
                                  virObjectFreeCallback) < 0) {
        virObjectUnref(outRd);
        goto cleanup;
    }

    if (virTimeBackOffStart(&timebackoff, 10, 30 * 1000) < 0)
        goto cleanup;

    while (!errRd->err && !outRd->err && !(errRd->eof && outRd->eof)) {
        if (virObjectWaitUntil(obj, &cond, timebackoff.limit_t) < 0) {
            if (errno == ETIMEDOUT)
                timeout = true;
            goto cleanup;
        }
    }

    if (!errRd->eof || !outRd->eof)
        goto cleanup;

    while (virTimeBackOffWaitUnlocked(&timebackoff, obj)) {
        if (callbacks->status(obj, pid, &status, opaque) < 0)
            goto cleanup;

        if (status.signal > 0) {
            ret = -1;
            goto cleanup;
        }

        if (status.exited) {
            ret = status.code;
            break;
        }
    }

    if (ret == -2)
        timeout = true;

 cleanup:
    if (ret == -1) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Command '%s' for domain '%s' was terminated by signal %d"),
                       cmd, obj->def->name, status.signal);
    } else if (ret == -2) {
        if (timeout)
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Command '%s' for domain '%s' failed to wait for completion"),
                           cmd, obj->def->name);
        else
            VIR_ERROR(_("Command '%s' for domain '%s' failed to execute"),
                      cmd, obj->def->name);
    } else if (ret) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Command '%s' for domain '%s' exited with code '%d':\nstdout:\n%s\nstderr:\n%s"),
                       cmd, obj->def->name, ret, outRd->buf, errRd->buf);
    }

    if (ret)
        virErrorPreserveLast(&err);

    if (pid > 0 && ret < 0)
        callbacks->terminate(obj, pid, opaque);

    if (output)
        *output = g_steal_pointer(&outRd->buf);

    if (errRd) {
        if (pid > 0 && !errRd->err && !errRd->eof) {
            virStreamEventRemoveCallback(errRd->stream);
            virStreamAbort(errRd->stream);
        }

        virObjectLock(errRd);
        errRd->cond = NULL;
        virObjectUnlock(errRd);

        virObjectUnref(errRd);
    }

    if (outRd) {
        if (pid > 0 && !outRd->err && !outRd->eof) {
            virStreamEventRemoveCallback(outRd->stream);
            virStreamAbort(outRd->stream);
        }

        virObjectLock(outRd);
        outRd->cond = NULL;
        virObjectUnlock(outRd);

        virObjectUnref(outRd);
    }

    virCondDestroy(&cond);

    if (ret)
        virErrorRestore(&err);

    return ret;
}
