/*
 * virthread.c: basic thread synchronization primitives
 *
 * Copyright (C) 2009-2010, 2014 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>

#include "virthread.h"

#ifdef __FreeBSD__
# include <pthread_np.h>
#endif

#include <unistd.h>
#include <inttypes.h>
#if WITH_SYS_SYSCALL_H
# include <sys/syscall.h>
#endif
#include <fcntl.h>

#include "viralloc.h"
#include "virthreadjob.h"
#include "virlog.h"
#include "virtime.h"
#include "vircommand.h"
#include "virfile.h"
#include "virstring.h"

#define VIR_FROM_THIS VIR_FROM_NONE

VIR_LOG_INIT("util.virthread");

int virOnce(virOnceControl *once, virOnceFunc init)
{
    int ret;

    ret = pthread_once(&once->once, init);
    if (ret != 0) {
        errno = ret;
        return -1;
    }

    return 0;
}


int virMutexInit(virMutex *m)
{
    int ret;
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_NORMAL);
    ret = pthread_mutex_init(&m->lock, &attr);
    pthread_mutexattr_destroy(&attr);
    if (ret != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

int virMutexInitRecursive(virMutex *m)
{
    int ret;
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
    ret = pthread_mutex_init(&m->lock, &attr);
    pthread_mutexattr_destroy(&attr);
    if (ret != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

void virMutexDestroy(virMutex *m)
{
    pthread_mutex_destroy(&m->lock);
}

char*
virBacktraceGather(void)
{
    virCommand *cmd = virCommandNew("/usr/bin/pstack");
    g_autofree char *tmpfile = NULL;
    g_autofree char *msg = NULL;
    int outfd = -1;
    char *ret = NULL;
    int status;
    int len;
    struct stat st;

    virCommandAddArgFormat(cmd, "%u", getpid());

    tmpfile = g_strdup("/tmp/libvirt.backtrace.XXXXXX");
    if ((outfd = g_mkstemp_full(tmpfile, O_WRONLY | O_CLOEXEC, S_IRUSR | S_IWUSR)) == -1) {
        virReportSystemError(errno, _("cannot create temporary file %s"),
                             tmpfile);
        goto cleanup;
    }

    virCommandSetOutputFD(cmd, &outfd);
    virCommandSetErrorFD(cmd, &outfd);

    if (virCommandRun(cmd, &status) < 0)
        goto cleanup;

    if (fstat(outfd, &st) < 0) {
        virReportSystemError(errno, _("stat(%s) failed"), tmpfile);
        goto cleanup;
    }

    /*
     * File description is shared so we need to rewind before
     * reading contents.
     */
    if (lseek(outfd, 0, SEEK_SET) < 0) {
        virReportSystemError(errno, _("lseek(%s) failed"), tmpfile);
        goto cleanup;
    }

    len = st.st_size;
    msg = g_new0(char, len + 1);
    if (saferead(outfd, msg, len) < 0) {
        virReportSystemError(errno, "%s",
                             _("failed to read output"));
        goto cleanup;
    }

    if (status) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("pstack failed, stream: %s"), msg);
        goto cleanup;
    }

    ret = g_steal_pointer(&msg);

 cleanup:
    VIR_FORCE_CLOSE(outfd);
    unlink(tmpfile);
    virCommandFree(cmd);
    return ret;
}

int virThreadDebug = 0;

void virMutexLock(virMutex *m)
{
    struct timespec until_ts;
    unsigned long long start;
    unsigned long long finish;
    unsigned long long until;
    g_autofree char *bt = NULL;

    if (!virThreadDebug) {
        pthread_mutex_lock(&m->lock);
        return;
    }

    ignore_value(virTimeMillisNow(&start));

    until = start + 5 * 1000ull;
    until_ts.tv_sec = until / 1000ull;
    until_ts.tv_nsec = (until % 1000ull) * 1000000ull;

    if (pthread_mutex_timedlock(&m->lock, &until_ts) == 0)
        return;

    bt = virBacktraceGather();
    if (!bt || bt[0] == '\0')
        VIR_WARN("Cannot take %p lock for 5000 ms", &m->lock);
    else
        VIR_WARN("Cannot take %p lock for 5000 ms\n%s", &m->lock, bt);
    VIR_FREE(bt);

    pthread_mutex_lock(&m->lock);

    ignore_value(virTimeMillisNow(&finish));

    VIR_WARN("Taking %p lock took %llu ms",
             &m->lock, finish > start ? finish - start : 0);
}

void virMutexUnlock(virMutex *m)
{
    pthread_mutex_unlock(&m->lock);
}


int virRWLockInit(virRWLock *m)
{
    int ret;
    ret = pthread_rwlock_init(&m->lock, NULL);
    if (ret != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

void virRWLockDestroy(virRWLock *m)
{
    pthread_rwlock_destroy(&m->lock);
}


void virRWLockRead(virRWLock *m)
{
    pthread_rwlock_rdlock(&m->lock);
}

void virRWLockWrite(virRWLock *m)
{
    pthread_rwlock_wrlock(&m->lock);
}


void virRWLockUnlock(virRWLock *m)
{
    pthread_rwlock_unlock(&m->lock);
}

int virCondInit(virCond *c)
{
    int ret;
    if ((ret = pthread_cond_init(&c->cond, NULL)) != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

int virCondDestroy(virCond *c)
{
    int ret;
    if ((ret = pthread_cond_destroy(&c->cond)) != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

int virCondWait(virCond *c, virMutex *m)
{
    int ret;
    if ((ret = pthread_cond_wait(&c->cond, &m->lock)) != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

int virCondWaitUntil(virCond *c, virMutex *m, unsigned long long whenms)
{
    int ret;
    struct timespec ts;

    ts.tv_sec = whenms / 1000;
    ts.tv_nsec = (whenms % 1000) * 1000000;

    if ((ret = pthread_cond_timedwait(&c->cond, &m->lock, &ts)) != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

void virCondSignal(virCond *c)
{
    pthread_cond_signal(&c->cond);
}

void virCondBroadcast(virCond *c)
{
    pthread_cond_broadcast(&c->cond);
}

struct virThreadArgs {
    virThreadFunc func;
    char *name;
    bool worker;
    void *opaque;
};

size_t virThreadMaxName(void)
{
#if defined(__FreeBSD__) || defined(__APPLE__)
    return 63;
#else
# ifdef __linux__
    return 15;
# else
    return 0; /* unlimited */
# endif
#endif
}

static void *virThreadHelper(void *data)
{
    struct virThreadArgs *args = data;
    struct virThreadArgs local = *args;
    g_autofree char *thname = NULL;
    size_t maxname = virThreadMaxName();

    /* Free args early, rather than tying it up during the entire thread.  */
    g_free(args);

    if (local.worker)
        virThreadJobSetWorker(local.name);
    else
        virThreadJobSet(local.name);

    if (maxname) {
        thname = g_strndup(local.name, maxname);
    } else {
        thname = g_strdup(local.name);
    }

#if defined(__linux__) || defined(WIN32)
    pthread_setname_np(pthread_self(), thname);
#else
# ifdef __FreeBSD__
    pthread_set_name_np(pthread_self(), thname);
# else
#  ifdef __APPLE__
    pthread_setname_np(thname);
#  endif
# endif
#endif

    local.func(local.opaque);

    if (!local.worker)
        virThreadJobClear(0);

    g_free(local.name);
    return NULL;
}

int virThreadCreateFull(virThread *thread,
                        bool joinable,
                        virThreadFunc func,
                        const char *name,
                        bool worker,
                        void *opaque)
{
    struct virThreadArgs *args;
    pthread_attr_t attr;
    int ret = -1;
    int err;

    if ((err = pthread_attr_init(&attr)) != 0)
        goto cleanup;

    args = g_new0(struct virThreadArgs, 1);
    args->func = func;
    args->name = g_strdup(name);
    args->worker = worker;
    args->opaque = opaque;

    if (!joinable)
        pthread_attr_setdetachstate(&attr, 1);

    err = pthread_create(&thread->thread, &attr, virThreadHelper, args);
    if (err != 0) {
        g_free(args->name);
        g_free(args);
        goto cleanup;
    }
    /* New thread owns 'args' in success case, so don't free */

    ret = 0;
 cleanup:
    pthread_attr_destroy(&attr);
    if (ret < 0)
        errno = err;
    return ret;
}

void virThreadSelf(virThread *thread)
{
    thread->thread = pthread_self();
}

bool virThreadIsSelf(virThread *thread)
{
    return pthread_equal(pthread_self(), thread->thread) ? true : false;
}

/* For debugging use only; this result is not guaranteed unique if
 * pthread_t is larger than a 64-bit pointer, nor does it always match
 * the pthread_self() id on Linux.  */
unsigned long long virThreadSelfID(void)
{
#if defined(WITH_SYS_SYSCALL_H) && defined(SYS_gettid) && defined(__linux__)
    pid_t tid = syscall(SYS_gettid);
    return tid;
#else
    union {
        unsigned long long l;
        pthread_t t;
    } u;
    u.t = pthread_self();
    return u.l;
#endif
}

/* For debugging use only; this result is not guaranteed unique if
 * pthread_t is larger than a 64-bit pointer, nor does it always match
 * the thread id of virThreadSelfID on Linux.  */
unsigned long long virThreadID(virThread *thread)
{
    union {
        unsigned long long l;
        pthread_t t;
    } u;
    u.t = thread->thread;
    return u.l;
}

void virThreadJoin(virThread *thread)
{
    pthread_join(thread->thread, NULL);
}

void virThreadCancel(virThread *thread)
{
    pthread_cancel(thread->thread);
}

int virThreadLocalInit(virThreadLocal *l,
                       virThreadLocalCleanup c)
{
    int ret;
    if ((ret = pthread_key_create(&l->key, c)) != 0) {
        errno = ret;
        return -1;
    }
    return 0;
}

void *virThreadLocalGet(virThreadLocal *l)
{
    return pthread_getspecific(l->key);
}

int virThreadLocalSet(virThreadLocal *l, void *val)
{
    int err = pthread_setspecific(l->key, val);
    if (err) {
        errno = err;
        return -1;
    }
    return 0;
}
