/*
 * vzct_driver.c: core driver functions for managing
 * Virtuozzo Containers
 *
 * Copyright (C) 2017 Virtuozzo, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>
#include <vzctl/libvzctl.h>
#include <vzctl/vzerror.h>
#include <linux/netlink.h>
#include <gio/gio.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <osinfo/osinfo.h>

#include "virlog.h"
#include "viraccessapicheck.h"
#include "viraccessapicheckext.h"
#include "viralloc.h"
#include "virdomainobjlist.h"
#include "vireventthread.h"
#include "virutil.h"
#include "datatypes.h"
#include "virfile.h"
#include "virstring.h"
#include "virthreadpool.h"
#include "configmake.h"
#include "vircommand.h"
#include "virportallocator.h"
#include "virpidfile.h"
#include "virdomainsnapshotobjlist.h"
#include "snapshot_conf.h"
#include "virtime.h"
#include "virthreadjob.h"
#include "virprocess.h"
#include "domain_event.h"
#include "virfdstream.h"
#include "virhook.h"
#include "domain_nwfilter.h"
#include "domain_validate.h"
#include "domain_exec.h"

#include "vzct_driver.h"

#define VIR_FROM_THIS VIR_FROM_VZCT

VIR_LOG_INIT("vzct.vzct_driver");

#define VZCT_ERROR_MESSAGE_CODE(func, code) \
do { \
    const char *check_err = vzctl2_get_last_error(); \
    if (check_err && *check_err) { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed with code %d: %s"), \
                       #func,  code, check_err); \
    } else { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed with code %d"), \
                       #func,  code); \
    } \
} while (0)

#define VZCT_CHECK(label, func) do { \
    int check_rc; \
    if ((check_rc = func)) { \
        VZCT_ERROR_MESSAGE_CODE(func, check_rc); \
        goto label; \
    } \
} while (0)

#define VZCT_CHECK_RET(ret, func) do { \
    int check_rc; \
    if ((check_rc = func)) { \
        VZCT_ERROR_MESSAGE_CODE(func, check_rc); \
        return ret; \
    } \
} while (0)

#define VZCT_ERROR_MESSAGE(func) \
do { \
    const char *check_err = vzctl2_get_last_error(); \
    if (check_err && *check_err) { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed: %s"), \
                       #func,  check_err); \
    } else { \
        virReportError(VIR_ERR_INTERNAL_ERROR, \
                       _("%s failed"), \
                       #func); \
    } \
} while (0)

#define VZCT_CHECK_PTR(label, ptr, func) do { \
    if (!(ptr = func)) { \
        VZCT_ERROR_MESSAGE(func); \
        goto label; \
    } \
} while (0)

#define VZCT_CHECK_PTR_RET(ret, ptr, func) do { \
    if (!(ptr = func)) { \
        VZCT_ERROR_MESSAGE(func); \
        return ret; \
    } \
} while (0)

#define VZCT_CHECK_LOCKED(label, env, status, func) do { \
    int check_rc; \
    int lockfd; \
    if ((lockfd = vzctl2_env_lock(env, status)) < 0) { \
        VZCT_ERROR_MESSAGE(vzctl2_env_lock); \
        goto label; \
    } \
    if ((check_rc = func)) { \
        VZCT_ERROR_MESSAGE_CODE(func, check_rc); \
    } \
    vzctl2_env_unlock(env, lockfd); \
    if (check_rc) \
        goto label; \
} while (0)

#define VZCT_PRIVATE_DIR ".libvirt"
#define VZCT_SNAPHOTS_DIR VZCT_PRIVATE_DIR"/snapshots"

#define VZCT_NET_VENET0 "venet0"

#define VZCT_DISK_ROOT_UUID "{00000000-0000-0000-0000-000000000000}"

#define VZCT_LIBOSINFO_URI "http://libosinfo.org/xmlns/libvirt/domain/1.0"
#define VZCT_VHS_URI "http://www.virtuozzo.com/vhs"

#define VZCT_VHS_XPATH_OSTEMPLATE "string(/vz/ostemplate/@name)"
#define VZCT_LIBOSINFO_XPATH_OSID "string(/libosinfo/os/@id)"

#define VZCT_LIBOSINFO_VERSION_PATH "/usr/share/osinfo/VERSION"

VIR_ENUM_DECL(vzctlEvent);
VIR_ENUM_IMPL(vzctlEvent,
              VZCTL_ENV_CONFIGURED + 1,
              "unknown",
              "created",
              "registered",
              "unregistered",
              "deleted",
              "config_changed",
              "started",
              "stopped",
              "suspended",
              "classes_config_changed",
              "shaping_config_changed",
              "net_config_changed",
              "umount",
              "configured"
);


typedef struct _vzctDriverConfig vzctDriverConfig;
struct _vzctDriverConfig {
    char *stateDir;
};

typedef struct _vzctDriver vzctDriver;
struct _vzctDriver {
    virMutex lock;
    virDomainXMLOption *xmlopt;
    vzctDriverConfig *config;
    virCaps *caps;
    virEventThread *eventThread;

    virDomainObjList *domains;

    vzevt_handle_t *eventdHandle;
    GSocket *eventdSocket;
    GSource *eventdSource;

    int netlinkFd;
    GSocket *netlinkSocket;
    GSource *netlinkSource;

    virPortAllocatorRange *vncPorts;
    virPortAllocatorRange *migrationPorts;

    virThreadPool *workers;
    virObjectEventState *eventsManager;

    OsinfoLoader *osinfoLoader;
    time_t osinfoCtime;

    virHashAtomic *migrating;

    int lastconnid;
};

typedef enum {
    VZCT_DOMAIN_JOB_NONE = 0,
    VZCT_DOMAIN_JOB_QUERY = 1,
    VZCT_DOMAIN_JOB_MODIFY = 2,
    VZCT_DOMAIN_JOB_MIGRATION = 3,
} vzctDomainJobType;

typedef struct _vzctDomainJobStats vzctDomainJobStats;
typedef vzctDomainJobStats *vzctDomainJobStatsPtr;
struct _vzctDomainJobStats {
    unsigned long long timeElapsed;
};

typedef struct _vzctDomainExecProcess vzctDomainExecProcess;
typedef vzctDomainExecProcess *vzctDomainExecProcessPtr;
struct _vzctDomainExecProcess {
    struct vzctl_exec_handle *handle;
    int conn_id;
    bool orphan;
    bool terminated;
};


#define VZCT_MIGRATION_CHANNELS_NUM 4

typedef struct _vzctDomainMigrationData vzctDomainMigrationData;
struct _vzctDomainMigrationData {
    int vncPort;

    pid_t socatPids[VZCT_MIGRATION_CHANNELS_NUM];
    int ports[VZCT_MIGRATION_CHANNELS_NUM];

    pid_t pid; /* pid of vzmdest server */
};

typedef struct _vzctDomainObjPrivate vzctDomainObjPrivate;
struct _vzctDomainObjPrivate {
    virDomainSnapshotObjList *snapshots;
    char *bundlePath;
    pid_t vncPid;

    virCond jobCond;
    vzctDomainJobType job;
    const char *jobAPI;
    vzctDomainJobType sideJob;
    const char *sideJobAPI;
    bool allowSideJob;
    pid_t migrationPid;
    pid_t socatPids[VZCT_MIGRATION_CHANNELS_NUM];
    int mig_conn_id;
    unsigned long long jobStarted;
    bool abortJob;

    vzctDomainExecProcess **exec_procs;
    size_t exec_nprocs;
    size_t exec_nprocs_max;
    int cleanupTimer;

    vzctDomainMigrationData *migdata;
};

#define VZCT_DOMAIN_DISK_UUID_SIZE 39

typedef struct _vzctDomainDiskPrivate vzctDomainDiskPrivate;
typedef vzctDomainDiskPrivate *vzctDomainDiskPrivatePtr;
struct _vzctDomainDiskPrivate {
    virObject parent;

    /* Size is copied from libvzctl header. Looks like common uuid format but
     * with braces. */
    char uuid[VZCT_DOMAIN_DISK_UUID_SIZE];
};

typedef enum {
    VZCT_DRIVER_JOB_EVENT = 0,
    VZCT_DRIVER_JOB_CONNECTION_CLOSED = 1,
    VZCT_DRIVER_JOB_CLEANUP = 2,
} vzctDriverJobType;

typedef enum {
    VZCT_DOMAIN_EVENT_NONE = 0,
    VZCT_DOMAIN_EVENT_STARTED,
    VZCT_DOMAIN_EVENT_STOPPED,
    VZCT_DOMAIN_EVENT_DEFINED,
    VZCT_DOMAIN_EVENT_UNDEFINED,
    VZCT_DOMAIN_EVENT_UPDATED,
    VZCT_DOMAIN_EVENT_CONFIGURED,
} vzctDomainEventType;

typedef struct _vzctDriverJob vzctDriverJob;
struct _vzctDriverJob {
    vzctDriverJobType type;
    vzctDomainEventType eventType;
    char *ctid;
    int conn_id;
    virDomainObj *dom;
};


static virClass *vzctDomainDiskPrivateClass;
static void vzctDomainDiskPrivateDispose(void *obj);

static int
vzctDomainDiskPrivateOnceInit(void)
{
    if (!VIR_CLASS_NEW(vzctDomainDiskPrivate, virClassForObject()))
        return -1;

    return 0;
}

VIR_ONCE_GLOBAL_INIT(vzctDomainDiskPrivate);

static vzctDriver *vzct_driver;

static void
vzctDriverConnectClose(vzctDriver *driver);

static virDomainObj *
vzctAddDomain(vzctDriver *driver,
              const char *id,
              virDomainObj *dom);

static int
vzctUpdateDomain(vzctDriver *driver,
                 virDomainObj *dom);

static int
vzctDomainStartFull(vzctDriver *driver,
                    virDomainObj *dom);

static void
vzctDomainCleanupTimeout(int timer, void *opaque);

G_DEFINE_AUTOPTR_CLEANUP_FUNC(virDomain, virObjectUnref);

static void
vzctDomainClearVolumes(virDomainDef *def,
                       vzctl_env_handle *env);

static int
vzctDomainCommandExecInternal(virConnectPtr conn,
                              virDomainObj *dom,
                              const char *cmd,
                              const char **args,
                              unsigned int nargs,
                              const char **envs,
                              unsigned int nenvs,
                              int rows,
                              int columns,
                              virStreamPtr in,
                              virStreamPtr out,
                              virStreamPtr err,
                              unsigned int flags);

static int
vzctDomainCommandSpawn(virConnectPtr conn,
                       virDomainObj *dom,
                       const char *cmd,
                       const char **args,
                       unsigned int nargs,
                       const char **envs,
                       unsigned int nenvs,
                       virStreamPtr in,
                       virStreamPtr out,
                       virStreamPtr err,
                       unsigned int flags,
                       void *opaque G_GNUC_UNUSED)
{
    return vzctDomainCommandExecInternal(conn, dom,
                                         cmd, args, nargs, envs, nenvs,
                                         0, 0, in, out, err, flags);
}

static int
vzctDomainCommandGetStatusInternal(virDomainObj *dom,
                                   int pid_value,
                                   virDomainCommandXStatusPtr status,
                                   void *opaque);

static int
vzctDomainCommandTerminateInternal(virDomainObj *dom,
                                   int pid_value,
                                   void *opaque);

static virDomainExecuteCommandCallbacks vzctExecuteCallbacks = {
    .spawn = vzctDomainCommandSpawn,
    .status = vzctDomainCommandGetStatusInternal,
    .terminate = vzctDomainCommandTerminateInternal,
};

static virObject *
vzctDomainDiskPrivateNew(void)
{
    vzctDomainDiskPrivate *priv;

    if (vzctDomainDiskPrivateInitialize() < 0)
        return NULL;

    if (!(priv = virObjectNew(vzctDomainDiskPrivateClass)))
        return NULL;

    return (virObject *) priv;
}


static void
vzctDomainDiskPrivateDispose(void *obj G_GNUC_UNUSED)
{
}

static void
vzctDomainFinishSnapshots(virDomainObj *dom);

static int
vzctDomainReadSnapshots(virDomainObj *dom);

static int
vzctDomainLoadSnapshotMetadata(vzctDriver *driver,
                               virDomainObj *dom);


static void
vzctDriverJobFree(vzctDriverJob *job)
{
    if (!job)
        return;

    VIR_FREE(job->ctid);
    VIR_FREE(job);
}


static void
vzctDomainStop(vzctDriver *driver,
               virDomainObj *dom);

static int
vzctMergeConfigs(virDomainDef *def,
                 virDomainDef *drvDef);

static virDomainDef *
vzctLoadDomainConfig(vzctDriver *driver,
                     const char *ctid,
                     char **bundlePath);


/*
 * We need this function to copy disk private.
 */
static virDomainDef *
vzctDomainDefCopy(virDomainDef *def,
                  virDomainXMLOption *xmlopt,
                  bool migratable)
{
    unsigned int format_flags = VIR_DOMAIN_DEF_FORMAT_SECURE |
                                VIR_DOMAIN_DEF_FORMAT_DISK_PRIVATE;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                               VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                                VIR_DOMAIN_DEF_PARSE_DISK_PRIVATE;
    g_autofree char *xml = NULL;
    virBuffer buf = VIR_BUFFER_INITIALIZER;

    if (migratable)
        format_flags |= VIR_DOMAIN_DEF_FORMAT_INACTIVE |
                        VIR_DOMAIN_DEF_FORMAT_MIGRATABLE;

    if (virDomainDefFormatInternal(def, xmlopt, &buf, format_flags) < 0 ||
        !(xml = virBufferContentAndReset(&buf)))
        return NULL;

    return virDomainDefParseString(xml, xmlopt, NULL, parse_flags);
}


static bool
vzctDomainJobAllowed(vzctDomainObjPrivate *priv,
                     vzctDomainJobType job)
{
    if (!priv->job)
        return true;

    return priv->allowSideJob &&
           !priv->sideJob &&
           job == VZCT_DOMAIN_JOB_QUERY;
}


/*
 * Acquire job.
 * @timeout: -1 - wait indefinetly
 *            0 - don't wait
 *           >0 - wait for specified timeout in milliseconds
 */
static int
vzctDomainBeginJobInternal(virDomainObj *dom,
                           vzctDomainJobType job,
                           long long timeout)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    unsigned long long then = 0;

    if (timeout > 0) {
        unsigned long long now;

        if (virTimeMillisNow(&now) < 0)
            return -1;

        then = now + timeout;
    }

    while (!vzctDomainJobAllowed(priv, job)) {
        int rc;

        if (timeout == 0)
            return -1;

        if (timeout > 0)
            rc = virObjectWaitUntil(dom, &priv->jobCond, then);
        else
            rc = virObjectWait(dom, &priv->jobCond);

        if (rc == 0)
            continue;

        if (errno == ETIMEDOUT) {
            if (priv->job && priv->sideJob) {
                virReportError(VIR_ERR_OPERATION_TIMEOUT,
                               _("cannot acquire lock for %s"
                                 " (held by job=%s sideJob=%s)"),
                               virThreadJobGet(),
                               priv->jobAPI, priv->sideJobAPI);
            } else {
                virReportError(VIR_ERR_OPERATION_TIMEOUT,
                               _("cannot acquire lock for %s"
                                 " (held by job=%s)"),
                               virThreadJobGet(), priv->jobAPI);
            }
        } else {
            virReportSystemError(errno, _("cannot acquire lock for %s"),
                                 virThreadJobGet());
        }

        return -1;
    }

    if (dom->removing) {
        char uuidstr[VIR_UUID_STRING_BUFLEN];

        virUUIDFormat(dom->def->uuid, uuidstr);
        if (timeout > 0)
            virReportError(VIR_ERR_NO_DOMAIN,
                           _("no domain name=%s uuid=%s ctid=%s"),
                           dom->def->name, uuidstr, dom->def->extraId);
        else
            VIR_DEBUG("no domain name=%s uuid=%s ctid=%s",
                      dom->def->name, uuidstr, dom->def->extraId);
        return -1;
    }

    if (job == VZCT_DOMAIN_JOB_MIGRATION &&
        virTimeMillisNow(&priv->jobStarted) < 0)
        return -1;

    if (priv->job) {
        VIR_DEBUG("begin side job %s", virThreadJobGet());
        priv->sideJob = job;
        priv->sideJobAPI = virThreadJobGet();
    } else {
        VIR_DEBUG("begin main job %s", virThreadJobGet());
        priv->job = job;
        priv->jobAPI = virThreadJobGet();
    }

    return 0;
}

#define VZCT_JOB_WAIT_TIME (30 * 1000)

static int
vzctDomainBeginJob(virDomainObj *dom,
                   vzctDomainJobType job)
{
    return vzctDomainBeginJobInternal(dom, job, VZCT_JOB_WAIT_TIME);
}


static void
vzctDomainAllowSideJob(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    priv->allowSideJob = true;
}


static void
vzctDomainDisableSideJob(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;

    /* Disable here so no new side job can start */
    priv->allowSideJob = false;
    while (priv->sideJob) {
        if (virObjectWait(dom, &priv->jobCond) < 0) {
            VIR_WARN("cannot wait side job finishing, errno=%d", errno);
            return;
        }
    }
}


static void
vzctDomainEndJob(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;

    if (priv->sideJob) {
        VIR_DEBUG("end side job %s", priv->sideJobAPI);
        priv->sideJob = VZCT_DOMAIN_JOB_NONE;
        priv->sideJobAPI = NULL;
    } else {
        VIR_DEBUG("end main job %s", priv->jobAPI);
        priv->job = VZCT_DOMAIN_JOB_NONE;
        priv->jobAPI = NULL;
        priv->mig_conn_id = 0;
        priv->jobStarted = 0;
        priv->abortJob = false;
        if (priv->allowSideJob)
            VIR_WARN("unexpected job end");
        priv->allowSideJob = false;
    }

    virCondBroadcast(&priv->jobCond);
}


static void
vzctDomainDropJob(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;

    VIR_DEBUG("drop job %s", priv->jobAPI);
    if (priv->sideJob) {
        VIR_DEBUG("convert side job to main job for %s", priv->sideJobAPI);
        priv->job = priv->sideJob;
        priv->jobAPI = priv->sideJobAPI;
        priv->sideJob = VZCT_DOMAIN_JOB_NONE;
        priv->sideJobAPI = NULL;
    } else {
        priv->job = VZCT_DOMAIN_JOB_NONE;
        priv->jobAPI = NULL;
        virCondBroadcast(&priv->jobCond);
    }

    priv->allowSideJob = false;
    priv->mig_conn_id = 0;
    priv->jobStarted = 0;
    priv->abortJob = false;
}


static gboolean
vzctNetlinkEvent(GSocket *socket G_GNUC_UNUSED,
                 GIOCondition cond,
                 gpointer opaque)
{
    vzctDriver *driver = opaque;
    struct sockaddr_nl addr;
    struct msghdr msg = { .msg_control = 0 };
    struct iovec iov;
    char buf[512];
    char *event, *ctid, *saveptr;
    vzctDriverJob *job = NULL;
    int len;

    virMutexLock(&driver->lock);

    if (!driver->netlinkSource)
        goto cleanup;

    if (cond & ~G_IO_IN) {
        if (cond & G_IO_OUT)
            VIR_WARN("Got writable event on vz event descriptor.");
        else if (cond & G_IO_HUP)
            VIR_DEBUG("Got HANGUP on vz event descriptor.");
        else if (cond & G_IO_ERR)
            VIR_WARN("Got ERROR on vz event descriptor.");

        /* Generally speaking we should not close connection as we go
         * out of sync. But these events should never happend on netlink
         * so this is just defencing from event loop eating 100% CPU. */
        vzctDriverConnectClose(driver);
        goto cleanup;
    }

    /* check for spurious */
    if (!(cond & G_IO_IN))
        goto cleanup;

    msg.msg_name = (void*)&addr;
    msg.msg_namelen = sizeof(addr);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    iov.iov_base = buf;
    iov.iov_len = sizeof(buf) - 1;

    while ((len = recvmsg(driver->netlinkFd, &msg, MSG_DONTWAIT)) < 0 &&
           errno == EINTR)
        ;

    if (len < 0) {
        if (errno != EAGAIN)
            VIR_ERROR(_("Can not read event message, error: %d"), errno);
        goto cleanup;
    }

    if (!len)
        goto cleanup;

    buf[len] = '\0';

    if (!(event = strtok_r(buf, "@", &saveptr))) {
        VIR_ERROR(_("Can not parse event type"));
        goto cleanup;
    }

    if (!(ctid = strtok_r(NULL, "@", &saveptr))) {
        VIR_ERROR(_("Can not parse container id"));
        goto cleanup;
    }

    VIR_LOG_VZ("received netlink event ctid=%s event=%s", ctid, event);

    /* Case for changing state in worker thread. During CT start
     * thru vzct driver first netlink event is arrived and blocked
     * on acquiring domain lock so the entire event loop is blocked.
     * Next in the process of starting domain prlsrvctl is called
     * to resolve network to bridge I guess. prlsrvctl in turn query
     * libvirtd and the latter is unresponsive as event loop is blocked */
    job = g_new0(vzctDriverJob, 1);
    job->type = VZCT_DRIVER_JOB_EVENT;
    job->ctid = g_strdup(ctid);
    if (STREQ(event, "ve-start")) {
        job->eventType = VZCT_DOMAIN_EVENT_STARTED;
    } else if (STREQ(event, "ve-stop")) {
        job->eventType = VZCT_DOMAIN_EVENT_STOPPED;
    } else {
        VIR_DEBUG("Unexpected event type %s", event);
        goto cleanup;
    }

    if (virThreadPoolSendJob(driver->workers, 0, job) < 0)
        goto cleanup;
    job = NULL;

 cleanup:
    vzctDriverJobFree(job);
    virMutexUnlock(&driver->lock);
    return G_SOURCE_CONTINUE;
}


static gboolean
vzctEventdEvent(GSocket *socket G_GNUC_UNUSED,
                GIOCondition cond,
                gpointer opaque)
{
    vzctDriver *driver = opaque;
    struct vzctl_state_evt vzevent;
    vzctDriverJob *job = NULL;

    virMutexLock(&driver->lock);

    if (!driver->eventdSource)
        goto cleanup;

    if (cond & ~G_IO_IN) {
        if (cond & G_IO_OUT)
            VIR_WARN("Got writable event on vz event descriptor.");
        else if (cond & G_IO_HUP)
            VIR_DEBUG("Got HANGUP on vz event descriptor.");
        else if (cond & G_IO_ERR)
            VIR_WARN("Got ERROR on vz event descriptor.");

        /* Generally speaking we should not close connection as we go
         * out of sync. But these events should never happend on netlink
         * so this is just defencing from event loop eating 100% CPU. */
        vzctDriverConnectClose(driver);
        goto cleanup;
    }

    /* check for spurious */
    if (!(cond & G_IO_IN))
        goto cleanup;

    /* FIXME Unfortunately this API returns -1 in case of no event and in
     * case of error thus we can miss some errors now. But errors are
     * OOM only so they are unlikely. Treat as no event. */
    if (vzctl2_get_state_evt(driver->eventdHandle,
                             &vzevent, sizeof(vzevent)) < 0)
        goto cleanup;

    VIR_LOG_VZ("received eventd event ctid=%s event=%s", vzevent.ctid,
               NULLSTR(vzctlEventTypeToString(vzevent.state)));

    job = g_new0(vzctDriverJob, 1);
    job->type = VZCT_DRIVER_JOB_EVENT;
    job->ctid = g_strdup(vzevent.ctid);
    switch (vzevent.state) {
    case VZCTL_ENV_CREATED:
    case VZCTL_ENV_REGISTERED:
        job->eventType = VZCT_DOMAIN_EVENT_DEFINED;
        break;
    case VZCTL_ENV_UNREGISTERED:
    case VZCTL_ENV_DELETED:
        job->eventType = VZCT_DOMAIN_EVENT_UNDEFINED;
        break;
    case VZCTL_ENV_CONFIG_CHANGED:
    case VZCTL_ENV_NET_CONFIG_CHANGED:
        job->eventType = VZCT_DOMAIN_EVENT_UPDATED;
        break;
    case VZCTL_ENV_CONFIGURED:
        job->eventType = VZCT_DOMAIN_EVENT_CONFIGURED;
        break;
    default:
        goto cleanup;
    }

    if (virThreadPoolSendJob(driver->workers, 0, job) < 0)
        goto cleanup;
    job = NULL;

 cleanup:
    vzctDriverJobFree(job);
    virMutexUnlock(&driver->lock);
    return G_SOURCE_CONTINUE;
}


static void
vzctDriverConnectClose(vzctDriver *driver)
{
    VIR_DEBUG("Closing connection to vzeventd.");

    if (driver->eventdSource) {
        g_source_destroy(driver->eventdSource);
        vir_g_source_unref(driver->eventdSource,
                           virEventThreadGetContext(driver->eventThread));
        driver->eventdSource = NULL;
        g_object_unref(driver->eventdSocket);
        driver->eventdSocket = NULL;
    }
    vzctl2_unregister_evt(driver->eventdHandle);
    driver->eventdHandle = NULL;

    if (driver->netlinkSource) {
        g_source_destroy(driver->netlinkSource);
        vir_g_source_unref(driver->netlinkSource,
                           virEventThreadGetContext(driver->eventThread));
        driver->netlinkSource = NULL;
        g_object_unref(driver->netlinkSocket);
        driver->netlinkSocket = NULL;
        driver->netlinkFd = -1;
    }
}


static char *
vzctStateFile(vzctDriverConfig *config, const unsigned char *uuid)
{
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    char *path;

    virUUIDFormat(uuid, uuidstr);
    path = g_strdup_printf("%s/%s.xml", config->stateDir, uuidstr);
    return path;
}


static char *
vzctMigFile(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;

    return g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/status.migrating.xml",
                           priv->bundlePath);
}

static char *
vzctMigActiveFile(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;

    return g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/migrating.active.xml",
                           priv->bundlePath);
}


static char *
vzctMigInactiveFile(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;

    return g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/migrating.inactive.xml",
                           priv->bundlePath);
}


static char *
vzctManagedSaveXMLFile(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;

    return g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/managedsave.xml",
                           priv->bundlePath);
}


static char *
vzctVncPidFile(vzctDriverConfig *config, const unsigned char *uuid)
{
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    char *path;

    virUUIDFormat(uuid, uuidstr);
    path = g_strdup_printf("%s/%s.vnc.pid", config->stateDir, uuidstr);
    return path;
}


static int
vzctDomainSaveConfig(vzctDriver *driver, virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    virDomainDef *def = virDomainObjIsActive(dom) ? dom->newDef : dom->def;
    g_autofree char *path = NULL;

    if (priv->migdata)
        path = vzctMigInactiveFile(dom);
    else
        path = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/config.xml", priv->bundlePath);

    return virDomainDefSavePath(def, driver->xmlopt, path,
                                VIR_DOMAIN_DEF_FORMAT_DISK_PRIVATE);
}


/*
 * Loading osinfo db take significant time. We cannot afford doing it on
 * any libosinfo name conversion. Thus we cache osinfo db.
 *
 * Unfortunatly checking osinfo package VERSION file which path is
 * vendor specific is only solution I have to invalidate cache without
 * restarting libvirtd.
 */
static OsinfoDb*
vzctGetOsinfoDb(void)
{
    struct stat sb;
    vzctDriver *driver = vzct_driver;
    g_autoptr(GError) err = NULL;
    OsinfoDb *db = NULL;

    virMutexLock(&driver->lock);
    if (stat(VZCT_LIBOSINFO_VERSION_PATH, &sb) < 0) {
        virReportSystemError(errno, _("Failed to stat '%s'"),
                             VZCT_LIBOSINFO_VERSION_PATH);
        goto cleanup;
    }

    if (!driver->osinfoLoader ||
        driver->osinfoCtime != sb.st_ctime) {

        if (driver->osinfoLoader)
            g_object_unref(driver->osinfoLoader);
        driver->osinfoLoader = osinfo_loader_new();
        osinfo_loader_process_default_path(driver->osinfoLoader, &err);

        if (err) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("cannot load libosinfo data: %s"), err->message);
            g_object_unref(driver->osinfoLoader);
            driver->osinfoLoader = NULL;
            goto cleanup;
        }

        driver->osinfoCtime = sb.st_ctime;
    }

    db = osinfo_loader_get_db(driver->osinfoLoader);

 cleanup:
    virMutexUnlock(&driver->lock);
    return db;
}


/*
 * Convert between libosinfo names. For example to convert
 * short-id=ubuntu14.04 to id=http://ubuntu.com/ubuntu/14.04.
 * In this case arguments are name='ubuntu14.04', from='short-id',
 * to='id'.
 */
static int
vzctOsinfoConvertName(const char *name,
                      const char *from,
                      const char *to,
                      char **result)
{
    OsinfoDb *db;
    g_autoptr(OsinfoList) all = NULL;
    g_autoptr(OsinfoFilter) filter = NULL;
    g_autoptr(OsinfoList) filtered = NULL;
    g_autoptr(GList) filtered_raw = NULL;
    g_autoptr(GList) idlist = NULL;
    GList *tmp = NULL;

    *result = NULL;

    db = vzctGetOsinfoDb();
    if (!(all = (OsinfoList *)osinfo_db_get_os_list(db))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("cannot get OS list"));
        return -1;
    }

    filter = g_object_new(OSINFO_TYPE_PRODUCTFILTER, NULL);
    filtered = g_object_new(OSINFO_TYPE_OSLIST, "element-type",
                            OSINFO_TYPE_OS, NULL);
    osinfo_filter_add_constraint(filter, from, name);
    osinfo_list_add_filtered(filtered, all, filter);

    filtered_raw = osinfo_list_get_elements(filtered);
    if (!filtered_raw)
        return 0;

    idlist = osinfo_entity_get_param_value_list(OSINFO_ENTITY(filtered_raw->data), to);
    if (!idlist) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("empty list is unexpected"));
        return -1;
    }

    /*
     * In case of short-id -> id there should be only single result.
     * In case of id -> short-id we should take last entry as
     * we modify osinfo-db specially for this purpuse.
     */
    tmp = idlist;
    while (tmp->next)
        tmp = tmp->next;

    *result = g_strdup(tmp->data);
    return 0;
}


static int
vzctDomainStateLoad(vzctDriver *driver,
                    virDomainObj *dom,
                    const unsigned char *uuid)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    g_autoptr(virDomainObj) saved = NULL;
    g_autofree char *statefile = vzctStateFile(driver->config, uuid);
    g_autofree char *pidfile = vzctVncPidFile(driver->config, uuid);
    unsigned int parseFlags = VIR_DOMAIN_DEF_PARSE_STATUS |
                              VIR_DOMAIN_DEF_PARSE_ACTUAL_NET |
                              VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                              VIR_DOMAIN_DEF_PARSE_ALLOW_POST_PARSE_FAIL;

    if (!(saved = virDomainObjParseFile(statefile, driver->xmlopt, parseFlags)))
        return -1;
    virObjectUnlock(saved);

    if (dom->def) {
        virDomainDef *def;

        /* FIXME Use active config from vzctl when it will be available. */
        if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
            return -1;
        dom->newDef = g_steal_pointer(&dom->def);
        dom->def = g_steal_pointer(&def);
        if (vzctMergeConfigs(dom->def, saved->def) < 0)
            return -1;
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            return -1;
    } else {
        dom->def = g_steal_pointer(&saved->def);
    }

    virDomainObjSetState(dom, VIR_DOMAIN_RUNNING, -1);
    dom->def->id = 0;

    if (dom->def->ngraphics) {
        virDomainGraphicsDef *gr = dom->def->graphics[0];

        gr->data.vnc.portReserved = true;
        virPortAllocatorSetUsed(gr->data.vnc.port);

        if (virFileExists(pidfile) &&
            virPidFileReadPath(pidfile, &priv->vncPid) < 0)
            return -1;
    }

    /*
     * Timer can be already initialized if we are recovering from unsynced
     * state.
     */
    if (!priv->cleanupTimer) {
        virObjectRef(dom);
        if ((priv->cleanupTimer = virEventAddTimeout(-1,
                                                     vzctDomainCleanupTimeout,
                                                     dom,
                                                     virObjectFreeCallback)) < 0) {
            virObjectUnref(dom);
            return -1;
        }
    }

    return 0;
}


static void
vzctCleanupDriverConfigs(vzctDriver *driver)
{
    DIR *dir;
    struct dirent *entry;
    g_autofree char *path = NULL;
    const char *stateDir = driver->config->stateDir;

    if (virDirOpenIfExists(&dir, stateDir) <= 0)
        return;

    while (virDirRead(dir, &entry, stateDir) > 0) {
        unsigned char uuid[VIR_UUID_BUFLEN];
        char *uuidstr = entry->d_name;
        virDomainObj *dom;

        /* Domain states are stored with filenames uuid.xml */
        if (!virStringStripSuffix(uuidstr, ".xml") ||
            virUUIDParse(uuidstr, uuid) < 0)
            continue;

        if ((dom = virDomainObjListFindByUUID(driver->domains, uuid))) {
            virDomainObjEndAPI(&dom);
            continue;
        }

        if (!(dom = virDomainObjNew(driver->xmlopt)))
            continue;

        if (vzctDomainStateLoad(driver, dom, uuid) < 0) {
            virObjectUnref(dom);
            continue;
        }

        vzctDomainStop(driver, dom);
        virObjectUnref(dom);
    }
    virDirClose(dir);
}


static int
vzctDriverConnectOpen(vzctDriver *driver)
{
    struct sockaddr_nl addr = { 0 };
    g_autoptr(GError) gerr = NULL;
    int fd;
    int ret = -1;

    virMutexLock(&driver->lock);

    VZCT_CHECK(cleanup, vzctl2_init_log("libvirtd"));
    VZCT_CHECK(cleanup, vzctl2_set_log_quiet(1));
    VZCT_CHECK(cleanup, vzctl2_lib_init());

    if (!(driver->eventThread = virEventThreadNew("vzct")))
        goto cleanup;

    VZCT_CHECK(cleanup,
               vzctl2_register_evt(&driver->eventdHandle));

    driver->netlinkFd = -1;

    fd = vzctl2_get_evt_fd(driver->eventdHandle);
    if (!(driver->eventdSocket = g_socket_new_from_fd(fd, &gerr))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unable to create socket object: %s"), gerr->message);
        goto cleanup;
    }

    driver->eventdSource = g_socket_create_source(driver->eventdSocket,
                                                  G_IO_IN, NULL);
    g_source_set_callback(driver->eventdSource,
                          (GSourceFunc)vzctEventdEvent,
                          driver, NULL);
    g_source_attach(driver->eventdSource,
                    virEventThreadGetContext(driver->eventThread));

    if ((driver->netlinkFd = socket(PF_NETLINK, SOCK_DGRAM, 31)) < 0) {
        virReportSystemError(errno,
                             "%s", _("failed to netlink socket"));
        goto cleanup;
    }

    if (virSetCloseExec(driver->netlinkFd) < 0) {
        virReportSystemError(errno, "%s",
                             _("Unable to set netlink socket close-on-exec flag"));
        goto cleanup;
    }

    addr.nl_family = AF_NETLINK;
    addr.nl_groups = 0x01; /* multicast */

    if (bind(driver->netlinkFd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        virReportSystemError(errno, "%s",
                             _("Unable to bind netlink socket"));
        goto cleanup;
    }

    if (!(driver->netlinkSocket = g_socket_new_from_fd(driver->netlinkFd, &gerr))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unable to create socket object: %s"), gerr->message);
        goto cleanup;
    }

    driver->netlinkSource = g_socket_create_source(driver->netlinkSocket,
                                                   G_IO_IN, NULL);
    g_source_set_callback(driver->netlinkSource,
                          (GSourceFunc)vzctNetlinkEvent,
                          driver, NULL);
    g_source_attach(driver->netlinkSource,
                    virEventThreadGetContext(driver->eventThread));

    if (!(driver->domains = virDomainObjListNewFlags(true)))
        goto cleanup;

    ret = 0;

 cleanup:
    virMutexUnlock(&driver->lock);
    return ret;
}


static virDrvOpenStatus
vzctConnectOpen(virConnectPtr conn,
                virConnectAuthPtr auth G_GNUC_UNUSED,
                virConf *conf G_GNUC_UNUSED,
                unsigned int flags)
{
    virCheckFlags(VIR_CONNECT_RO, VIR_DRV_OPEN_ERROR);

    if (!conn->uri)
        return VIR_DRV_OPEN_DECLINED;

    if (!conn->uri->scheme)
        return VIR_DRV_OPEN_DECLINED;

    if (STRNEQ(conn->uri->scheme, "vzct"))
        return VIR_DRV_OPEN_DECLINED;

    /* Remote driver should handle these. */
    if (conn->uri->server)
        return VIR_DRV_OPEN_DECLINED;

    /* From this point on, the connection is for us. */
    if (STRNEQ_NULLABLE(conn->uri->path, "/system")) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unexpected Virtuozzo URI path '%s', try vzct:///system"),
                       conn->uri->path);
        return VIR_DRV_OPEN_ERROR;
    }

    if (virConnectOpenEnsureACL(conn) < 0)
        return VIR_DRV_OPEN_ERROR;

    conn->id = g_atomic_int_add(&vzct_driver->lastconnid, 1) + 1;
    conn->privateData = vzct_driver;

    return VIR_DRV_OPEN_SUCCESS;
}


static int
vzctConnectClose(virConnectPtr conn)
{
    vzctDriver *driver = conn->privateData;
    vzctDriverJob *job;

    job = g_new0(vzctDriverJob, 1);
    job->type = VZCT_DRIVER_JOB_CONNECTION_CLOSED;
    job->conn_id = conn->id;
    if (virThreadPoolSendJob(driver->workers, 0, job) < 0) {
        VIR_WARN("cannot run cleanup job");
        vzctDriverJobFree(job);
    }

    conn->privateData = NULL;
    return 0;
}


static int
vzctDomainNetAddIpAddress(virDomainNetDef *net,
                          vzctl_ip_iterator ips)
{
    virSocketAddr address;
    virNetDevIPAddr *ip;
    unsigned int prefix = 0;
    char *p;
    char buf[512];

    if (vzctl2_env_get_ipstr(ips, buf, sizeof(buf))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("can not get ip address string for net '%s'"),
                       net->ifname_guest);
        return -1;
    }

    /* The above function returns ip in format address/mask */
    if ((p = strchr(buf, '/')))
        *p = '\0';

    if (virSocketAddrParse(&address, buf, AF_UNSPEC) < 0)
        return -1;

    if (p) {
        p++;
        if (VIR_SOCKET_ADDR_IS_FAMILY(&address, AF_INET6)) {
            /* mask is in prefix form for ipv6 */
            if (virStrToLong_uip(p, NULL, 10, &prefix) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("cannot parse prefix '%s'"), p);
                return -1;
            }
        } else if (VIR_SOCKET_ADDR_IS_FAMILY(&address, AF_INET)) {
            virSocketAddr netmask;
            int iprefix;

            /* mask is in address form for ipv4 */
            if (virSocketAddrParseIPv4(&netmask, p) < 0)
                return -1;

            if ((iprefix = virSocketAddrGetNumNetmaskBits(&netmask)) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("netmask '%s' is not valid"), p);
                return -1;
            }

            prefix = iprefix;
        }
    }

    ip = g_new0(virNetDevIPAddr, 1);
    ip->address = address;
    ip->prefix = prefix;
    VIR_APPEND_ELEMENT(net->guestIP.ips, net->guestIP.nips, ip);

    return 0;
}


static int
vzctLoadNetsConfig(vzctl_env_handle_ptr env,
                   virDomainDef *def)
{
    vzctl_env_param_ptr param = vzctl2_get_env_param(env);
    vzctl_veth_dev_iterator it = NULL;
    vzctl_ip_iterator ips;
    vzctl_str_iterator isn;
    vzctl_rate_iterator irs;
    virDomainNetDef *net;
    const char *hostname;
    size_t nnets_max = 0;
    size_t nsearchs = 0;
    int ratebound = 0;

    net = g_new0(virDomainNetDef, 1);
    net->type = VIR_DOMAIN_NET_TYPE_VENET0;
    net->ifname_guest = g_strdup(VZCT_NET_VENET0);
    VIR_RESIZE_N(def->nets, nnets_max, def->nnets, 1);
    def->nets[def->nnets++] = net;

    ips = NULL;
    while ((ips = vzctl2_env_get_ipaddress(param, ips)))
        vzctDomainNetAddIpAddress(net, ips);

    irs = NULL;
    while ((irs = vzctl2_env_get_rate(param, irs))) {
        struct vzctl_rate_param rate_param;

        VZCT_CHECK_RET(-1, vzctl2_env_get_rate_param(irs, &rate_param));
        def->net_classes.rate[rate_param.net_class] = rate_param.rate;
    }

    vzctl2_env_get_ratebound(param, &ratebound);
    def->net_classes.bound = ratebound ? true : false;

    if (!vzctl2_env_get_hostname(param, &hostname))
        def->dns.hostname = g_strdup(hostname);

    isn = NULL;
    while ((isn = vzctl2_env_get_nameserver(param, isn))) {
        const char *server = vzctl2_env_get_str_param(isn);

        VIR_EXPAND_N(def->dns.servers, def->dns.nservers, 1);
        virSocketAddrParse(&def->dns.servers[def->dns.nservers - 1],
                           server, AF_UNSPEC);
    }

    isn = NULL;
    while ((isn = vzctl2_env_get_searchdomain(param, isn))) {
        VIR_EXPAND_N(def->dns.searchs, nsearchs, 1);
        def->dns.searchs[nsearchs - 1] = g_strdup(vzctl2_env_get_str_param(isn));
    }

    /* add NULL termination */
    if (def->dns.searchs)
        VIR_EXPAND_N(def->dns.searchs, nsearchs, 1);

    while ((it = vzctl2_env_get_veth(param, it))) {
        struct vzctl_veth_dev_param net_param;

        vzctl2_env_get_veth_param(it, &net_param, sizeof(net_param));

        /*
         * Bridge or network is required in libvirt. In libvzctl this
         * parameter can be NULL.
         */
        if (!net_param.network)
            net_param.network = "";

        net = g_new0(virDomainNetDef, 1);
        net->ifname = g_strdup(net_param.dev_name);
        net->ifname_guest = g_strdup(net_param.dev_name_ve);
        net->dhcp = net_param.dhcp;
        net->dhcp6 = net_param.dhcp6;
        /* TODO do we need to handle invalid mac? */
        ignore_value(virMacAddrParse(net_param.mac_ve, &net->mac));
        switch (net_param.nettype) {
        case VZCTL_NETTYPE_BRIDGE:
            net->type = VIR_DOMAIN_NET_TYPE_BRIDGE;
            net->data.bridge.brname = g_strdup(net_param.network);
            break;

        case VZCTL_NETTYPE_VNET:
            /* TODO We need to expand network to bridge for running
             * domains. */
            net->type = VIR_DOMAIN_NET_TYPE_NETWORK;
            net->data.network.name = g_strdup(net_param.network);
            break;

        default:
            /* ignore unknown types */
            VIR_DEBUG("Unknown nettype: %d", net_param.nettype);
            virDomainNetDefFree(net);
            continue;
        }
        if (net_param.vporttype == VZCTL_VPORTTYPE_OVS) {

            net->virtPortProfile = g_new0(virNetDevVPortProfile, 1);
            net->virtPortProfile->virtPortType = VIR_NETDEV_VPORT_PROFILE_OPENVSWITCH;
            if (net_param.ifaceid) {
                if (virUUIDParse(net_param.ifaceid,
                                 net->virtPortProfile->interfaceID) < 0) {
                    virReportError(VIR_ERR_INTERNAL_ERROR,
                                   _("cannot parse interfaceid, domain '%s', iface '%s' "),
                                   def->name, net->ifname);
                    return -1;
                }
                net->virtPortProfile->interfaceID_specified = true;
            }
        }

        if (net_param.gw) {
            virNetDevIPRoute *route;

            route = g_new0(virNetDevIPRoute, 1);
            virSocketAddrParse(&route->gateway, net_param.gw, AF_INET);
            virSocketAddrParseIPv4(&route->address, VIR_SOCKET_ADDR_IPV4_ALL);
            route->family = g_strdup("ipv4");
            VIR_APPEND_ELEMENT_COPY(net->guestIP.routes,
                                    net->guestIP.nroutes, route);
        }

        if (net_param.gw6) {
            virNetDevIPRoute *route6;

            route6 = g_new0(virNetDevIPRoute, 1);
            virSocketAddrParse(&route6->gateway, net_param.gw6, AF_INET6);
            virSocketAddrParseIPv6(&route6->address, VIR_SOCKET_ADDR_IPV6_ALL);
            route6->family = g_strdup("ipv6");
            VIR_APPEND_ELEMENT_COPY(net->guestIP.routes,
                                    net->guestIP.nroutes, route6);
        }

        ips = NULL;
        while ((ips = vzctl2_env_get_veth_ipaddress(it, ips)))
            vzctDomainNetAddIpAddress(net, ips);

        VIR_RESIZE_N(def->nets, nnets_max, def->nnets, 1);
        def->nets[def->nnets++] = net;
    }
    VIR_SHRINK_N(def->nets, nnets_max, nnets_max - def->nnets);

    return 0;
}


static void
vzctLoadDisksConfig(vzctDriver *driver,
                    vzctl_env_handle_ptr env,
                    virDomainDef *def)
{
    vzctl_env_param_ptr param = vzctl2_get_env_param(env);
    struct vzctl_disk *it = NULL;
    size_t ndisks_max = 0;

    while ((it = vzctl2_env_get_disk(param, it))) {
        struct vzctl_disk_param disk_param;
        virDomainDiskDef *disk = virDomainDiskDefNew(driver->xmlopt);
        vzctDomainDiskPrivate *diskPrivate =
            (vzctDomainDiskPrivate *)disk->privateData;

        vzctl2_env_get_disk_param(it, &disk_param, sizeof(disk_param));

        disk->device = VIR_DOMAIN_DISK_DEVICE_DISK;
        virDomainDiskSetType(disk, VIR_STORAGE_TYPE_FILE);
        virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_PLOOP);
        virDomainDiskSetDriver(disk, "vzct");
        virDomainDiskSetSource(disk, disk_param.path);
        memcpy(diskPrivate->uuid, disk_param.uuid, sizeof(diskPrivate->uuid));

        if (STREQ_NULLABLE(disk_param.mnt, "/"))
            disk->info.bootIndex = 1;

        VIR_RESIZE_N(def->disks, ndisks_max, def->ndisks, 1);
        def->disks[def->ndisks++] = disk;
    }
    VIR_SHRINK_N(def->disks, ndisks_max, ndisks_max - def->ndisks);
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC(vzctl_env_handle, vzctl2_env_close);

static void
vzctLoadDisksIotunes(vzctl_env_param_ptr param,
                     virDomainDef *def)
{
    unsigned int io, iops;
    size_t i;

    vzctl2_env_get_iolimit(param, &io);
    vzctl2_env_get_iopslimit(param, &iops);

    if (io == UINT_MAX)
        io = 0;
    if (iops == UINT_MAX)
        iops = 0;

    if (!io && !iops)
        return;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        virDomainBlockIoTuneInfo *iotune = &disk->blkdeviotune;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        iotune->total_bytes_sec = io;
        iotune->total_iops_sec = iops;

        if (def->ndisks > 1)
            iotune->group_name = g_strdup("virtuozzo");
    }
}


/* Weight values are taken from libvzctl */
static unsigned int vzctIOPrioWeight[] = {
    320, 365, 410, 460, 500, 550, 600, 640
};


static int
vzctConvertWeightToIOPrio(unsigned int weight)
{
    size_t i;

    for (i = 0; i < G_N_ELEMENTS(vzctIOPrioWeight); i++)
        if (weight <= vzctIOPrioWeight[i])
            return i;

    return G_N_ELEMENTS(vzctIOPrioWeight) - 1;
}


static unsigned int
vzctConvertIOPrioToWeight(int ioprio)
{
    if (ioprio < 0)
        return vzctIOPrioWeight[0];

    if (ioprio >= G_N_ELEMENTS(vzctIOPrioWeight))
        return vzctIOPrioWeight[G_N_ELEMENTS(vzctIOPrioWeight) - 1];

    return vzctIOPrioWeight[ioprio];
}


static virDomainDef *
vzctLoadDomainConfig(vzctDriver *driver G_GNUC_UNUSED,
                     const char *ctid,
                     char **bundlePath)
{
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    virDomainDef *ret = NULL;
    vzctl_env_param_ptr param;
    struct vzctl_mem_guarantee mem_guarantee;
    struct vzctl_cpulimit_param cpu_limit;
    const char *uuid = NULL;
    const char *name = NULL;
    const char *ostemplate;
    const char *path;
    char mask[512];
    g_autoptr(virBitmap) nodeset = NULL;
    unsigned long v;
    int err, rc, ioprio;

    VZCT_CHECK_PTR_RET(NULL, env, vzctl2_env_open(ctid, 0, &err));
    param = vzctl2_get_env_param(env);
    def = virDomainDefNew();

    vzctl2_env_get_uuid(param, &uuid);
    if (!uuid || virUUIDParse(uuid, def->uuid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("failed to parse uuid '%s' from vzctl2_env_get_uuid"),
                       NULLSTR(uuid));
        return NULL;
    }
    vzctl2_env_get_name(env, &name);
    if (name)
        def->name = g_strdup(name);
    else
        def->name = g_strdup(ctid);
    if (bundlePath) {
        VZCT_CHECK_RET(NULL, vzctl2_env_get_ve_private_path(param, &path));
        *bundlePath = g_strdup(path);
    }
    def->extraId = g_strdup(ctid);

    def->virtType = VIR_DOMAIN_VIRT_VZCT;
    def->os.type = VIR_DOMAIN_OSTYPE_EXE;
    def->os.arch = virArchFromHost();
    def->os.init = g_strdup("/sbin/init");

    vzctl2_env_get_ramsize(param, &v);
    def->mem.cur_balloon = ((unsigned long long)v) << 10;
    virDomainDefSetMemoryTotal(def, def->mem.cur_balloon);

    VZCT_CHECK_RET(NULL, vzctl2_env_get_memguarantee(param, &mem_guarantee));
    switch (mem_guarantee.type) {
    case VZCTL_MEM_GUARANTEE_PCT:
        if (mem_guarantee.value > 100)
            mem_guarantee.value = 100;
        def->mem.min_guarantee = def->mem.cur_balloon * mem_guarantee.value / 100;
        break;

    case VZCTL_MEM_GUARANTEE_BYTES:
        /* bytes to KiB */
        def->mem.min_guarantee = mem_guarantee.value >> 10;
        break;

    case VZCTL_MEM_GUARANTEE_AUTO:
        def->mem.min_guarantee_auto = true;
        break;

    default:
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("unexpected memguarantee type=%d"),
                       mem_guarantee.type);
        return NULL;
    }

    rc = vzctl2_env_get_cpulimit(param, &cpu_limit);
    if (rc != 0 && rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_cpulimit);
        return NULL;
    }
    if (rc == 0) {
        def->cputune.period = 100000;

        if (!cpu_limit.limit) {
            def->cputune.quota = -1;
        } else if (cpu_limit.type == VZCTL_CPULIMIT_PCT) {
            def->cputune.quota = (cpu_limit.limit * def->cputune.period) / 100;
        } else {
            struct vzctl_cpuinfo info;

            if (vzctl2_get_cpuinfo(&info)) {
                VZCT_ERROR_MESSAGE(vzctl2_get_cpuinfo);
                return NULL;
            }

            def->cputune.quota = (cpu_limit.limit * def->cputune.period *
                                  1000 * info.ncpu) / info.freq;
        }
    }

    rc = vzctl2_env_get_cpuunits(param, &v);
    if (rc != 0 && rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_cpuunits);
        return NULL;
    }
    if (rc == 0) {
        def->cputune.shares = v;
        def->cputune.sharesSpecified = true;
    }

    vzctl2_env_get_cpu_count(param, &v);
    def->maxvcpus = v;

    rc = vzctl2_env_get_cpumask(param, mask, sizeof(mask));
    if (rc != 0 && rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_cpumask);
        return NULL;
    }
    if (rc == 0 && strlen(mask) > 0 &&
        virBitmapParse(mask, &def->cpumask, VIR_DOMAIN_CPUMASK_LEN) < 0)
        return NULL;

    rc = vzctl2_env_get_nodemask(param, mask, sizeof(mask));
    if (rc != 0 && rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_nodemask);
        return NULL;
    }
    if (rc == 0 && strlen(mask) > 0) {
        if (virBitmapParse(mask, &nodeset, VIR_DOMAIN_CPUMASK_LEN) < 0)
            return NULL;

        if (virDomainNumatuneSet(def->numa,
                                 true,
                                 VIR_DOMAIN_NUMATUNE_PLACEMENT_STATIC,
                                 VIR_DOMAIN_NUMATUNE_MEM_STRICT,
                                 nodeset) < 0)
            return NULL;
    }

    def->onReboot = VIR_DOMAIN_LIFECYCLE_ACTION_RESTART;
    def->onPoweroff = VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY;
    def->onCrash = VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY;

    vzctLoadDisksConfig(driver, env, def);
    vzctLoadDisksIotunes(param, def);
    if (vzctLoadNetsConfig(env, def) < 0)
        return NULL;

    rc = vzctl2_env_get_ioprio(param, &ioprio);
    if (rc != 0 && rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_ioprio);
        return NULL;
    }
    if (rc == 0)
        def->blkio.weight = vzctConvertIOPrioToWeight(ioprio);

    def->ostemplate = NULL;
    rc = vzctl2_env_get_ostemplate(param, &ostemplate);
    if (rc == 0) {
        if (*ostemplate != '\0')
            def->ostemplate = g_strdup(ostemplate);
    } else if (rc != VZCTL_E_NOT_INITIALIZED) {
        VZCT_ERROR_MESSAGE(vzctl2_env_get_ostemplate);
        return NULL;
    }

    ret = g_steal_pointer(&def);
    return ret;
}


static void
vzctGenerateDiskNames(virDomainDef *def)
{
    size_t i;
    int max = -1;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        int idx;

        /*
         * All disks added NOT thru libvirt will be set on virtio
         * bus. This is handy for SolusIO. In future we can make
         * default bus configurable if we need to.
         */
        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK ||
            !disk->dst ||
            !STRPREFIX(disk->dst, "vd"))
            continue;

        idx = virDiskNameToIndex(disk->dst);
        if (idx > max)
            max = idx;
    }

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK ||
            disk->dst)
            continue;

        disk->dst = virIndexToDiskName(++max, "vd");
        disk->bus = VIR_DOMAIN_DISK_BUS_VIRTIO;
    }
}


/*
 * Convert libvzctl ostemplate name to libosinfo short-id.
 * Like from .ubuntu-14.04-x86_64 to ubuntu14.04.
 * More precisely:
 *
 * - strip . prefix
 * - strip . -x86_64 suffix
 * - remove first dash before digit
 */
static char *
vzctConvertFromOstemplateName(const char *name)
{
    g_autoptr(GRegex) regex = NULL;
    g_autoptr(GError) err = NULL;
    g_autoptr(GMatchInfo) info = NULL;

    if (!(regex = g_regex_new("^\\.?(.*?)-([0-9].*?)(-x86_64)?$", 0, 0, &err))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Failed to compile regex %s"), err->message);
        goto fail;
    }

    if (!g_regex_match(regex, name, 0, &info)) {
        VIR_WARN("cannot convert ostemplate '%s'", name);
        goto fail;
    }

    return g_strdup_printf("%s%s",
                           g_match_info_fetch(info, 1),
                           g_match_info_fetch(info, 2));

 fail:
    return g_strdup(name);
}


static int
vzctUpdateLibosinfoMetadata(virDomainDef *def,
                            const char *libosid)
{
    g_autofree char *meta = NULL;

    if (virXMLExtractNamespaceXML(def->metadata, VZCT_LIBOSINFO_URI, &meta) < 0)
        return -1;

    if (meta) {
        g_autoptr(xmlDoc) doc = NULL;
        g_autoptr(xmlXPathContext) ctxt = NULL;
        xmlNodePtr root_e;
        xmlNodePtr os_e;

        if (!(doc = virXMLParseStringCtxt(meta, "libosinfo", &ctxt)))
            return -1;

        if (!(root_e = virXPathNode("/libosinfo", ctxt))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("invalid root element"));
            return -1;
        }

        if (!(os_e = virXPathNode("/libosinfo/os", ctxt)) &&
            !(os_e = xmlNewChild(root_e, NULL, BAD_CAST "os", NULL))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("cannot create 'os' element"));
            return -1;
        }

        if (!xmlSetProp(os_e, BAD_CAST "id", BAD_CAST(libosid ?: ""))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("cannot set 'id' attribute"));
            return -1;
        }

        VIR_FREE(meta);
        if (!(meta = virXMLNodeToString(doc, root_e)))
            return -1;
    } else if (libosid) {
        g_auto(virBuffer) buf = VIR_BUFFER_INITIALIZER;

        virBufferAddLit(&buf, "<libosinfo>\n");
        virBufferAdjustIndent(&buf, 2);
        virBufferAsprintf(&buf, "<os id='%s'/>\n", libosid);
        virBufferAdjustIndent(&buf, -2);
        virBufferAddLit(&buf, "</libosinfo>\n");

        meta = virBufferContentAndReset(&buf);
    }

    if (virDomainDefSetMetadata(def, VIR_DOMAIN_METADATA_ELEMENT,
                                meta, "libosinfo", VZCT_LIBOSINFO_URI) < 0)
        return -1;

    return 0;
}


/*
 * Show ostemplate in VHS metadata in form:
 *
 * <vz>
 *   <ostemplate name="foo"/>
 * </vz>
 */
static int
vzctUpdateVHSMetadata(virDomainDef *def,
                      const char *ostemplate)
{
    g_autofree char *meta = NULL;

    if (virXMLExtractNamespaceXML(def->metadata, VZCT_VHS_URI, &meta) < 0)
        return -1;

    if (meta) {
        g_autoptr(xmlDoc) doc = NULL;
        g_autoptr(xmlXPathContext) ctxt = NULL;
        xmlNodePtr root_e;
        xmlNodePtr os_e;

        if (!(doc = virXMLParseStringCtxt(meta, "vz", &ctxt)))
            return -1;

        if (!(root_e = virXPathNode("/vz", ctxt))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("invalid root element"));
            return -1;
        }

        if (!(os_e = virXPathNode("/vz/ostemplate", ctxt)) &&
            !(os_e = xmlNewChild(root_e, NULL, BAD_CAST "ostemplate", NULL))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("cannot create 'ostemplate' element"));
            return -1;
        }

        if (!xmlSetProp(os_e, BAD_CAST "name", BAD_CAST(ostemplate ?: ""))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("cannot set 'name' attribute"));
            return -1;
        }

        VIR_FREE(meta);
        if (!(meta = virXMLNodeToString(doc, root_e)))
            return -1;
    } else if (ostemplate) {
        g_auto(virBuffer) buf = VIR_BUFFER_INITIALIZER;

        virBufferAddLit(&buf, "<vz>\n");
        virBufferAdjustIndent(&buf, 2);
        virBufferAsprintf(&buf, "<ostemplate name='%s'/>\n", ostemplate);
        virBufferAdjustIndent(&buf, -2);
        virBufferAddLit(&buf, "</vz>\n");

        meta = virBufferContentAndReset(&buf);
    }

    if (virDomainDefSetMetadata(def, VIR_DOMAIN_METADATA_ELEMENT,
                                meta, "vz", VZCT_VHS_URI) < 0)
        return -1;

    return 0;
}


static int
vzctUpdateMetadata(virDomainDef *def)
{
    g_autofree char *ostemplate = g_steal_pointer(&def->ostemplate);
    g_autofree char *libosid = NULL;

    if (ostemplate) {
        g_autofree char *libosname = NULL;

        libosname = vzctConvertFromOstemplateName(ostemplate);
        if (vzctOsinfoConvertName(libosname, "short-id", "id", &libosid) < 0)
            return -1;

        if (libosid)
            VIR_FREE(ostemplate);
    }

    if (vzctUpdateVHSMetadata(def, ostemplate) < 0 ||
        vzctUpdateLibosinfoMetadata(def, libosid) < 0)
        return -1;

    return 0;
}


static int
vzctMergeConfigs(virDomainDef *def,
                 virDomainDef *drvDef)
{
    size_t i;

    if (!drvDef)
        goto finish;

    def->id = drvDef->id;

    if (drvDef->cputune.period) {
        def->cputune.quota = def->cputune.quota > 0 ?
                             ((def->cputune.quota * drvDef->cputune.period) /
                              def->cputune.period) : drvDef->cputune.quota;
        def->cputune.period = drvDef->cputune.period;
    }

    if (drvDef->ngraphics) {
        def->ngraphics = drvDef->ngraphics;
        drvDef->ngraphics = 0;
        def->graphics = g_steal_pointer(&drvDef->graphics);
    }

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        vzctDomainDiskPrivate *diskPrivate =
            (vzctDomainDiskPrivate *)disk->privateData;
        size_t j;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        for (j = 0; j < drvDef->ndisks; j++) {
            virDomainDiskDef *drvDisk = drvDef->disks[j];
            vzctDomainDiskPrivate *drvDiskPrivate =
                (vzctDomainDiskPrivate *)drvDisk->privateData;

            if ((*drvDiskPrivate->uuid &&
                 STREQ_NULLABLE(diskPrivate->uuid, drvDiskPrivate->uuid)) ||
                (!*drvDiskPrivate->uuid &&
                 STREQ_NULLABLE(disk->src->path, drvDisk->src->path))) {
                disk->dst = g_strdup(drvDisk->dst);
                disk->bus = drvDisk->bus;

                if (drvDisk->blkdeviotune.group_name)
                    disk->blkdeviotune.group_name =
                        g_strdup(drvDisk->blkdeviotune.group_name);

                disk->info.alias = g_strdup(drvDisk->info.alias);

                disk->src->type = drvDisk->src->type;
                disk->src->srcpool = g_steal_pointer(&drvDisk->src->srcpool);
                break;
            }
        }
    }

    for (i = 0; i < drvDef->ndisks; i++) {
        if (drvDef->disks[i]->device == VIR_DOMAIN_DISK_DEVICE_CDROM) {
            VIR_EXPAND_N(def->disks, def->ndisks, 1);
            /*
             * This will make a hole in drvDef but the latter is not supposed
             * to be used after call to this function anyway.
             */
            def->disks[def->ndisks - 1] = g_steal_pointer(&drvDef->disks[i]);
            break;
        }

    }

    for (i = 0; i < def->nnets; i++) {
        virDomainNetDef *net = def->nets[i];
        size_t j;

        for (j = 0; j < drvDef->nnets; j++) {
            virDomainNetDef *drvNet = drvDef->nets[j];

            /* venet0 does not have a device on host */
            if ((net->ifname && drvNet->ifname &&
                 STREQ(net->ifname, drvNet->ifname)) ||
                (net->ifname_guest && drvNet->ifname_guest &&
                 STREQ(net->ifname_guest, drvNet->ifname_guest))) {
                net->info.alias = g_strdup(drvNet->info.alias);
                net->filter = g_steal_pointer(&drvNet->filter);
                net->filterparams = g_steal_pointer(&drvNet->filterparams);
                break;
            }
        }
    }

    def->metadata = g_steal_pointer(&drvDef->metadata);

 finish:
    if (vzctUpdateMetadata(def) < 0)
        return -1;

    vzctGenerateDiskNames(def);
    return 0;
}


#define VZCT_VNC_SERVER_BIN "/usr/bin/prl_vzvncserver_app"

static int
vzctStartVNCServer(vzctDriver *driver,
                   virDomainObj *dom)
{
    g_autofree char *pidfile = vzctVncPidFile(driver->config, dom->def->uuid);
    vzctDomainObjPrivate *priv = dom->privateData;
    virDomainGraphicsDef *gr;
    g_autoptr(virCommand) cmd = NULL;
    int outfd = 1;
    int errfd = 2;

    if (!dom->def->ngraphics)
        return 0;

    gr = dom->def->graphics[0];

    if (priv->migdata) {
        if (priv->migdata->vncPort) {
            /* FIXME data from buggy peer can be inconsistent and
             * migdata.vncPort != gr.data.vnc.port for manual port */
            gr->data.vnc.port = priv->migdata->vncPort;
            priv->migdata->vncPort = 0;
        }
    } else if (gr->data.vnc.autoport) {
        unsigned short port;

        if (virPortAllocatorAcquire(driver->vncPorts, &port) < 0)
            return -1;

        gr->data.vnc.port = port;
    } else if (virPortAllocatorSetUsed(gr->data.vnc.port) < 0) {
        return -1;
    }
    gr->data.vnc.portReserved = true;

    cmd = virCommandNew(VZCT_VNC_SERVER_BIN);

    virCommandAddArgList(cmd, "--listen", gr->listens[0].address, NULL);
    virCommandAddArgFormat(cmd, "--port=%d", gr->data.vnc.port);
    /* otherwise ipv6 on :: will be started */
    virCommandAddArg(cmd, "--portv6=-1");

    if (gr->data.vnc.auth.passwd) {
        virCommandAddArg(cmd, "--passwd");
        virCommandSetInputBuffer(cmd, gr->data.vnc.auth.passwd);
    }

    /*
     * prl_vzvncserver_app closes stdout and stderr after writing
     * initial startup messages if no verbosity mode is set.
     */
    virCommandSetOutputFD(cmd, &outfd);
    virCommandSetErrorFD(cmd, &errfd);

    virCommandAddArg(cmd, dom->def->name);

    if (unlink(pidfile) < 0 && errno != ENOENT) {
        virReportSystemError(errno,
                             _("Cannot remove stale VNC PID file %s"), pidfile);
        return -1;
    }

    virCommandSetPidFile(cmd, pidfile);
    virCommandDaemonize(cmd);

    if (virCommandRun(cmd, NULL) < 0)
        return -1;

    if (virPidFileReadPath(pidfile, &priv->vncPid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Failed to read vnc server pid"));
        return -1;
    }

    return 0;
}


static void
vzctStopVNCServer(vzctDriver *driver, virDomainObj *dom)
{
    g_autofree char *pidfile = vzctVncPidFile(driver->config, dom->def->uuid);
    vzctDomainObjPrivate *priv = dom->privateData;
    virDomainGraphicsDef *gr;

    if (!dom->def->ngraphics)
        return;

    gr = dom->def->graphics[0];
    /* FIXME Check pid belongs to process of VZCT_VNC_SERVER_BIN. */
    if (priv->vncPid) {
        virProcessKillPainfully(priv->vncPid, true);
        priv->vncPid = 0;
    }
    if (gr->data.vnc.portReserved)
        virPortAllocatorRelease(gr->data.vnc.port);

    if (unlink(pidfile) < 0 && errno != ENOENT)
        VIR_WARN("Cannot remove VNC PID file: %s, errno: %d", pidfile, errno);
}


static void
vzctDomainFinishExecProcs(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    size_t i;

    for (i = 0; i < priv->exec_nprocs; i++) {
        vzctDomainExecProcess *exec_proc = priv->exec_procs[i];
        int status;
        int rc;

        if (!exec_proc->terminated)
            vzctl2_env_exec_terminate(exec_proc->handle);

        VIR_DEBUG("Reaping process pid=%d", exec_proc->handle->pid);
        rc = vzctl2_env_waitpid(exec_proc->handle, 0, &status);
        if (rc < 0)
            VIR_WARN("Fail reaping process pid=%d, err: %s",
                     exec_proc->handle->pid, NULLSTR(vzctl2_get_last_error()));

        vzctl2_release_exec_handle(exec_proc->handle);
    }

    if (priv->cleanupTimer) {
        virEventRemoveTimeout(priv->cleanupTimer);
        priv->cleanupTimer = 0;
    }
    VIR_FREE(priv->exec_procs);
    priv->exec_nprocs = 0;
    priv->exec_nprocs_max = 0;
}


static void
vzctMigrationReapServer(pid_t pid)
{
    int status;
    size_t i;
    int rc;

    kill(pid, SIGTERM);

    /* Try to reap for 10s */
    for (i = 0; i < 50; i++) {
        while ((rc = waitpid(pid, &status, WNOHANG)) == -1 &&
               errno == EINTR);

        if (rc < 0) {
            VIR_WARN("Error reaping migration server pid '%d', errno '%d'",
                     pid, errno);
            return;
        }
        else if (rc == pid)
            return;

        usleep(200 * 1000);
    }

    VIR_WARN("Cannot reap migration server in time pid '%d'", pid);
    kill(pid, SIGKILL);
}


static void
vzctMigrationDataCleanup(vzctDomainMigrationData *migdata)
{
    size_t i;

    if (!migdata)
        return;

    if (migdata->vncPort)
        virPortAllocatorRelease(migdata->vncPort);

    for (i = 0; i < VZCT_MIGRATION_CHANNELS_NUM; i++) {
        virProcessAbort(migdata->socatPids[i]);

        if (migdata->ports[i])
            virPortAllocatorRelease(migdata->ports[i]);
    }

    /*
     * Because virProcessAbort gives very little time to terminate.
     * If we terminate migration server too fast it may fail to cleanup
     * disk data which can result in further errors.
     */
    vzctMigrationReapServer(migdata->pid);

    VIR_FREE(migdata);
}


static void
vzctDomainStopFlags(vzctDriver *driver,
                    virDomainObj *dom,
                    bool skipEnv)
{
    g_autofree char *statefile = vzctStateFile(driver->config, dom->def->uuid);
    g_autoptr(vzctl_env_handle) env = NULL;
    int err;

    if (!skipEnv && !(env = vzctl2_env_open(dom->def->extraId, 0, &err)))
        VZCT_ERROR_MESSAGE(vzctl2_env_open);

    vzctDomainFinishExecProcs(dom);
    vzctStopVNCServer(driver, dom);
    virDomainConfVMNWFilterTeardown(dom);

    if (env)
        vzctDomainClearVolumes(dom->def, env);
    virDomainObjRemoveTransientDef(dom);
    virDomainObjSetState(dom, VIR_DOMAIN_SHUTOFF, -1);

    if (unlink(statefile) < 0 && errno != ENOENT)
        VIR_WARN("Cannot remove state, file: %s, error: %d",
                 statefile, errno);
}


static void
vzctDomainStop(vzctDriver *driver,
               virDomainObj *dom)
{
    vzctDomainStopFlags(driver, dom, false);
}


static int
vzctDomainStart(vzctDriver *driver,
                virDomainObj *dom,
                virDomainDef *drvDef)
{
    g_autofree char *statefile = vzctStateFile(driver->config, dom->def->uuid);
    g_autoptr(virDomainDef) drvDefCopy = NULL;
    vzctDomainObjPrivate *priv = dom->privateData;
    virDomainDef *def;
    int ret = -1;

    /* FIXME Use active config from vzctl when it will be available. */
    if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
        return -1;
    dom->newDef = g_steal_pointer(&dom->def);
    dom->def = g_steal_pointer(&def);

    if (!drvDef)
        drvDef = dom->newDef;
    /* Make a copy as merge damages second argument. */
    if (!(drvDefCopy = vzctDomainDefCopy(drvDef, driver->xmlopt, false)))
        goto cleanup;
    if (vzctMergeConfigs(dom->def, drvDefCopy) < 0)
        goto cleanup;

    virDomainObjSetState(dom, VIR_DOMAIN_RUNNING, -1);
    dom->def->id = 0;

    if (vzctStartVNCServer(driver, dom) < 0)
        goto cleanup;

    virObjectRef(dom);
    if ((priv->cleanupTimer = virEventAddTimeout(-1,
                                                 vzctDomainCleanupTimeout,
                                                 dom,
                                                 virObjectFreeCallback)) < 0) {
        virObjectUnref(dom);
        goto cleanup;
    }

    if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    if (ret < 0)
        vzctDomainStop(driver, dom);

    return ret;
}


static virDomainObj *
vzctAddDomain(vzctDriver *driver,
              const char *ctid,
              virDomainObj *dom)
{
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(virDomainDef) drvDef = NULL;
    g_autoptr(virDomainDef) migActiveDef = NULL;
    vzctl_env_status_t status;
    virDomainState state = VIR_DOMAIN_SHUTOFF;
    vzctDomainObjPrivate *priv;
    g_autofree char *path = NULL;
    g_autofree char *statefile = NULL;
    g_autoptr(virDomainObj) mig = NULL;
    g_autofree char *autostartFile = NULL;
    virDomainDef *migDef = NULL;
    bool failure = true;
    int rc = 0;

    if (!dom) {
        g_autofree char *bundlePath = NULL;
        char uuidstr[VIR_UUID_STRING_BUFLEN];

        if (!(def = vzctLoadDomainConfig(driver, ctid, &bundlePath)))
            goto cleanup;

        if (!(dom = virDomainObjListAdd(driver->domains, def, driver->xmlopt, 0)))
            goto cleanup;

        /* When we create new domain thru defineXML ADDED event will be fired
         * before creating is finished thus the def readed will be not complete.
         * So after creation is finished and dom unlocked we update config
         * to the outdated version. Thus let's skip update in this case.
         * May be we just should create and config domain in one go in defineXML.
         */
        if (dom->def)
            return dom;

        priv = dom->privateData;
        priv->bundlePath = g_steal_pointer(&bundlePath);
        dom->persistent = true;

        virUUIDFormat(def->uuid, uuidstr);
        priv->migdata = virHashAtomicSteal(driver->migrating, uuidstr);
        VIR_DEBUG("domain CTID '%s', migrating '%d'", ctid, !!priv->migdata);

        vzctDomainLoadSnapshotMetadata(driver, dom);
    } else {
        priv = dom->privateData;
        VIR_FREE(dom->newDef);

        if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
            goto cleanup;
    }
    VIR_FREE(dom->def);
    dom->def = g_steal_pointer(&def);
    dom->def->id = -1;
    virDomainObjSetState(dom, VIR_DOMAIN_SHUTOFF, -1);

    if (priv->migdata) {
        path = vzctMigInactiveFile(dom);

        /* Backcompat case, just use config file at usual place */
        if (!virFileExists(path))
            VIR_FREE(path);
    }

    if (!path)
        path = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/config.xml", priv->bundlePath);
    if (virFileExists(path)) {
        unsigned int parseFlags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                                  VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                                  VIR_DOMAIN_DEF_PARSE_ALLOW_POST_PARSE_FAIL |
                                  VIR_DOMAIN_DEF_PARSE_DISK_PRIVATE;
        VIR_DEBUG("loading driver def from file '%s'", path);
        if (!(drvDef = virDomainDefParseFile(path, driver->xmlopt,
                                             NULL, parseFlags)))
            goto cleanup;
    } else if (priv->migdata) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("cannot find migrating inactive config, CTID '%s'"),
                       ctid);
    }

    if (vzctMergeConfigs(dom->def, drvDef) < 0)
        goto cleanup;
    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_RUNNING));
    /* TODO examine state more carefully */
    VIR_DEBUG("container: %s, status: %d", dom->def->name, status.mask);
    if (status.mask & ENV_STATUS_RUNNING)
        state = VIR_DOMAIN_RUNNING;

    if (priv->migdata) {
        g_autofree char *migactive = vzctMigActiveFile(dom);
        g_autofree char *migfile = vzctMigFile(dom);

        if (virFileExists(migactive)) {
            if (!(migActiveDef = virDomainDefParseFile(
                                        migactive, driver->xmlopt,
                                        NULL, VIR_DOMAIN_DEF_PARSE_INACTIVE)))
                goto cleanup;
            migDef = migActiveDef;
        } else if (virFileExists(migfile)) {
            /* backcompat case */
            if (!(mig = virDomainObjParseFile(migfile, driver->xmlopt,
                                              VIR_DOMAIN_DEF_PARSE_STATUS |
                                              VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE)))
                goto cleanup;
            virObjectUnlock(mig);
            migDef = mig->def;
        } else {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("cannot find migrating active config, CTID '%s'"),
                           ctid);
            goto cleanup;
        }
    }

    statefile = vzctStateFile(driver->config, dom->def->uuid);
    if (virFileExists(statefile))
        rc = vzctDomainStateLoad(driver, dom, dom->def->uuid);
    else if (state == VIR_DOMAIN_RUNNING)
        rc = vzctDomainStart(driver, dom, migDef ?: NULL);
    if (rc < 0)
        goto cleanup;
    if (state == VIR_DOMAIN_SHUTOFF &&
        dom->state.state == VIR_DOMAIN_RUNNING) {
        vzctDomainStop(driver, dom);
    }

    autostartFile = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/autostart",
                                    priv->bundlePath);
    if (virFileExists(autostartFile))
        dom->autostart = true;

    failure = false;

 cleanup:
    if (dom && failure) {
        VIR_WARN("Can not sync state for domain ctid=%s", dom->def->extraId);
        virDomainObjSetState(dom, VIR_DOMAIN_NOSTATE, -1);
        dom->def->id = -1;
    }

    return dom;
}


/*
 * Try to sync domain if it is out of sync.
 * Returns:
 *  1 - domain is in sync
 * -1 - sync error
 *  0 - domain synced sucessfully
 */
static int
vzctCheckSync(vzctDriver *driver, virDomainObj *dom)
{
    g_autofree char *oldname = NULL;
    virObjectEvent *event;

    if (dom->state.state != VIR_DOMAIN_NOSTATE)
        return 1;

    oldname = g_strdup(dom->def->name);

    vzctAddDomain(driver, NULL, dom);
    if (dom->state.state == VIR_DOMAIN_NOSTATE) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("domain is out of sync"));
        return -1;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_DEFINED,
                                     VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    if (STRNEQ(dom->def->name, oldname) &&
        virDomainObjListRename(driver->domains, dom, dom->def->name,
                               0, NULL, NULL) < 0) {
        virDomainObjSetState(dom, VIR_DOMAIN_NOSTATE, -1);
        dom->def->id = -1;
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("domain is out of sync"));
        return -1;
    }

    return 0;
}


static int
vzctUpdateDomain(vzctDriver *driver,
                 virDomainObj *dom)
{
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(virDomainDef) oldDef = NULL;
    g_autoptr(virDomainDef) oldNewDef = NULL;

    /*
     * If domain is active then use active config from
     * vzctl when it will be available.
     */
    if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
        return -1;

    if (STRNEQ(dom->def->name, def->name) &&
        virDomainObjListRename(driver->domains, dom, def->name,
                               0, NULL, NULL) < 0)
        return -1;
    if (memcmp(dom->def->uuid, def->uuid, VIR_UUID_BUFLEN) != 0 &&
        virDomainObjListUpdateUUID(driver->domains, dom, def->uuid) < 0)
        return -1;
    oldDef = g_steal_pointer(&dom->def);

    if (vzctMergeConfigs(def, oldDef) < 0)
        return -1;
    dom->def = g_steal_pointer(&def);

    if (virDomainObjIsActive(dom)) {
        if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
            return -1;
        oldNewDef = g_steal_pointer(&dom->newDef);
        if (vzctMergeConfigs(def, oldNewDef) < 0)
            return -1;
        dom->newDef = g_steal_pointer(&def);
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        return -1;
    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            return -1;
    }

    return 0;
}


static int
vzctStateCleanup(void)
{
    vzctDriver *driver = vzct_driver;

    /*
     * In unpriviledged mode vzct driver initialization is skipped. Also
     * cleanup is called even on initialization failure.
     */
    if (!vzct_driver)
        return 0;

    virThreadPoolFree(driver->workers);
    vzctDriverConnectClose(driver);
    virObjectUnref(driver->xmlopt);
    virObjectUnref(driver->caps);
    virMutexDestroy(&driver->lock);
    virObjectUnref(driver->domains);
    VIR_FREE(driver->config->stateDir);
    VIR_FREE(driver->config);
    g_object_unref(driver->eventThread);
    virPortAllocatorRangeFree(driver->vncPorts);
    virPortAllocatorRangeFree(driver->migrationPorts);
    virObjectUnref(driver->eventsManager);
    if (driver->osinfoLoader)
        g_object_unref(driver->osinfoLoader);
    virObjectUnref(driver->migrating);
    VIR_FREE(driver);
    vzct_driver = NULL;
    return 0;
}


static void*
vzctDomainObjPrivateAlloc(void *opaque G_GNUC_UNUSED)
{
    vzctDomainObjPrivate *priv = g_new0(vzctDomainObjPrivate, 1);

    if (virCondInit(&priv->jobCond) < 0) {
        virReportSystemError(errno, "%s", _("cannot initialize job condition"));
        VIR_FREE(priv);
        return NULL;
    }

    priv->snapshots = virDomainSnapshotObjListNew();
    return priv;
}


static void
vzctDomainObjPrivateFree(void *data)
{
    vzctDomainObjPrivate *priv = data;

    virCondDestroy(&priv->jobCond);
    virDomainSnapshotObjListFree(priv->snapshots);
    VIR_FREE(priv->bundlePath);
    VIR_FREE(priv);
}


static int
vzctDomainDeviceDefPostParse(virDomainDeviceDef *dev,
                             const virDomainDef *def G_GNUC_UNUSED,
                             unsigned int parseFlags G_GNUC_UNUSED,
                             void *opaque G_GNUC_UNUSED,
                             void *parseOpaque G_GNUC_UNUSED)
{
    virDomainDiskDef *disk;

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        disk = dev->data.disk;

        if (!virDomainDiskGetDriver(disk))
            virDomainDiskSetDriver(disk, "vzct");

        if (!disk->src->format) {
            if (disk->device == VIR_DOMAIN_DISK_DEVICE_DISK)
                virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_PLOOP);
            else
                virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_RAW);
        }

        break;
    }

    return 0;
}


static int
vzctValidateDiskIoTune(const virDomainDef *def,
                       const virDomainDiskDef *disk)
{
    const virDomainBlockIoTuneInfo *iotune = &disk->blkdeviotune;
    size_t i;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK) {
        if (virDomainBlockIoTuneInfoHasAny(iotune)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("iotune is only supported for disks devices"));
            return -1;
        }

        return 0;
    }

    if (iotune->read_bytes_sec ||
        iotune->write_bytes_sec ||
        iotune->read_iops_sec ||
        iotune->write_iops_sec ||
        virDomainBlockIoTuneInfoHasMax(iotune) ||
        virDomainBlockIoTuneInfoHasMaxLength(iotune)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only total_bytes_sec and total_iops_sec"
                         " can be tuned for disks"));
        return -1;
    }

    if (!iotune->group_name)
        return 0;

    for (i = 0; i < def->ndisks; i++) {
        const virDomainDiskDef *d = def->disks[i];

        if (STRNEQ_NULLABLE(iotune->group_name, d->blkdeviotune.group_name))
            continue;

        if (!virDomainBlockIoTuneInfoEqual(iotune, &d->blkdeviotune)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("different iotunes for disks %s and %s"),
                           disk->dst, d->dst);
            return -1;
        }
    }

    return 0;
}


static int
vzctDeviceDefValidateCallback(const virDomainDeviceDef *dev,
                              const virDomainDef *def G_GNUC_UNUSED,
                              void *opaque G_GNUC_UNUSED,
                              void *parseOpaque G_GNUC_UNUSED)
{
    virDomainDiskDef *disk;
    virDomainNetDef *net;
    virDomainGraphicsDef *gr;
    virSocketAddr address;
    bool set_IPv4 = false;
    bool set_IPv6 = false;
    size_t i;

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        disk = dev->data.disk;
        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK &&
            disk->device != VIR_DOMAIN_DISK_DEVICE_CDROM) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only cdrom and disks are supported"));
            return -1;
        }

        if (disk->src->type != VIR_STORAGE_TYPE_FILE &&
            disk->src->type != VIR_STORAGE_TYPE_VOLUME) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only files and volume are supported for disks"));
            return -1;
        }

        if (disk->device == VIR_DOMAIN_DISK_DEVICE_DISK &&
            disk->src->format != VIR_STORAGE_FILE_PLOOP) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only ploop files are supported for disks"));
            return -1;
        }

        if (disk->device == VIR_DOMAIN_DISK_DEVICE_CDROM &&
            disk->src->format != VIR_STORAGE_FILE_RAW) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only raw files are supported for cdroms"));
            return -1;
        }

        if (STRNEQ(virDomainDiskGetDriver(disk), "vzct")) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only vzct driver is supported for disks"));
            return -1;
        }

        if ((disk->info.bootIndex > 1 &&
             disk->device == VIR_DOMAIN_DISK_DEVICE_DISK) ||
            (disk->info.bootIndex > 0 &&
             disk->device == VIR_DOMAIN_DISK_DEVICE_CDROM)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only one boot disk is supported and it must"
                             " have boot index 1"));
            return -1;
        }

        if (vzctValidateDiskIoTune(def, disk) < 0)
            return -1;

        break;

    case VIR_DOMAIN_DEVICE_NET:
        net = dev->data.net;
        if (net->type != VIR_DOMAIN_NET_TYPE_BRIDGE &&
            net->type != VIR_DOMAIN_NET_TYPE_NETWORK &&
            net->type != VIR_DOMAIN_NET_TYPE_VENET0) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only interfaces of type bridge, network"
                             " and venet0 are supported"));
            return -1;
        }

        if (net->type == VIR_DOMAIN_NET_TYPE_VENET0) {
            virMacAddr zeromac;

            if (net->ifname_guest &&
                STRNEQ(net->ifname_guest, VZCT_NET_VENET0)) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Guest dev name can not be other than venet0"));
                return -1;
            }

            if (net->ifname) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Host dev name can not be specified for venet0"));
                return -1;
            }

            memset(zeromac.addr, 0, sizeof(zeromac.addr));
            if (!net->mac_generated &&
                memcmp(net->mac.addr, zeromac.addr, sizeof(zeromac.addr))) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("mac can not be specified for venet0"));
                return -1;
            }

            break;
        }

        if (!net->ifname_guest) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Guest dev name must be specified"));
            return -1;
        }
        if (net->virtPortProfile &&
            net->virtPortProfile->virtPortType != VIR_NETDEV_VPORT_PROFILE_OPENVSWITCH) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only explicit openvswitch virtual port type"
                             " is supported"));
            return -1;
        }

        for (i = 0; i < net->guestIP.nroutes; i++) {
            virNetDevIPRoute *route = net->guestIP.routes[i];

            if (VIR_SOCKET_ADDR_IS_FAMILY(&route->gateway, AF_INET)) {
                if (set_IPv4) {
                    virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                                   _("Only one IPv4 gateway is supported"));
                    return -1;
                }

                set_IPv4 = true;
            }

            if (VIR_SOCKET_ADDR_IS_FAMILY(&route->gateway, AF_INET6)) {
                if (set_IPv6) {
                    virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                                   _("Only one IPv6 gateway is supported"));
                    return -1;
                }

                set_IPv6 = true;
            }

            if (virNetDevIPRouteGetPrefix(route)) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Only the default gateway is supported"));
                return -1;
            }
        }
        break;

    case VIR_DOMAIN_DEVICE_GRAPHICS:
        gr = dev->data.graphics;
        if (gr->type != VIR_DOMAIN_GRAPHICS_TYPE_VNC) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only VNC graphics is supported."));
            return -1;
        }
        if (gr->data.vnc.websocket != 0) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Websocket VNC ports are not supported."));
            return -1;
        }
        if (gr->data.vnc.keymap != 0 &&
            STRNEQ(gr->data.vnc.keymap, "en-us")) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only 'en-us' VNC keymap is supported."));
            return -1;
        }
        if (gr->data.vnc.sharePolicy == VIR_DOMAIN_GRAPHICS_VNC_SHARE_ALLOW_EXCLUSIVE) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("VNC exclusive share policy is not supported."));
            return -1;
        }
        if (gr->data.vnc.auth.connected == VIR_DOMAIN_GRAPHICS_AUTH_CONNECTED_FAIL ||
            gr->data.vnc.auth.connected == VIR_DOMAIN_GRAPHICS_AUTH_CONNECTED_KEEP) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("VNC supports only disconnecton on passwd change."));
            return -1;
        }
        if (gr->data.vnc.auth.expires) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("VNC expiring passwords is not supported."));
            return -1;
        }
        if (gr->nListens > 1) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only one graphics listen address is supported."));
            return -1;
        }
        if (gr->listens[0].type != VIR_DOMAIN_GRAPHICS_LISTEN_TYPE_ADDRESS) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only listening on inet address is supported for graphics"));
            return -1;
        }
        if (virSocketAddrParse(&address, gr->listens[0].address, AF_UNSPEC) < 0)
            return -1;
        if (VIR_SOCKET_ADDR_IS_FAMILY(&address, AF_INET6)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Listening on IPv6 addresses is not supported"));
            return -1;
        }

    }

    return 0;
}


static int
vzctDomainDefValidateGraphics(const virDomainDef *def)
{
    virDomainGraphicsDef *gr;

    if (!def->ngraphics)
        return 0;

    if (def->ngraphics > 1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only one graphics per domain is supported."));
        return -1;
    }

    gr = def->graphics[0];

    if (!gr->data.vnc.autoport &&
        gr->data.vnc.port == 0) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("invalid vnc port %d"), gr->data.vnc.port);
        return -1;
    }

    return 0;
}


static int
vzctDomainDefValidateIoTune(const virDomainDef *def)
{
    size_t i;
    int tuneCount = 0;
    int groupCount = 0;
    int diskCount = 0;
    const char *group_name = NULL;

    for (i = 0; i < def->ndisks; i++) {
        const virDomainDiskDef *disk = def->disks[i];
        const virDomainBlockIoTuneInfo *iotune  = &disk->blkdeviotune;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        diskCount++;

        if (iotune->group_name) {
            if (group_name) {
                if (STRNEQ(group_name, iotune->group_name))
                    goto error;
            } else {
                group_name = iotune->group_name;
            }
            groupCount++;
        }

        if (virDomainBlockIoTuneInfoHasAny(iotune))
            tuneCount++;
    }

    if ((!tuneCount && !groupCount) ||
        (tuneCount == 1 && diskCount == 1) ||
        (group_name && tuneCount && groupCount == diskCount))
        return 0;

 error:
    virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                   _("Supported configurations are either single disk"
                     " with iotunes or all disks in same iotune group"));
    return -1;
}


static int
vzctDomainDefGetMetaAttr(const virDomainDef *def,
                         const char *uri,
                         const char *xpath,
                         const char *meta,
                         char **attr)
{
    g_autofree char *defmeta = NULL;

    *attr = NULL;

    if (!meta) {
        if (virXMLExtractNamespaceXML(def->metadata, uri,
                                      &defmeta) < 0)
            return -1;

        meta = defmeta;
    }

    if (meta) {
        g_autoptr(xmlDoc) xml = NULL;
        g_autoptr(xmlXPathContext) ctxt = NULL;

        if (!(xml = virXMLParseStringCtxt(meta, _("(meta)"), &ctxt)))
            return -1;

        *attr = virXPathString(xpath, ctxt);
        if (*attr && **attr == '\0')
            VIR_FREE(*attr);
    }

    return 0;
}


static int
vzctDomainDefValidateCallback(const virDomainDef *def,
                              void *opaque G_GNUC_UNUSED,
                              void *parseOpaque G_GNUC_UNUSED)
{
    size_t i;
    bool hasCdrom = false;
    bool hasVenet0 = false;
    virDomainNumatuneMemMode mode;
    g_autofree char *osid = NULL;
    g_autofree char *ostemplate = NULL;

    if (def->cputune.quota > 0 && !def->cputune.period) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("If 'quota' is positive, 'period' must be "
                         "set as well"));
        return -1;
    }
    if (def->cputune.global_period || def->cputune.global_quota ||
        def->cputune.emulator_period || def->cputune.emulator_quota ||
        def->cputune.iothread_period || def->cputune.iothread_quota) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only shares, quota, and period are supported "
                         " in the cputune section"));
        return -1;
    }
    if (def->onReboot != VIR_DOMAIN_LIFECYCLE_ACTION_RESTART) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only restart on reboot is supported"));
        return -1;
    }
    if (def->onPoweroff != VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only destroy on poweroff is supported"));
        return -1;
    }
    if (def->onCrash != VIR_DOMAIN_LIFECYCLE_ACTION_DESTROY) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only destroy on crash is supported"));
        return -1;
    }
    if (virDomainDefHasVcpusOffline(def)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Offline vcpus are not supported"));
        return -1;
    }
    if (def->placement_mode != VIR_DOMAIN_CPU_PLACEMENT_MODE_STATIC) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only static cpu placement mode is supported"));
        return -1;
    }
    if (virDomainNumatuneGetMode(def->numa, -1, &mode) == 0 &&
        mode != VIR_DOMAIN_NUMATUNE_MEM_STRICT) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only strict numatune mode is supported"));
        return -1;
    }
    if (virDomainNumatuneHasPlacementAuto(def->numa)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Auto numatune placement mode is not supported"));
        return -1;
    }
    if (def->mem.cur_balloon != virDomainDefGetMemoryTotal(def)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Memory balloning is not supported"));
        return -1;
    }
    if (STRNEQ("/sbin/init", def->os.init)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("init must be /sbin/init"));
        return -1;
    }

    if (vzctDomainDefValidateGraphics(def) < 0)
        return -1;

    for (i = 0; i < def->ndisks; i++) {
        const virDomainDiskDef *disk = def->disks[i];

        if (disk->device == VIR_DOMAIN_DISK_DEVICE_CDROM) {
            if (hasCdrom) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("Only one CD-ROM is supported."));
                return -1;
            }
            hasCdrom = true;
        }
    }

    for (i = 0; i < def->nnets; i++) {
        if (def->nets[i]->type == VIR_DOMAIN_NET_TYPE_VENET0) {
            if (hasVenet0) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("A domain can only have one venet0."));
                return -1;
            }
            hasVenet0 = true;
        }
    }


    if (vzctDomainDefValidateIoTune(def) < 0)
        return -1;

    if (vzctDomainDefGetMetaAttr(def,
                                 VZCT_VHS_URI, VZCT_VHS_XPATH_OSTEMPLATE,
                                 NULL, &ostemplate) < 0)
        return -1;

    if (vzctDomainDefGetMetaAttr(def,
                                 VZCT_LIBOSINFO_URI, VZCT_LIBOSINFO_XPATH_OSID,
                                 NULL, &osid) < 0)
        return -1;

    if (ostemplate && osid) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("cannot set both VHS ostemplate and libosinfo OS ID at the same time"));
        return -1;
    }

    return 0;
}


static int
vzctDomainDiskParse(xmlXPathContextPtr ctxt,
                    virDomainDiskDef *disk)
{
    vzctDomainDiskPrivate *diskPrivate =
        (vzctDomainDiskPrivate *)disk->privateData;
    g_autofree char *uuid = NULL;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
        return 0;

    if (!(uuid = virXPathString("string(./uuid[1])", ctxt)) ||
        virStrcpyStatic(diskPrivate->uuid, uuid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("cannot parse uuid"));
        return -1;
    }

    return 0;
}


static int
vzctDomainDiskFormat(virDomainDiskDef *disk,
                     virBuffer *buf)
{
    vzctDomainDiskPrivate *diskPrivate =
        (vzctDomainDiskPrivate *)disk->privateData;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
        return 0;

    virBufferAsprintf(buf, "<uuid>%s</uuid>\n", diskPrivate->uuid);
    return 0;
}

virDomainDefParserConfig vzctDriverDomainDefParserConfig = {
    .devicesPostParseCallback = vzctDomainDeviceDefPostParse,
    .deviceValidateCallback = vzctDeviceDefValidateCallback,
    .domainValidateCallback = vzctDomainDefValidateCallback,
    .features = VIR_DOMAIN_DEF_FEATURE_USER_ALIAS,
};

virDomainXMLPrivateDataCallbacks vzctDriverPrivateDataCallbacks = {
    .alloc = vzctDomainObjPrivateAlloc,
    .free = vzctDomainObjPrivateFree,
    .diskNew = vzctDomainDiskPrivateNew,
    .diskParse = vzctDomainDiskParse,
    .diskFormat = vzctDomainDiskFormat,
};


static virCaps *
vzctDriverCapsInit(void)
{
    virCaps *caps;
    virCapsGuest *guest;

    if ((caps = virCapabilitiesNew(virArchFromHost(),
                                   false, false)) == NULL)
        goto error;

    if (!(caps->host.numa = virCapabilitiesHostNUMANewHost()))
        goto error;

    if (virGetHostUUID(caps->host.host_uuid)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       "%s", _("cannot get the host uuid"));
        goto error;
    }

    if ((guest = virCapabilitiesAddGuest(caps,
                                         VIR_DOMAIN_OSTYPE_EXE,
                                         caps->host.arch,
                                         NULL, NULL, 0, NULL)) == NULL)
        goto error;

    if (virCapabilitiesAddGuestDomain(guest,
                                      VIR_DOMAIN_VIRT_VZCT,
                                      NULL, NULL, 0, NULL) == NULL)
        goto error;

    return caps;

 error:
    virObjectUnref(caps);
    return NULL;
}


static void
vzctDriverJobHandleEvent(vzctDriver *driver,
                         int event,
                         const char *ctid)
{
    virDomainObj *dom;
    virObjectEvent *domEvent;
    bool job = false;
    bool filtersLocked = false;
    int rc;

    if (event == VZCT_DOMAIN_EVENT_DEFINED) {
        dom = vzctAddDomain(driver, ctid, NULL);
        if (dom) {
            if (virDomainObjIsActive(dom)) {
                domEvent = virDomainEventLifecycleNewFromObj(dom,
                                             VIR_DOMAIN_EVENT_STARTED,
                                             VIR_DOMAIN_EVENT_STARTED_BOOTED);
                virObjectEventStateQueue(driver->eventsManager, domEvent);
            }
            domEvent = virDomainEventLifecycleNewFromObj(dom,
                                             VIR_DOMAIN_EVENT_DEFINED,
                                             VIR_DOMAIN_EVENT_DEFINED_ADDED);
            virObjectEventStateQueue(driver->eventsManager, domEvent);
        }
        goto cleanup;
    }

    if (!(dom = virDomainObjListFindByExtraId(driver->domains, ctid))) {
        /* Not an error - container can be deleted at the moment
         * this event is arrived */
        VIR_DEBUG("Can not find container by ctid: %s", ctid);
        return;
    }

    if (vzctDomainBeginJobInternal(dom, VZCT_DOMAIN_JOB_MODIFY, -1) < 0)
        goto cleanup;
    job = true;

    if (event == VZCT_DOMAIN_EVENT_UNDEFINED) {
        virDomainObjListRemove(driver->domains, dom);
        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_UNDEFINED,
                                         VIR_DOMAIN_EVENT_UNDEFINED_REMOVED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        goto cleanup;
    }

    rc = vzctCheckSync(driver, dom);
    if (rc < 0 || (rc == 0 && event == VZCT_DOMAIN_EVENT_UPDATED))
        goto cleanup;

    switch (event) {
    case VZCT_DOMAIN_EVENT_STARTED:
        if (virDomainObjIsActive(dom))
            break;

        if (vzctDomainStart(driver, dom, NULL) < 0)
            goto error;

        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_STARTED,
                                         VIR_DOMAIN_EVENT_STARTED_BOOTED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        break;

    case VZCT_DOMAIN_EVENT_STOPPED:
        if (!virDomainObjIsActive(dom))
            break;

        vzctDomainStop(driver, dom);
        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_STOPPED,
                                         VIR_DOMAIN_EVENT_STOPPED_DESTROYED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        break;

    case VZCT_DOMAIN_EVENT_UPDATED:
        if (vzctUpdateDomain(driver, dom) < 0)
            goto error;
        domEvent = virDomainEventLifecycleNewFromObj(dom,
                                         VIR_DOMAIN_EVENT_DEFINED,
                                         VIR_DOMAIN_EVENT_DEFINED_UPDATED);
        virObjectEventStateQueue(driver->eventsManager, domEvent);
        break;

    case VZCT_DOMAIN_EVENT_CONFIGURED:
        filtersLocked = true;
        virNWFilterReadLockFilterUpdates();
        if (virDomainConfVMNWFilterInstantiate(dom, false) < 0)
            goto error;
    }

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    if (filtersLocked)
        virNWFilterUnlockFilterUpdates();
    virDomainObjEndAPI(&dom);
    return;

 error:
    VIR_WARN("Can not sync state for domain ctid=%s", ctid);
    virDomainObjSetState(dom, VIR_DOMAIN_NOSTATE, -1);
    dom->def->id = -1;
    domEvent = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_DEFINED,
                                     VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, domEvent);
    goto cleanup;
}


static void
vzctDomainCleanupTimeout(int timer, void *opaque)
{
    vzctDriverJob *job;

    job = g_new0(vzctDriverJob, 1);
    job->type = VZCT_DRIVER_JOB_CLEANUP;
    job->dom = opaque;
    if (virThreadPoolSendJob(vzct_driver->workers, 0, job) < 0) {
        VIR_WARN("cannot run cleanup job");
        vzctDriverJobFree(job);
    }
    virEventUpdateTimeout(timer, -1);
}


static void
vzctDomainCleanup(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    size_t i;

    VIR_DEBUG("Cleanup domain name=%s", dom->def->name);

    for (i = 0; i < priv->exec_nprocs; i++) {
        vzctDomainExecProcess *exec_proc = priv->exec_procs[i];
        int status;
        int rc;

        if (!exec_proc->orphan)
            continue;

        if (!exec_proc->terminated) {
            vzctl2_env_exec_terminate(exec_proc->handle);
            exec_proc->terminated = true;
        }

        VIR_DEBUG("Reaping process pid=%d", exec_proc->handle->pid);
        rc = vzctl2_env_waitpid(exec_proc->handle, 1, &status);
        if (rc == 0)
            continue;

        if (rc < 0)
            VIR_WARN("Fail reaping process pid=%d, err: %s",
                     exec_proc->handle->pid, NULLSTR(vzctl2_get_last_error()));

        /* pid is reset in handle after reaping is finished */
        VIR_DEBUG("Process reaped.");
        vzctl2_release_exec_handle(exec_proc->handle);
        VIR_DELETE_ELEMENT_INPLACE(priv->exec_procs, i, priv->exec_nprocs);
    }

    for (i = 0; i < priv->exec_nprocs; i++) {
        if (priv->exec_procs[i]->orphan) {
            VIR_DEBUG("Cleanup is not finished. Retry after timeout.");
            virEventUpdateTimeout(priv->cleanupTimer, 30 * 1000);
            break;
        }
    }
}


static void
vzctDriverJobHandleConnectionClosed(vzctDriver *driver,
                                    int conn_id)
{
    virDomainObj **doms;
    size_t ndoms;
    size_t i;
    size_t j;

    /*
     * Like for qemu driver collect domains instead of using iterator.
     * Because if we stuck in iterator for some domain then read lock
     * for the list is hold and define/undefine API for example for
     * another domain will stuck which is not good.
     */
    virDomainObjListCollect(driver->domains, NULL, &doms, &ndoms, NULL, 0);
    VIR_DEBUG("run connection closed hanlder");

    for (i = 0; i < ndoms; i++) {
        virDomainObj *dom = doms[i];
        vzctDomainObjPrivate *priv = dom->privateData;

        virObjectLock(dom);

        if (priv->job == VZCT_DOMAIN_JOB_MIGRATION &&
            priv->mig_conn_id == conn_id)
            vzctDomainDropJob(dom);

        for (j = 0; j < priv->exec_nprocs; j++) {
            vzctDomainExecProcess *exec_proc = priv->exec_procs[j];

            if (exec_proc->conn_id == conn_id)
                exec_proc->orphan = true;
        }

        vzctDomainCleanup(dom);
        virObjectUnlock(dom);
    }

    virObjectListFreeCount(doms, ndoms);
}


static void
vzctDriverJobHandler(void *data, void *opaque)
{
    vzctDriverJob *job = data;
    vzctDriver *driver = opaque;

    switch (job->type) {
    case VZCT_DRIVER_JOB_EVENT:
        vzctDriverJobHandleEvent(driver, job->eventType, job->ctid);
        break;
    case VZCT_DRIVER_JOB_CONNECTION_CLOSED:
        vzctDriverJobHandleConnectionClosed(driver, job->conn_id);
        break;
    case VZCT_DRIVER_JOB_CLEANUP:
        virObjectLock(job->dom);
        vzctDomainCleanup(job->dom);
        virObjectUnlock(job->dom);
        break;
    }

    vzctDriverJobFree(job);
}


static int
vzctDomainLoadSnapshotMetadata(vzctDriver *driver,
                               virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    g_autofree char *dirpath = NULL;
    struct dirent *entry;
    DIR *dir = NULL;

    dirpath = g_strdup_printf("%s/"VZCT_SNAPHOTS_DIR, priv->bundlePath);

    if (virDirOpenIfExists(&dir, dirpath) <= 0)
        goto cleanup;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    while (virDirRead(dir, &entry, NULL) > 0) {
        unsigned int flags = VIR_DOMAIN_SNAPSHOT_PARSE_REDEFINE |
                             VIR_DOMAIN_SNAPSHOT_PARSE_INTERNAL;
        g_autofree char *path = NULL;
        g_autofree char *xml = NULL;
        virDomainSnapshotDef *def;
        virDomainMomentObj *snap;
        char uuid[VIR_UUID_STRING_BUFLEN + 2];

        if (virStrcpy(uuid, entry->d_name, sizeof(uuid)) < 0) {
            VIR_WARN("unexpected snapshot file name=%s", entry->d_name);
            continue;
        }

        path = g_strdup_printf("%s/%s", dirpath, entry->d_name);
        if (!virDomainSnapshotFindByName(priv->snapshots, uuid)) {
            if (unlink(path) < 0)
                VIR_WARN("Failed to unlink path=%s, errno=%d", path, errno);
            continue;
        }

        if (!(xml = virFileRead(path))) {
            VIR_WARN("Failed to read snapshot file=%s, errno=%d", path, errno);
            continue;
        }
        if (!(def = virDomainSnapshotDefParseString(xml,
                                                    driver->xmlopt,
                                                    NULL, NULL, flags)))
            continue;

        if (!(snap = virDomainSnapshotAssignDef(dom->snapshots, def)))
            virObjectUnref(def);
    }

 cleanup:
    virDomainSnapshotObjListRemoveAll(priv->snapshots);
    virDirClose(dir);
    return 0;
}


static int
vzctAutostartDomain(virDomainObj *dom,
                    void *opaque)
{
    vzctDriver *driver = opaque;

    virObjectLock(dom);

    if (!dom->autostart || virDomainObjIsActive(dom))
        goto cleanup;

    if (vzctDomainStartFull(driver, dom) < 0)
        goto cleanup;

 cleanup:
    virObjectUnlock(dom);
    return 0;
}


static int
vzctStateInitialize(bool privileged,
                    const char *root G_GNUC_UNUSED,
                    virStateInhibitCallback callback G_GNUC_UNUSED,
                    void *opaque G_GNUC_UNUSED)
{
    vzctDriver *driver = NULL;
    struct vzctl_ids *ids = NULL;
    bool autostart;
    int ndoms;
    size_t i;

    /* Check that the user is root, silently disable if not */
    if (!privileged) {
        VIR_DEBUG("Not running privileged, disabling driver");
        return 0;
    }

    driver = g_new0(vzctDriver, 1);
    if (virMutexInit(&driver->lock) < 0) {
        VIR_FREE(driver);
        return -1;
    }
    vzct_driver = driver;

    if (!(driver->xmlopt = virDomainXMLOptionNew(&vzctDriverDomainDefParserConfig,
                                                 &vzctDriverPrivateDataCallbacks,
                                                 NULL, NULL, NULL))) {
        goto error;
    }

    driver->config = g_new0(vzctDriverConfig, 1);
    driver->config->stateDir = g_strdup(LOCALSTATEDIR"/run/libvirt/vzct");

    if (g_mkdir_with_parents(driver->config->stateDir, 0777) < 0) {
        virReportSystemError(errno, _("Failed to create state dir %s"),
                             driver->config->stateDir);
        goto error;
    }

    if (!(driver->caps = vzctDriverCapsInit()))
        goto error;

    if (!(driver->vncPorts = virPortAllocatorRangeNew(_("vnc"), 5900, 65535)))
        goto error;

    if (!(driver->migrationPorts = virPortAllocatorRangeNew(_("vnc"), 49152, 49215)))
        goto error;

    if (!(driver->workers = virThreadPoolNewFull(0, 1, 0,
                                                 vzctDriverJobHandler,
                                                 "vzct-event",
                                                 vzct_driver)))
        goto error;

    if (!(driver->eventsManager = virObjectEventStateNew()))
        goto error;

    driver->migrating = virHashAtomicNew(g_free);

    if (vzctDriverConnectOpen(driver) < 0)
        goto error;

    VZCT_CHECK_PTR(error, ids, vzctl2_alloc_env_ids());

    if ((ndoms = vzctl2_get_env_ids_by_state(ids, ENV_STATUS_EXISTS)) < 0) {
        VZCT_ERROR_MESSAGE(vzctl2_get_env_ids_by_state);
        vzctl2_free_env_ids(ids);
        goto error;
    }

    for (i = 0; i < ndoms; i++) {
        virDomainObj *dom = vzctAddDomain(driver, ids->ids[i], NULL);
        virDomainObjEndAPI(&dom);
    }
    vzctl2_free_env_ids(ids);

    vzctCleanupDriverConfigs(driver);

    if (virDriverShouldAutostart(driver->config->stateDir, &autostart) < 0)
        goto error;

    if (autostart)
        virDomainObjListForEach(driver->domains, false,
                                vzctAutostartDomain, driver);

    return 0;

 error:
    vzctStateCleanup();
    return -1;
}


static int
vzctConnectSupportsFeature(virConnectPtr conn,
                           int feature)
{
    if (virConnectSupportsFeatureEnsureACL(conn) < 0)
        return -1;

    switch (feature) {
    case VIR_DRV_FEATURE_MIGRATION_PARAMS:
    case VIR_DRV_FEATURE_MIGRATION_P2P:
    case VIR_DRV_FEATURE_MIGRATE_CHANGE_PROTECTION:
    case VIR_DRV_FEATURE_TYPED_PARAM_STRING:
        return 1;
    default:
        return 0;
    }
}


static const char *
vzctConnectGetType(virConnectPtr conn)
{
    if (virConnectGetTypeEnsureACL(conn) < 0)
        return NULL;

    return "vzct";
}


static virDomainObj *
vzctDomObjFromDomain(virDomainPtr domain)
{
    vzctDriver *driver = domain->conn->privateData;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    virDomainObj *dom;

    if (!(dom = virDomainObjListFindByUUID(driver->domains, domain->uuid))) {
        virUUIDFormat(domain->uuid, uuidstr);
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("no domain with matching uuid '%s' (%s)"),
                       uuidstr, domain->name);
        return NULL;
    }

    return dom;
}


static int
vzctConnectListAllDomains(virConnectPtr conn,
                          virDomainPtr **doms,
                          unsigned int flags)
{
    vzctDriver *driver = conn->privateData;
    int ret = -1;

    virCheckFlags(VIR_CONNECT_LIST_DOMAINS_FILTERS_ALL, -1);

    if (virConnectListAllDomainsEnsureACL(conn) < 0)
        return -1;

    ret = virDomainObjListExport(driver->domains, conn, doms,
                                 virConnectListAllDomainsCheckACL, flags);

    return ret;
}


static int
vzctDomainGetState(virDomainPtr domain,
                   int *state,
                   int *reason,
                   unsigned int flags)
{
    virDomainObj *dom;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetStateEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    *state = virDomainObjGetState(dom, reason);
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static virDomainPtr
vzctDomainLookupByUUID(virConnectPtr conn,
                       const unsigned char *uuid)
{
    vzctDriver *driver = conn->privateData;
    virDomainObj *dom;
    virDomainPtr domain = NULL;

    if (!(dom = virDomainObjListFindByUUID(driver->domains, uuid))) {
        char uuidstr[VIR_UUID_STRING_BUFLEN];
        virUUIDFormat(uuid, uuidstr);
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("No domain with matching uuid '%s'"), uuidstr);
        goto cleanup;
    }

    if (virDomainLookupByUUIDEnsureACL(conn, dom->def) < 0)
        goto cleanup;

    domain = virGetDomain(conn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    virDomainObjEndAPI(&dom);
    return domain;
}


static virDomainPtr
vzctDomainLookupByName(virConnectPtr conn,
                       const char *name)
{
    vzctDriver *driver = conn->privateData;
    virDomainObj *dom;
    virDomainPtr domain = NULL;

    if (!(dom = virDomainObjListFindByName(driver->domains, name))) {
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("No domain with matching name '%s'"), name);
        goto cleanup;
    }

    if (virDomainLookupByNameEnsureACL(conn, dom->def) < 0)
        goto cleanup;

    domain = virGetDomain(conn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    virDomainObjEndAPI(&dom);
    return domain;
}


static char *
vzctDomainGetXMLDesc(virDomainPtr domain,
                     unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom;
    char *ret = NULL;
    bool job = false;

    /* Flags checked by virDomainDefFormat */

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetXMLDescEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    ret = virDomainDefFormat((flags & VIR_DOMAIN_XML_INACTIVE) &&
                             dom->newDef ? dom->newDef : dom->def,
                             driver->xmlopt,
                             virDomainDefFormatConvertXMLFlags(flags));

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC(vzctl_env_param, vzctl2_free_env_param);

static int
vzctlDomainCreate(virDomainObj *dom,
                  virDomainDef *def)
{
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    g_autoptr(vzctl_env_param) env_param = NULL;
    struct vzctl_env_create_param create_param = {
        /* FIXME */
        .config = (char*) "vswap.1024MB",
        .layout = VZCTL_LAYOUT_5,
        .no_root_disk = 1,
    };

    virUUIDFormat(def->uuid, uuidstr);
    create_param.uuid = uuidstr;
    SET_CTID(create_param.ctid, def->extraId);
    create_param.name = def->name;

    VZCT_CHECK_PTR_RET(-1, env_param, vzctl2_alloc_env_param());
    VZCT_CHECK_RET(-1, vzctl2_env_create(env_param, &create_param, 0));

    def->extraId = g_strdup(uuidstr);
    virDomainObjSetState(dom, VIR_DOMAIN_SHUTOFF, -1);

    return 0;
}


static char *
vzctDomainDiskDescription(const virDomainDiskDef *disk)
{
    g_autofree char *ret = NULL;

    switch (disk->src->type) {
    case VIR_STORAGE_TYPE_FILE:
        ret = g_strdup_printf("file='%s'", disk->src->path);
        break;

    case VIR_STORAGE_TYPE_VOLUME:
        ret = g_strdup_printf("pool='%s', volume='%s'",
                              disk->src->srcpool->pool,
                              disk->src->srcpool->volume);
        break;

    default:
        ret = g_strdup("type 'unsupported'");
        break;
    }

    return g_steal_pointer(&ret);
}


static void
vzctGenerateDiskUUID(const virDomainDiskDef *disk)
{
    unsigned char uuid[VIR_UUID_BUFLEN];
    vzctDomainDiskPrivate *diskPriv = (vzctDomainDiskPrivate *)disk->privateData;
    char *diskUUID = diskPriv->uuid;

    virUUIDGenerate(uuid);
    diskUUID[0] = '{';
    virUUIDFormat(uuid, diskUUID + 1);
    diskUUID[VIR_UUID_STRING_BUFLEN] = '}';
    diskUUID[VIR_UUID_STRING_BUFLEN + 1] = '\0';
}


static int
vzctConfigDiskAdd(vzctl_env_handle_ptr env,
                  const virDomainDiskDef *disk)
{
    /* just { 0 } is known bug for gcc less then 5.1 */
    struct vzctl_disk_param params = { .enabled = 0 };
    vzctDomainDiskPrivate *diskPrivate =
        (vzctDomainDiskPrivate *)disk->privateData;
    char rootpath[2] = "/";
    char emptypath[1] = "";

    VIR_DEBUG("Adding disk to config %s", vzctDomainDiskDescription(disk));

    params.enabled = 1;

    switch (disk->src->type) {
    case VIR_STORAGE_TYPE_FILE:
        params.path = disk->src->path;
        break;

    case VIR_STORAGE_TYPE_VOLUME:
        params.path = emptypath;
        break;

    default:
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("unsupported disk source type '%s'"),
                       virStorageTypeToString(disk->src->type));
        return -1;
    }

    if (disk->info.bootIndex == 1)
        params.mnt = rootpath;

    if (!diskPrivate->uuid[0])
        vzctGenerateDiskUUID(disk);
    memcpy(params.uuid, diskPrivate->uuid, sizeof(params.uuid));

    VZCT_CHECK_RET(-1, vzctl2_env_attach_disk(env, &params));
    return 0;
}


/*
 * Convert libosinfo short-id to libvzctl ostemplate name.
 * Like ubuntu14.04 to ubuntu-14.04. More precisely:
 *
 * - add dash before first digit
 */
static char *
vzctConvertToOstemplateName(const char *name)
{
    g_autoptr(GRegex) regex = NULL;
    g_autoptr(GError) err = NULL;
    g_autoptr(GMatchInfo) info = NULL;

    if (!(regex = g_regex_new("^(.*?)([0-9].*?)$", 0, 0, &err))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Failed to compile regex %s"), err->message);
        goto fail;
    }

    if (!g_regex_match(regex, name, 0, &info)) {
        VIR_WARN("cannot convert libosinfo: '%s'", name);
        goto fail;
    }

    return g_strdup_printf("%s-%s",
                           g_match_info_fetch(info, 1),
                           g_match_info_fetch(info, 2));

 fail:
    return g_strdup(name);
}


/*
 * Both @libosid and @ostemplate can't be not NULL. This is checked
 * outside scope of this function.
 */
static int
vzctConfigSetOsDistribution(vzctl_env_param_ptr param,
                            const char *libosid,
                            const char *ostemplate)
{
    g_autofree char *tmp = NULL;

    if (libosid) {
        g_autofree char *libosname  = NULL;

        if (vzctOsinfoConvertName(libosid, "id", "short-id", &libosname))
            return -1;

        if (!libosname) {
            virReportError(VIR_ERR_INVALID_ARG, _("unknown os id '%s'"), libosid);
            return -1;
        }

        ostemplate = tmp = vzctConvertToOstemplateName(libosname);
    }

    VZCT_CHECK_RET(-1, vzctl2_env_set_ostemplate(param, ostemplate ?: ""));

    return 0;
}


static int
vzctConfigDiskDelete(vzctl_env_handle_ptr env,
                     const virDomainDiskDef *disk)
{
    const vzctDomainDiskPrivate *diskPrivate =
        (const vzctDomainDiskPrivate *)disk->privateData;

    VIR_DEBUG("Deleting disk from config %s", vzctDomainDiskDescription(disk));

    VZCT_CHECK_RET(-1, vzctl2_env_detach_disk(env, diskPrivate->uuid));
    return 0;
}


static int
vzctConfigDiskUpdate(vzctl_env_handle_ptr env,
                     const virDomainDiskDef *olddisk,
                     virDomainDiskDef *disk)
{
    /* just { 0 } is known bug for gcc less then 5.1 */
    struct vzctl_disk_param param = { .enabled = 0 };
    const vzctDomainDiskPrivate *oldDiskPrivate =
        (const vzctDomainDiskPrivate *)olddisk->privateData;
    vzctDomainDiskPrivate *diskPrivate =
        (vzctDomainDiskPrivate *)disk->privateData;
    char rootpath[2] = "/";
    char emptypath[1] = "";

    VIR_DEBUG("Updating disk config '%s'", vzctDomainDiskDescription(disk));

    memcpy(diskPrivate->uuid, oldDiskPrivate->uuid, sizeof(diskPrivate->uuid));

    if (disk->info.bootIndex == olddisk->info.bootIndex)
        return 0;

    memcpy(param.uuid, oldDiskPrivate->uuid, sizeof(param.uuid));

    if (disk->info.bootIndex == 1)
        param.mnt = rootpath;
    else
        param.mnt = emptypath;

    VZCT_CHECK(error,
               vzctl2_env_set_disk(env, &param));

    return 0;

 error:
    return -1;
}


static int
vzctDomainDiskMatch(const virDomainDiskDef *a,
                    const virDomainDiskDef *b)
{
    if (a->src->type != b->src->type)
        return false;

    switch (a->src->type) {
    case VIR_STORAGE_TYPE_FILE:
        return STREQ_NULLABLE(a->src->path, b->src->path);

    case VIR_STORAGE_TYPE_VOLUME:
        return STREQ_NULLABLE(a->src->srcpool->pool,
                              b->src->srcpool->pool) &&
               STREQ_NULLABLE(a->src->srcpool->volume,
                              b->src->srcpool->volume);

    default:
         return false;
    }

    return false;
}


static int
vzctConfigSetDisks(vzctl_env_handle_ptr env,
                   const virDomainDef *olddef,
                   const virDomainDef *def)
{
    size_t i, j;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        const virDomainDiskDef *olddisk = NULL;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        for (j = 0; olddef && j < olddef->ndisks; j++) {
            const virDomainDiskDef *d = olddef->disks[j];

            if (vzctDomainDiskMatch(disk, d)) {
                olddisk = olddef->disks[j];
                break;
            }
        }

        if (olddisk) {
            if (vzctConfigDiskUpdate(env, olddisk, disk) < 0)
                return -1;
        } else {
            if (vzctConfigDiskAdd(env, disk) < 0)
                return -1;
        }
    }

    for (i = 0; olddef && i < olddef->ndisks; i++) {
        const virDomainDiskDef *olddisk = olddef->disks[i];
        const virDomainDiskDef *disk = NULL;

        if (olddisk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        for (j = 0; j < def->ndisks; j++) {
            if (vzctDomainDiskMatch(olddisk, def->disks[j])) {
                disk = def->disks[j];
                break;
            }
        }

        if (!disk && vzctConfigDiskDelete(env, olddisk) < 0)
            return -1;
    }

    return 0;
}


static char *
virNetDevIpAddrFormat(const virNetDevIPAddr *ip)
{
    g_autofree char *iponly = NULL;
    g_autofree char *ipmask = NULL;

    if (!(iponly = virSocketAddrFormat(&ip->address)))
        return NULL;

    if (ip->prefix) {
        if (VIR_SOCKET_ADDR_IS_FAMILY(&ip->address, AF_INET6)) {
            ipmask = g_strdup_printf("%s/%u", iponly, ip->prefix);
        } else if (VIR_SOCKET_ADDR_IS_FAMILY(&ip->address, AF_INET)) {
            g_autofree char *maskstr = NULL;
            virSocketAddr mask;

            if (virSocketAddrPrefixToNetmask(ip->prefix, &mask, AF_INET) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR,
                               _("invalid prefix %u"), ip->prefix);
                return NULL;
            }

            if (!(maskstr = virSocketAddrFormat(&mask)))
                return NULL;

            ipmask = g_strdup_printf("%s/%s", iponly, maskstr);
        }
    }

    if (ipmask)
        return g_steal_pointer(&ipmask);
    else
        return g_steal_pointer(&iponly);
}


static int
vzctConfigNetAdd(vzctl_env_param_ptr param,
                 const virDomainNetDef *net,
                 struct vzctl_veth_dev_param *oldnet_param)
{
    struct vzctl_veth_dev_param net_param;
    vzctl_veth_dev_iterator it = NULL;
    char mac[VIR_MAC_STRING_BUFLEN];
    g_autofree char *gwstr = NULL;
    g_autofree char *gw6str = NULL;
    char interfaceid[VIR_UUID_STRING_BUFLEN];
    size_t i;

    VIR_DEBUG("Updating/adding network interface: %s", net->ifname);

    if (oldnet_param)
        memcpy(&net_param, oldnet_param, sizeof(net_param));
    else
        memset(&net_param, 0, sizeof(net_param));

    virMacAddrFormat(&net->mac, mac);
    net_param.mac_ve = mac;

    net_param.dev_name = net->ifname;
    net_param.dev_name_ve = net->ifname_guest;
    net_param.dhcp = net->dhcp;
    net_param.dhcp6 = net->dhcp6;
    switch ((int)net->type) {
    case VIR_DOMAIN_NET_TYPE_BRIDGE:
        net_param.nettype = VZCTL_NETTYPE_BRIDGE;
        net_param.network = net->data.bridge.brname;
        break;

    case VIR_DOMAIN_NET_TYPE_NETWORK:
        net_param.nettype = VZCTL_NETTYPE_VNET;
        net_param.network = net->data.network.name;
        break;

    default:
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unexpected net interface type: %s"),
                       virDomainNetTypeToString(net->type));
        return -1;
    }
    if (net->virtPortProfile &&
        net->virtPortProfile->virtPortType == VIR_NETDEV_VPORT_PROFILE_OPENVSWITCH) {
        net_param.vporttype = VZCTL_VPORTTYPE_OVS;

        if (net->virtPortProfile->interfaceID_specified) {
            virUUIDFormat(net->virtPortProfile->interfaceID, interfaceid);
            net_param.ifaceid = interfaceid;
        } else {
            net_param.ifaceid = NULL;
        }
    } else {
        net_param.vporttype = VZCTl_VPORTTYPE_NONE;
    }

    for (i = 0; i < net->guestIP.nroutes; i++) {
        virNetDevIPRoute *route = net->guestIP.routes[i];

        if (VIR_SOCKET_ADDR_IS_FAMILY(&route->gateway, AF_INET)) {
            if (!(gwstr = virSocketAddrFormat(&route->gateway)))
                return -1;

            net_param.gw = gwstr;
        } else if (VIR_SOCKET_ADDR_IS_FAMILY(&route->gateway, AF_INET6)) {
            if (!(gw6str = virSocketAddrFormat(&route->gateway)))
                return -1;

            net_param.gw6 = gw6str;
        }
    }

    VZCT_CHECK_PTR_RET(-1, it, vzctl2_create_veth_dev(&net_param, sizeof(net_param)));

    VZCT_CHECK_RET(-1, vzctl2_env_del_veth_ipaddress(it, "all"));
    for (i = 0; i < net->guestIP.nips; i++) {
        g_autofree char *ipstr = NULL;

        if (!(ipstr = virNetDevIpAddrFormat(net->guestIP.ips[i])))
            return -1;

        VZCT_CHECK_RET(-1, vzctl2_env_add_veth_ipaddress(it, ipstr));
    }

    VZCT_CHECK_RET(-1, vzctl2_env_add_veth(param, it));
    return 0;
}


static int
vzctConfigNetDelete(vzctl_env_param_ptr param,
                    const char *ifname)
{
    VIR_DEBUG("Deleting network interface: %s", ifname);

    VZCT_CHECK_RET(-1, vzctl2_env_del_veth(param, ifname));
    return 0;
}


static int
vzctConfigUpdateVenet0(vzctl_env_param_ptr param,
                       const virDomainNetDef *net)
{
    size_t i;

    VZCT_CHECK_RET(-1, vzctl2_env_del_ipaddress(param, "all"));

    for (i = 0; i < net->guestIP.nips; i++) {
        g_autofree char *ipstr = NULL;

        if (!(ipstr = virNetDevIpAddrFormat(net->guestIP.ips[i])))
            return -1;

        VZCT_CHECK_RET(-1, vzctl2_env_add_ipaddress(param, ipstr));
    }

    return 0;
}


static int
vzctConfigSetNets(vzctl_env_param_ptr param,
                  vzctl_env_param_ptr oldparam,
                  const virDomainDef *def)
{
    size_t i;
    vzctl_veth_dev_iterator it;
    bool setrate = false;

    for (i = 0; i < VIR_DOMAIN_NETWORK_CLASS_MAX; i++) {
        if (def->net_classes.rate[i]) {
            struct vzctl_rate_param rate_param;
            vzctl_rate_iterator irs;

            rate_param.dev = "*";
            rate_param.net_class = i;
            rate_param.rate = def->net_classes.rate[i];

            VZCT_CHECK_PTR_RET(-1, irs, vzctl2_create_rate(&rate_param));
            VZCT_CHECK_RET(-1, vzctl2_env_add_rate(param, irs));

            setrate = true;
        }
    }

    if (!setrate)
        VZCT_CHECK_RET(-1, vzctl2_env_add_rate(param, NULL));

    VZCT_CHECK_RET(-1, vzctl2_env_set_ratebound(param, def->net_classes.bound ? 1 : 0));

    if (def->dns.hostname)
        vzctl2_env_set_hostname(param, def->dns.hostname);

    if (def->dns.servers) {
        for (i = 0; i < def->dns.nservers; i++) {
            g_autofree char *server = virSocketAddrFormat(&def->dns.servers[i]);

            VZCT_CHECK_RET(-1, vzctl2_env_add_nameserver(param, server));
        }
    } else {
        VZCT_CHECK_RET(-1, vzctl2_env_add_nameserver(param, ""));
    }

    if (def->dns.searchs) {
        for (i = 0; def->dns.searchs[i]; i++)
            VZCT_CHECK_RET(-1, vzctl2_env_add_searchdomain(param,
                                                           def->dns.searchs[i]));
    } else {
        VZCT_CHECK_RET(-1, vzctl2_env_add_searchdomain(param, ""));
    }

    for (i = 0; i < def->nnets; i++) {
        const virDomainNetDef *net = def->nets[i];
        struct vzctl_veth_dev_param oldnet_param;

        if (net->type == VIR_DOMAIN_NET_TYPE_VENET0) {
            if (vzctConfigUpdateVenet0(param, net) < 0)
                return -1;

            continue;
        }

        it = NULL;
        while ((it = vzctl2_env_get_veth(oldparam, it))) {
            vzctl2_env_get_veth_param(it, &oldnet_param, sizeof(oldnet_param));
            if (STREQ_NULLABLE(net->ifname, oldnet_param.dev_name))
                break;
        }

        if (vzctConfigNetAdd(param, net, it ? &oldnet_param : NULL) < 0)
            return -1;
    }

    it = NULL;
    while ((it = vzctl2_env_get_veth(oldparam, it))) {
        struct vzctl_veth_dev_param oldnet_param;
        const virDomainNetDef *net = NULL;

        vzctl2_env_get_veth_param(it, &oldnet_param, sizeof(oldnet_param));
        for (i = 0; i < def->nnets; i++) {
            if (STREQ_NULLABLE(oldnet_param.dev_name, def->nets[i]->ifname)) {
                net = def->nets[i];
                break;
            }
        }
        if (!net && vzctConfigNetDelete(param, oldnet_param.dev_name_ve) < 0)
            return -1;
    }

    return 0;
}


static void
vzctConfigSetIotunes(vzctl_env_param_ptr param,
                     const virDomainDef *def)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        virDomainBlockIoTuneInfo *iotune = &disk->blkdeviotune;
        unsigned int io = iotune->total_bytes_sec;
        unsigned int iops = iotune->total_iops_sec;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        /*
         * Quite strange interface. We set 0 here for unlimited but
         * get back UINT_MAX on read.
         */
        vzctl2_env_set_iolimit(param, io);
        vzctl2_env_set_iopslimit(param, iops);
        break;
    }
}


static int
vzctConfigSet(const virDomainDef *olddef,
              const virDomainDef *def)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    struct vzctl_mem_guarantee mem_guarantee;
    struct vzctl_cpulimit_param cpu_limit;
    vzctl_env_param_ptr oldparam;
    const char *ctid;
    g_autofree char *mask = NULL;
    g_autofree char *osid = NULL;
    g_autofree char *ostemplate = NULL;
    int ioprio;
    int err;

    ctid = olddef ? olddef->extraId : def->extraId;
    VZCT_CHECK_PTR_RET(-1, param, vzctl2_alloc_env_param());
    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(ctid, 0, &err));
    oldparam = vzctl2_get_env_param(env);

    /* Ceiling memory size from Kb to Mb */
    VZCT_CHECK_RET(-1, vzctl2_env_set_ramsize(param,
                                              (def->mem.cur_balloon + 1023)
                                              / 1024));
    memset(&mem_guarantee, 0, sizeof(mem_guarantee));
    if (def->mem.min_guarantee_auto) {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_AUTO;
    } else {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_BYTES;
        /* KiB to bytes */
        mem_guarantee.value = def->mem.min_guarantee << 10;
    }
    VZCT_CHECK_RET(-1, vzctl2_env_set_memguarantee(param, &mem_guarantee));

    cpu_limit.type = VZCTL_CPULIMIT_PCT;
    cpu_limit.limit = def->cputune.quota > 0 ?
                      ((def->cputune.quota * 100) /
                       def->cputune.period) : 0;
    VZCT_CHECK_RET(-1, vzctl2_env_set_cpulimit(param, &cpu_limit));

    VZCT_CHECK_RET(-1, vzctl2_env_set_cpuunits(param, def->cputune.shares ?
                                                      def->cputune.shares :
                                                      VZCTL_CPUUNITS_DEFAULT));

    VZCT_CHECK_RET(-1, vzctl2_env_set_cpu_count(param, def->maxvcpus));
    if (!(mask = virBitmapFormat(def->cpumask)))
        return -1;
    VZCT_CHECK_RET(-1, vzctl2_env_set_cpumask(param, mask));

    VIR_FREE(mask);
    if (!(mask = virDomainNumatuneFormatNodeset(def->numa, NULL, -1)))
        return -1;
    VZCT_CHECK_RET(-1, vzctl2_env_set_nodemask(param, mask));

    if (vzctConfigSetDisks(env, olddef, def) < 0)
        return -1;

    if (vzctDomainDefGetMetaAttr(def,
                                 VZCT_VHS_URI, VZCT_VHS_XPATH_OSTEMPLATE,
                                 NULL, &ostemplate) < 0)
        return -1;

    if (vzctDomainDefGetMetaAttr(def,
                                 VZCT_LIBOSINFO_URI, VZCT_LIBOSINFO_XPATH_OSID,
                                 NULL, &osid) < 0)
        return -1;

    if (vzctConfigSetOsDistribution(param, osid, ostemplate) < 0)
        return -1;

    if (vzctConfigSetNets(param, oldparam, def) < 0)
        return -1;

    vzctConfigSetIotunes(param, def);

    ioprio = vzctConvertWeightToIOPrio(def->blkio.weight);
    VZCT_CHECK_RET(-1, vzctl2_env_set_ioprio(param, ioprio));

    VZCT_CHECK_RET(-1, vzctl2_apply_param(env, param, VZCTL_SAVE));
    return 0;
}


static virDomainPtr
vzctDomainDefineXMLFlags(virConnectPtr conn,
                         const char *xml,
                         unsigned int flags)
{
    vzctDriver *driver = conn->privateData;
    vzctDomainObjPrivate *priv;
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(virDomainDef) newdef = NULL;
    virDomainObj *dom = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE;
    virDomainPtr domain = NULL;
    g_autofree char *bundlePath = NULL;
    virObjectEvent *event;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    const char *ctid;
    bool job = false;
    int detail;

    if (flags & VIR_DOMAIN_DEFINE_VALIDATE)
        parse_flags |= VIR_DOMAIN_DEF_PARSE_VALIDATE_SCHEMA;

    if ((def = virDomainDefParseString(xml, driver->xmlopt,
                                       NULL, parse_flags)) == NULL)
        return NULL;

    if (virDomainDefineXMLFlagsEnsureACL(conn, def) < 0)
        return NULL;

    virUUIDFormat(def->uuid, uuidstr);
    def->extraId = g_strdup(uuidstr);
    if (!(dom = virDomainObjListAdd(driver->domains, def, driver->xmlopt, 0)))
        return NULL;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (dom->def && virDomainObjGetState(dom, NULL) != VIR_DOMAIN_SHUTOFF) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("define is supportend only for shutoff domains"));
        goto cleanup;
    }

    if (!dom->def && vzctlDomainCreate(dom, def) < 0)
        goto cleanup;

    if (vzctConfigSet(dom->def, def) < 0)
        goto cleanup;

    if (dom->def)
        detail = VIR_DOMAIN_EVENT_DEFINED_UPDATED;
    else
        detail = VIR_DOMAIN_EVENT_DEFINED_ADDED;

    priv = dom->privateData;
    ctid = dom->def ? dom->def->extraId : def->extraId;
    if (!(newdef = vzctLoadDomainConfig(driver, ctid, &bundlePath)))
        goto cleanup;
    virDomainDefFree(dom->def);
    if (vzctMergeConfigs(newdef, def) < 0)
        goto cleanup;
    dom->def = g_steal_pointer(&newdef);
    if (!priv->bundlePath)
        priv->bundlePath = g_steal_pointer(&bundlePath);
    dom->persistent = true;

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_DEFINED,
                                     detail);
    virObjectEventStateQueue(driver->eventsManager, event);
    domain = virGetDomain(conn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    if (!domain && !dom->def) {
        /* Removing needs def for name and uuid */
        dom->def = g_steal_pointer(&def);
        virDomainObjListRemove(driver->domains, dom);
    }

    virDomainObjEndAPI(&dom);
    return domain;
}


static virDomainPtr
vzctDomainDefineXML(virConnectPtr conn, const char *xml)
{
    return vzctDomainDefineXMLFlags(conn, xml, 0);
}


static char *
vzctConnectGetCapabilities(virConnectPtr conn)
{
    vzctDriver *driver = conn->privateData;
    char *xml;

    if (virConnectGetCapabilitiesEnsureACL(conn) < 0)
        return NULL;

    xml = virCapabilitiesFormatXML(driver->caps);
    return xml;
}


static int
vzctDomainUndefineFlags(virDomainPtr domain,
                        unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom;
    g_autoptr(vzctl_env_handle) env = NULL;
    virObjectEvent *event;
    vzctl_env_status_t status;
    size_t i;
    int ret = -1;
    int err;

    virCheckFlags(VIR_DOMAIN_UNDEFINE_MANAGED_SAVE |
                  VIR_DOMAIN_UNDEFINE_SNAPSHOTS_METADATA |
                  VIR_DOMAIN_UNDEFINE_NVRAM |
                  VIR_DOMAIN_UNDEFINE_KEEP_NVRAM |
                  VIR_DOMAIN_UNDEFINE_CHECKPOINTS_METADATA, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainUndefineFlagsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("cannot undefine running domain"));
        goto cleanup;
    }

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_SUSPENDED));

    if ((status.mask & ENV_STATUS_SUSPENDED) &&
        !(flags & VIR_DOMAIN_UNDEFINE_MANAGED_SAVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("cannot undefine while domain managed "
                         "save image exists"));
        goto cleanup;
    }

    if (virDomainSnapshotObjListNum(dom->snapshots, NULL, 0) &&
        !(flags & VIR_DOMAIN_UNDEFINE_SNAPSHOTS_METADATA)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("cannot undefine inactive domain with "
                         "snapshots metadata"));
        goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env,
                   vzctl2_env_open(dom->def->extraId, 0, &err));

    for (i = 0; i < dom->def->ndisks; i++) {
        virDomainDiskDef *disk = dom->def->disks[i];
        vzctDomainDiskPrivate *diskPrivate =
            (vzctDomainDiskPrivate *)disk->privateData;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        /*
         * Disk are detached because vzctl2_env_destroy unlinks disks
         * as well. But in case of volumes we don't store path to disks
         * in offline configuration. And currently undefine works only
         * for offline containers.
         */
        if (disk->src->type == VIR_STORAGE_TYPE_VOLUME)
            continue;

        VZCT_CHECK(cleanup, vzctl2_env_detach_disk(env, diskPrivate->uuid));
    }

    VZCT_CHECK(cleanup, vzctl2_env_destroy(env, 0));
    virDomainObjListRemove(driver->domains, dom);
    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_UNDEFINED,
                                     VIR_DOMAIN_EVENT_UNDEFINED_REMOVED);
    virObjectEventStateQueue(driver->eventsManager, event);
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainUndefine(virDomainPtr domain)
{
    return vzctDomainUndefineFlags(domain, 0);
}


static int
vzctDomainIsActive(virDomainPtr domain)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom;
    int ret = -1;
    bool job = false;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainIsActiveEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    ret = virDomainObjIsActive(dom);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainStopVzctl(virDomainObj *dom,
                    bool gracefull)
{
    int mode = gracefull ? M_HALT : M_KILL;
    g_autoptr(vzctl_env_handle) env = NULL;
    int ret = -1;
    int err;

    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_RET(-1, vzctl2_env_stop(env, mode, 0));
    ret = 0;
    return ret;
}


static int
vzctDomainAddCdrom(vzctl_env_handle_ptr env,
                   const virDomainDef *def)
{
    vzctl_env_param_ptr param = vzctl2_get_env_param(env);
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        const virDomainDiskDef *disk = def->disks[i];

        if (disk->device == VIR_DOMAIN_DISK_DEVICE_CDROM) {
            VZCT_CHECK_RET(-1, vzctl2_env_set_cidata_fname(param, disk->src->path));
            break;
        }
    }

    return 0;
}


static int
vzctDomainCleanupGarbage(virDomainObj *dom)
{
    g_autofree char *migfile = vzctMigFile(dom);

    if (unlink(migfile) < 0 && errno != ENOENT) {
        virReportSystemError(errno,
                             _("cannot remove migrating status file path=%s"),
                             migfile);
        return -1;
    }

    return 0;
}


static int
vzctDomainResolveVolumes(virDomainDef *def,
                         vzctl_env_handle *env)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        struct vzctl_disk_param param = { .enabled = 0 };
        const vzctDomainDiskPrivate *diskPrivate =
            (const vzctDomainDiskPrivate *)disk->privateData;

        if (virDomainDiskTranslateSourcePool(disk) < 0)
            return -1;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        if (disk->src->srcpool->actualtype != VIR_STORAGE_TYPE_FILE) {
            virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                           _("only ploop volumes are supported"));
            return -1;
        }

        memcpy(param.uuid, diskPrivate->uuid, sizeof(param.uuid));
        param.path = disk->src->path;
        VZCT_CHECK_RET(-1, vzctl2_env_set_disk(env, &param));
    }

    return 0;
}


static void
vzctDomainClearVolumes(virDomainDef *def,
                       vzctl_env_handle *env)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        struct vzctl_disk_param param = { .enabled = 0 };
        const vzctDomainDiskPrivate *diskPrivate =
            (const vzctDomainDiskPrivate *)disk->privateData;
        char emptypath[1] = "";
        int rc;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        memcpy(param.uuid, diskPrivate->uuid, sizeof(param.uuid));
        param.path = emptypath;

        if ((rc = vzctl2_env_set_disk(env, &param)))
            VZCT_ERROR_MESSAGE_CODE(vzctl2_env_set_disk, rc);
    }
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC(virObjectEvent, virObjectUnref);

static int
vzctDomainStartFull(vzctDriver *driver,
                    virDomainObj *dom)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(virObjectEvent) event = NULL;
    g_autoptr(virDomainDef) def = NULL;
    vzctl_env_status_t status;
    bool started = false;
    bool suspended;
    bool job = false;
    int ret = -1;
    int err;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is already running"));
        goto cleanup;
    }

    if (vzctDomainCleanupGarbage(dom) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_SUSPENDED));

    suspended = !!(status.mask & ENV_STATUS_SUSPENDED);

    if (suspended) {
        struct vzctl_cpt_param param;
        g_autofree char *xmlfile = NULL;
        bool saveXml = false;

        xmlfile = vzctManagedSaveXMLFile(dom);
        saveXml = virFileExists(xmlfile);
        if (saveXml) {
            if (!(def = virDomainDefParseFile(xmlfile, driver->xmlopt, NULL,
                                              VIR_DOMAIN_DEF_PARSE_INACTIVE |
                                              VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                                              VIR_DOMAIN_DEF_PARSE_ALLOW_POST_PARSE_FAIL)))
                goto cleanup;
        } else {
            VIR_WARN("restoring without libvirt xml for saved image");
        }

        memset(&param, 0, sizeof(param));
        param.cmd = VZCTL_CMD_RESTORE;
        VZCT_CHECK(cleanup, vzctl2_env_restore(env, &param, 0));

        if (saveXml && unlink(xmlfile) < 0)
            VIR_WARN("cannot unlink managed save, name: %s, errno: %d",
                     dom->def->name, errno);

        event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STARTED,
                                     VIR_DOMAIN_EVENT_STARTED_RESTORED);
    } else {
        if (vzctDomainAddCdrom(env, dom->def) < 0)
            goto cleanup;

        VZCT_CHECK(cleanup, vzctl2_env_start(env, 0));

        event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STARTED,
                                     VIR_DOMAIN_EVENT_STARTED_BOOTED);
    }
    started = true;
    if (vzctDomainStart(driver, dom, def) < 0)
        goto cleanup;

    if (virHookPresent(VIR_HOOK_DRIVER_VZCT) &&
        virHookCall(VIR_HOOK_DRIVER_VZCT, dom->def->name,
                    VIR_HOOK_VZCT_OP_STARTED,
                    VIR_HOOK_SUBOP_NONE, NULL, NULL, NULL) < 0) {
        vzctDomainStop(driver, dom);
        goto cleanup;
    }


    virObjectEventStateQueue(driver->eventsManager, event);
    event = NULL;
    ret = 0;

 cleanup:
    if (ret < 0 && started)
        vzctDomainStopVzctl(dom, true);
    if (ret < 0 && env)
        vzctDomainClearVolumes(dom->def, env);
    if (job)
        vzctDomainEndJob(dom);

    return ret;
}


static int
vzctDomainCreateWithFlags(virDomainPtr domain,
                          unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCreateWithFlagsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainStartFull(driver, dom) < 0)
        goto cleanup;

    ret = 0;

 cleanup:

    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCreate(virDomainPtr domain)
{
    return vzctDomainCreateWithFlags(domain, 0);
}


static int
vzctDomainDestroyFlags(virDomainPtr domain,
                       unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virObjectEvent *event;
    virDomainObj *dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_DESTROY_GRACEFUL, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainDestroyFlagsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if (vzctDomainStopVzctl(dom, flags & VIR_DOMAIN_DESTROY_GRACEFUL) < 0)
        goto cleanup;

    vzctDomainStop(driver, dom);
    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STOPPED,
                                     VIR_DOMAIN_EVENT_STOPPED_DESTROYED);
    virObjectEventStateQueue(driver->eventsManager, event);
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainDestroy(virDomainPtr domain)
{
    return vzctDomainDestroyFlags(domain, 0);
}


static int
vzctDomainShutdownFlags(virDomainPtr domain,
                        unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virObjectEvent *event;
    virDomainObj *dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainShutdownFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if (vzctDomainStopVzctl(dom, true) < 0)
        goto cleanup;

    vzctDomainStop(driver, dom);
    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STOPPED,
                                     VIR_DOMAIN_EVENT_STOPPED_SHUTDOWN);
    virObjectEventStateQueue(driver->eventsManager, event);
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainShutdown(virDomainPtr domain)
{
    return vzctDomainShutdownFlags(domain, 0);
}


static int
vzctDomainGetInfo(virDomainPtr domain,
                  virDomainInfoPtr info)
{
    vzctDriver *driver = domain->conn->privateData;
    struct vzctl_cpustat cpustat;
    g_autoptr(vzctl_env_handle) env = NULL;
    virDomainObj *dom;
    bool job = false;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetInfoEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    info->state = virDomainObjGetState(dom, NULL);
    info->memory = dom->def->mem.cur_balloon;
    info->maxMem = virDomainDefGetMemoryTotal(dom->def);
    info->nrVirtCpu = virDomainDefGetVcpusMax(dom->def);


    if (virDomainObjIsActive(dom)) {
        int err;

        VZCT_CHECK_PTR(cleanup, env,
                       vzctl2_env_open(dom->def->extraId,
                                       VZCTL_CONF_SKIP_PARSE, &err));

        VZCT_CHECK(cleanup, vzctl2_env_cpustat(env, &cpustat, sizeof(cpustat)));
        /* secs to nanosecs */
        info->cpuTime = (cpustat.user + cpustat.system + cpustat.nice) * 1000000000ULL;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainIsPersistent(virDomainPtr domain)
{
    virDomainObj *dom;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainIsPersistentEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = 1;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static char *
vzctDomainGetOSType(virDomainPtr domain)
{
    virDomainObj *dom;
    char *ret = NULL;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetOSTypeEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = g_strdup(virDomainOSTypeToString(dom->def->os.type));

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCollectNetStats(virTypedParameterPtr *params,
                          int *nparams,
                          int *maxparams,
                          vzctl_env_handle_ptr env,
                          virDomainNetDef *net,
                          size_t idx)
{
    const char *ctid = vzctl2_env_get_ctid(env);
    char name[VIR_TYPED_PARAM_FIELD_LENGTH];
    struct vzctl_netstat stats;
    const char *ifname;

    if (net->type == VIR_DOMAIN_NET_TYPE_VENET0)
        ifname = VZCT_NET_VENET0;
    else
        ifname = net->ifname;

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.name", idx);
    if (virTypedParamsAddString(params, nparams, maxparams, name,
                                ifname) < 0)
        return -1;

    if (!env)
        return 0;

    VZCT_CHECK(error,
               vzctl2_get_env_netstat(ctid, ifname, &stats, sizeof(stats)));

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.rx.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.incoming) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.rx.pkts", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.incoming_pkt) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.tx.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.outgoing) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "net.%zu.tx.pkts", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.outgoing_pkt) < 0)
        return -1;

    return 0;

 error:
    return -1;
}


static int
vzctDomainCollectDiskStats(virTypedParameterPtr *params,
                           int *nparams,
                           int *maxparams,
                           vzctl_env_handle_ptr env,
                           virDomainDiskDef *disk,
                           size_t idx,
                           bool active)
{
    vzctDomainDiskPrivate *priv = (vzctDomainDiskPrivate *)disk->privateData;
    char name[VIR_TYPED_PARAM_FIELD_LENGTH];
    struct vzctl_disk_stats stats;
    unsigned long long usage;
    int rc;

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.name", idx);
    if (virTypedParamsAddString(params, nparams, maxparams, name,
                                disk->dst) < 0)
        return -1;

    /*
     * In case of offline domain volume path etc can be outdated. So don't
     * report anything besides name. This is on par with qemu driver.
     */
    if (disk->src->type == VIR_STORAGE_TYPE_VOLUME && !env)
        return 0;

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.path", idx);
    if (virTypedParamsAddString(params, nparams, maxparams, name,
                                disk->src->path) < 0)
        return -1;

    VZCT_CHECK_RET(-1, vzctl2_get_disk_usage(disk->src->path, &usage));

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.physical", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name, usage) < 0)
        return -1;

    if (!env)
        return 0;

    rc = vzctl2_env_get_disk_stats(env, priv->uuid, &stats, sizeof(stats));
    if (rc == VZCTL_E_NOTSUPP) {
        /* disk is attached backup probably, stats are not provided */
        return 0;
    } else if (rc) {
        VZCT_ERROR_MESSAGE_CODE(vzctl2_env_get_disk_stats, rc);
        return -1;
    }

    /* convert kB to bytes */
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.capacity", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.total << 10) < 0)
        return -1;

    if (!active)
        return 0;

    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.rd.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.read) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.wr.bytes", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.write) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.rd.reqs", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.read_reqs) < 0)
        return -1;
    g_snprintf(name, VIR_TYPED_PARAM_FIELD_LENGTH, "block.%zu.wr.reqs", idx);
    if (virTypedParamsAddULLong(params, nparams, maxparams, name,
                                stats.io.write_reqs) < 0)
        return -1;

    return 0;
}


static virDomainStatsRecordPtr
vzctDomainCollectStats(virConnectPtr conn,
                       virDomainObj *dom,
                       int stats,
                       unsigned int flags)
{
    virDomainDef *def;
    virDomainStatsRecordPtr tmp;
    virDomainStatsRecordPtr ret = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    int maxparams = 0;
    long long timeout;
    bool job = false;
    size_t i;
    int err;

    virObjectLock(dom);
    tmp = g_new0(virDomainStatsRecord, 1);

    if (flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_NOWAIT)
        timeout = 0;
    else
        timeout = VZCT_JOB_WAIT_TIME;

    if (vzctDomainBeginJobInternal(dom, VZCT_DOMAIN_JOB_QUERY, timeout) < 0)
        goto cleanup;
    job = true;

    def = dom->def;

    if (dom->state.state == VIR_DOMAIN_NOSTATE)
        stats &= VIR_DOMAIN_STATS_STATE;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (stats & VIR_DOMAIN_STATS_STATE) {
        int state;
        int reason;

        state = virDomainObjGetState(dom, &reason);
        if (virTypedParamsAddInt(&tmp->params, &tmp->nparams, &maxparams,
                                 "state.state", state) < 0 ||
            virTypedParamsAddInt(&tmp->params, &tmp->nparams, &maxparams,
                                 "state.reason", reason) < 0)
            goto cleanup;
    }

    if ((stats & VIR_DOMAIN_STATS_CPU_TOTAL) &&
         virDomainObjIsActive(dom)) {
        struct vzctl_cpustat cpustats;
        unsigned long long time;

        VZCT_CHECK(cleanup, vzctl2_env_cpustat(env, &cpustats, sizeof(cpustats)));

        /* secs to nanosecs */
        time = (cpustats.user + cpustats.nice + cpustats.system) * 1000000000ULL;
        if (virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "cpu.time", time) < 0 ||
            virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "cpu.user",
                                    cpustats.user * 1000000000ULL) < 0 ||
            virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "cpu.system",
                                    cpustats.system * 1000000000ULL) < 0)
            goto cleanup;
    }

    if (stats & VIR_DOMAIN_STATS_BALLOON) {
        unsigned long long mem = virDomainDefGetMemoryTotal(def);

        if (virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "balloon.current", mem) < 0 ||
            virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                    "balloon.maximum", mem) < 0)
            goto cleanup;

        if (virDomainObjIsActive(dom)) {
            struct vzctl_meminfo meminfo;

            VZCT_CHECK(cleanup,
                       vzctl2_get_env_meminfo(dom->def->extraId,
                                              &meminfo, sizeof(meminfo)));

            /* bytes to KiB */
            if (virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.available",
                                        meminfo.total >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.unused",
                                        meminfo.free >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.disk_caches",
                                        meminfo.cached >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.swap_in",
                                        meminfo.swap_in >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.swap_out",
                                        meminfo.swap_out >> 10) < 0 ||
                virTypedParamsAddULLong(&tmp->params, &tmp->nparams, &maxparams,
                                        "balloon.rss",
                                        meminfo.rss >> 10) < 0)
            goto cleanup;
        }
    }

    if (stats & VIR_DOMAIN_STATS_VCPU) {
        int nvcpus = def->maxvcpus;

        /* unlimited */
        if (!nvcpus && (nvcpus = virHostCPUGetCount()) < 0)
            goto cleanup;

        if (virTypedParamsAddUInt(&tmp->params, &tmp->nparams, &maxparams,
                                  "vcpu.current", nvcpus) < 0 ||
            virTypedParamsAddUInt(&tmp->params, &tmp->nparams, &maxparams,
                                  "vcpu.maximum", nvcpus) < 0)
            goto cleanup;
    }

    if (stats & VIR_DOMAIN_STATS_INTERFACE) {
        if (virTypedParamsAddUInt(&tmp->params, &tmp->nparams,
                                  &maxparams, "net.count", def->nnets) < 0)
            goto cleanup;

        for (i = 0; i < def->nnets; i++) {
            if (vzctDomainCollectNetStats(&tmp->params,
                                          &tmp->nparams, &maxparams,
                                          virDomainObjIsActive(dom) ? env : NULL,
                                          def->nets[i], i) < 0)
                goto cleanup;
        }
    }

    if (stats & VIR_DOMAIN_STATS_BLOCK) {
        size_t ndisks = 0;
        size_t idx = 0;

        for (i = 0; i < def->ndisks; i++) {
            if (def->disks[i]->device == VIR_DOMAIN_DISK_DEVICE_DISK)
                ndisks++;
        }

        if (virTypedParamsAddUInt(&tmp->params, &tmp->nparams,
                                  &maxparams, "block.count", ndisks) < 0)
            goto cleanup;

        for (i = 0; i < def->ndisks; i++) {
            virDomainDiskDef *disk = def->disks[i];

            if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
                continue;

            if (vzctDomainCollectDiskStats(&tmp->params,
                                           &tmp->nparams, &maxparams,
                                           env, disk, idx++,
                                           virDomainObjIsActive(dom)) < 0)
                goto cleanup;
        }
    }

    if (!(tmp->dom = virGetDomain(conn, def->name, def->uuid, def->id)))
        goto cleanup;

    ret = g_steal_pointer(&tmp);

 cleanup:
    if (tmp) {
        virTypedParamsFree(tmp->params, tmp->nparams);
        VIR_FREE(tmp);
    }
    if (job)
        vzctDomainEndJob(dom);
    virObjectUnlock(dom);
    return ret;
}


static int
vzctConnectGetAllDomainStats(virConnectPtr conn,
                             virDomainPtr *domains,
                             unsigned int ndomains,
                             unsigned int stats,
                             virDomainStatsRecordPtr **retStats,
                             unsigned int flags)
{
    vzctDriver *driver = conn->privateData;
    unsigned int supported = VIR_DOMAIN_STATS_STATE |
                             VIR_DOMAIN_STATS_CPU_TOTAL |
                             VIR_DOMAIN_STATS_BALLOON |
                             VIR_DOMAIN_STATS_VCPU |
                             VIR_DOMAIN_STATS_INTERFACE |
                             VIR_DOMAIN_STATS_BLOCK;
    unsigned int lflags = flags & (VIR_CONNECT_LIST_DOMAINS_FILTERS_ACTIVE |
                                   VIR_CONNECT_LIST_DOMAINS_FILTERS_PERSISTENT |
                                   VIR_CONNECT_LIST_DOMAINS_FILTERS_STATE);
    virDomainObj **doms = NULL;
    size_t ndoms;
    size_t i;
    int ret = -1;

    virCheckFlags(VIR_CONNECT_LIST_DOMAINS_FILTERS_ACTIVE |
                  VIR_CONNECT_LIST_DOMAINS_FILTERS_PERSISTENT |
                  VIR_CONNECT_LIST_DOMAINS_FILTERS_STATE |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_NOWAIT |
                  VIR_CONNECT_GET_ALL_DOMAINS_STATS_ENFORCE_STATS, -1);

    if (virConnectGetAllDomainStatsEnsureACL(conn) < 0)
        return -1;

    if (!stats) {
        stats = supported;
    } else if ((flags & VIR_CONNECT_GET_ALL_DOMAINS_STATS_ENFORCE_STATS) &&
        !(stats & supported)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED,
                       _("Stats types bits 0x%x are not supported"),
                       stats & ~supported);
        return -1;
    }

    if (ndomains) {
        if (virDomainObjListConvert(driver->domains, conn,
                                    domains, ndomains, &doms, &ndoms,
                                    virConnectGetAllDomainStatsCheckACL,
                                    lflags, true) < 0)
            return -1;
    } else {
        if (virDomainObjListCollect(driver->domains, conn, &doms, &ndoms,
                                    virConnectGetAllDomainStatsCheckACL,
                                    lflags) < 0)
            return -1;
    }

    *retStats = g_new0(virDomainStatsRecordPtr, ndoms + 1);

    ret = 0;
    for (i = 0; i < ndoms; i++) {
        if (!((*retStats)[ret] = vzctDomainCollectStats(conn, doms[i],
                                                        stats, flags))) {
            virResetLastError();
        } else {
            ret++;
        }
    }

    virObjectListFreeCount(doms, ndoms);

    return ret;
}


static int
vzctDomainSnapshotObjListStealMetaIter(void *payload,
                                       const char *name G_GNUC_UNUSED,
                                       void *data)
{
    virDomainMomentObj *snap = payload;
    virDomainMomentObj *peer;
    virDomainSnapshotObjList *src = data;

    if ((peer = virDomainSnapshotFindByName(src, snap->def->name))) {
        virDomainSnapshotDef *src_def = virDomainSnapshotObjGetDef(peer);
        virDomainSnapshotDef *dst_def = virDomainSnapshotObjGetDef(snap);

        dst_def->state = src_def->state;
        dst_def->memory = src_def->memory;
        if (!snap->def->dom)
            snap->def->dom = g_steal_pointer(&peer->def->dom);
        if (!snap->def->inactiveDom)
            snap->def->inactiveDom = g_steal_pointer(&peer->def->inactiveDom);
    }

    return 0;
}


static void
vzctDomainSnapshotObjListStealMeta(virDomainSnapshotObjList *dst,
                                   virDomainSnapshotObjList *src)
{
    virDomainSnapshotForEach(dst, vzctDomainSnapshotObjListStealMetaIter, src);
}


static void
vzctDomainFinishSnapshots(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv;

    if (!dom)
        return;

    priv  = dom->privateData;
    vzctDomainSnapshotObjListStealMeta(dom->snapshots, priv->snapshots);
    virDomainSnapshotObjListRemoveAll(priv->snapshots);
}


static int
vzctDomainReadSnapshots(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    struct vzctl_snapshot_tree* tree = NULL;
    g_autofree char *path = NULL;
    virDomainMomentObj *current = NULL;
    size_t i;
    int ret = -1;
    int err;

    path = g_strdup_printf("%s/Snapshots.xml", priv->bundlePath);
    if (!virFileExists(path))
        return 0;

    VZCT_CHECK_PTR(cleanup, tree,
                   vzctl2_open_snapshot_tree(path, &err));

    for (i = 0; i < tree->nsnapshots; i++) {
        struct vzctl_snapshot_data *data = tree->snapshots[i];
        virDomainSnapshotDef *def = NULL;
        virDomainMomentObj *snapshot;
        struct tm tm;
        const char *tmp;

        if (!(def = virDomainSnapshotDefNew()))
            goto cleanup;

        def->parent.name = g_strdup(data->guid);
        if (data->parent_guid && data->parent_guid[0] != '\0')
            def->parent.parent_name = g_strdup(data->parent_guid);
        def->parent.description = g_strdup(data->desc);
        tmp = strptime(data->date, "%Y-%m-%d %H:%M:%S", &tm);
        if (!tmp || *tmp != '\0') {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("unexpected datetime format: '%s'"), data->date);
            virObjectUnref(def);
            goto cleanup;
        }
        def->parent.creationTime = mktime(&tm);
        /* libvzctl does not provide state and thus memory */
        def->state = VIR_DOMAIN_SNAPSHOT_NOSTATE;
        def->memory = VIR_DOMAIN_SNAPSHOT_LOCATION_DEFAULT;

        if (!(snapshot = virDomainSnapshotAssignDef(priv->snapshots, def))) {
            virObjectUnref(def);
            goto cleanup;
        }

        if (data->current) {
            if (current) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("more then one current snapshots"));
                goto cleanup;
            }
            current = snapshot;
        }
    }

    virDomainSnapshotSetCurrent(priv->snapshots, current);
    if (virDomainSnapshotUpdateRelations(priv->snapshots) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("snapshots have inconsistent relations"));
        goto cleanup;
    }

    vzctDomainSnapshotObjListStealMeta(priv->snapshots, dom->snapshots);
    ret = 0;

 cleanup:
    if (ret < 0)
        virDomainSnapshotObjListRemoveAll(priv->snapshots);

    if (tree)
        vzctl2_close_snapshot_tree(tree);

    return ret;
}


static virDomainObj *
vzctDomObjFromSnapshot(virDomainSnapshotPtr snapshot)
{
    return vzctDomObjFromDomain(snapshot->domain);
}


static virDomainMomentObj *
vzctSnapObjFromName(virDomainSnapshotObjList *snapshots,
                    const char *name)
{
    virDomainMomentObj *snap = virDomainSnapshotFindByName(snapshots, name);

    if (!(snap = virDomainSnapshotFindByName(snapshots, name)))
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT,
                       _("no domain snapshot with matching name '%s'"),
                       name);

    return snap;
}


static virDomainMomentObj *
vzctSnapObjFromSnapshot(virDomainSnapshotObjList *snapshots,
                        virDomainSnapshotPtr snapshot)
{
    return vzctSnapObjFromName(snapshots, snapshot->name);
}


static char *
vzctDomainSnapshotGetXMLDesc(virDomainSnapshotPtr snapshot,
                             unsigned int flags)
{
    vzctDriver *driver = snapshot->domain->conn->privateData;
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    char *xml = NULL;
    virDomainMomentObj *snap = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    bool job = false;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_XML_SECURE, NULL);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return NULL;

    if (virDomainSnapshotGetXMLDescEnsureACL(snapshot->domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    virUUIDFormat(snapshot->domain->uuid, uuidstr);

    xml = virDomainSnapshotDefFormat(uuidstr, virDomainSnapshotObjGetDef(snap),
                                     driver->xmlopt,
                                     virDomainSnapshotFormatConvertXMLFlags(flags));

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return xml;
}


static int
vzctDomainSnapshotListNames(virDomainPtr domain,
                            char **names,
                            int nameslen,
                            unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSnapshotListNamesEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    n = virDomainSnapshotObjListGetNames(priv->snapshots, NULL, names, nameslen,
                                         flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotNum(virDomainPtr domain,
                      unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSnapshotNumEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    n = virDomainSnapshotObjListNum(priv->snapshots, NULL, flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainListAllSnapshots(virDomainPtr domain,
                           virDomainSnapshotPtr **snaps,
                           unsigned int flags)
{
    virDomainObj *dom = NULL;
    vzctDomainObjPrivate *priv;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainListAllSnapshotsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    n = virDomainListSnapshots(priv->snapshots, NULL, domain, snaps, flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotNumChildren(virDomainSnapshotPtr snapshot,
                              unsigned int flags)
{
    virDomainObj *dom = NULL;
    vzctDomainObjPrivate *priv;
    virDomainMomentObj *snap = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotNumChildrenEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    n = virDomainSnapshotObjListNum(priv->snapshots, snap, flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotListChildrenNames(virDomainSnapshotPtr snapshot,
                                    char **names,
                                    int nameslen,
                                    unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    virDomainMomentObj *snap = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotListChildrenNamesEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    n = virDomainSnapshotObjListGetNames(priv->snapshots, snap, names, nameslen,
                                         flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static int
vzctDomainSnapshotListAllChildren(virDomainSnapshotPtr snapshot,
                                  virDomainSnapshotPtr **snaps,
                                  unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    virDomainMomentObj *snap = NULL;
    bool job = false;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_LIST_TOPOLOGICAL |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotListAllChildrenEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    n = virDomainListSnapshots(priv->snapshots, snap, snapshot->domain, snaps,
                               flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return n;
}


static virDomainSnapshotPtr
vzctDomainSnapshotLookupByName(virDomainPtr domain,
                               const char *name,
                               unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom;
    virDomainMomentObj *snap = NULL;
    virDomainSnapshotPtr snapshot = NULL;
    bool job = false;

    virCheckFlags(0, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotLookupByNameEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromName(priv->snapshots, name)))
        goto cleanup;

    snapshot = virGetDomainSnapshot(domain, snap->def->name);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return snapshot;
}


static int
vzctDomainHasCurrentSnapshot(virDomainPtr domain,
                             unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainHasCurrentSnapshotEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    ret = (virDomainSnapshotGetCurrent(priv->snapshots) != NULL);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static virDomainSnapshotPtr
vzctDomainSnapshotGetParent(virDomainSnapshotPtr snapshot,
                            unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom;
    virDomainMomentObj *snap = NULL;
    virDomainSnapshotPtr parent = NULL;
    bool job = false;

    virCheckFlags(0, NULL);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return NULL;

    if (virDomainSnapshotGetParentEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    if (!snap->def->parent_name) {
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT,
                       _("snapshot '%s' does not have a parent"),
                       snap->def->name);
        goto cleanup;
    }

    parent = virGetDomainSnapshot(snapshot->domain, snap->def->parent_name);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return parent;
}


static virDomainSnapshotPtr
vzctDomainSnapshotCurrent(virDomainPtr domain,
                          unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom;
    virDomainSnapshotPtr snapshot = NULL;
    const char *name;
    bool job = false;

    virCheckFlags(0, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotCurrentEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(name = virDomainSnapshotGetCurrentName(priv->snapshots))) {
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT, "%s",
                       _("the domain does not have a current snapshot"));
        goto cleanup;
    }

    snapshot = virGetDomainSnapshot(domain, name);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return snapshot;
}


static int
vzctDomainSnapshotIsCurrent(virDomainSnapshotPtr snapshot,
                            unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    int ret = -1;
    virDomainMomentObj *snap = NULL;
    bool job = false;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotIsCurrentEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    ret = snap == virDomainSnapshotGetCurrent(priv->snapshots);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSnapshotHasMetadata(virDomainSnapshotPtr snapshot,
                              unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    int ret = -1;
    virDomainMomentObj *snap = NULL;
    bool job = false;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotHasMetadataEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    ret = 1;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


/*
 * We don't store `parent` and `is current` values in metadata for
 * the sake on simplicity. These values are taken from libvzctl.
 */
static int
vzctDomainSnapshotWriteMetadata(vzctDriver *driver,
                                virDomainObj *dom,
                                virDomainMomentObj *snapshot)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    virDomainSnapshotDef *def = virDomainSnapshotObjGetDef(snapshot);
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    g_autofree char *xml = NULL;
    g_autofree char *dir = NULL;
    g_autofree char *file = NULL;
    unsigned int flags = VIR_DOMAIN_SNAPSHOT_FORMAT_SECURE |
                         VIR_DOMAIN_SNAPSHOT_FORMAT_INTERNAL;

    virUUIDFormat(dom->def->uuid, uuidstr);
    if (!(xml = virDomainSnapshotDefFormat(uuidstr, def,
                                           driver->xmlopt, flags)))
        return -1;

    dir = g_strdup_printf("%s/"VZCT_SNAPHOTS_DIR, priv->bundlePath);
    if (g_mkdir_with_parents(dir, 0777) < 0) {
        virReportSystemError(errno, _("cannot create directory '%s'"), dir);
        return -1;
    }

    file = g_strdup_printf("%s/%s.xml", dir, def->parent.name);
    return virXMLSaveFile(file, NULL, "snapshot-edit", xml);
}


static virDomainSnapshotPtr
vzctDomainSnapshotCreateXML(virDomainPtr domain,
                            const char *xmlDesc,
                            unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainSnapshotPtr ret = NULL;
    g_autoptr(virDomainSnapshotDef) def = NULL;
    virDomainMomentObj *snap = NULL;
    virDomainObj *dom = NULL;
    struct vzctl_snapshot_param param = { 0 };
    g_autoptr(vzctl_env_handle) env = NULL;
    unsigned int parse_flags = 0;
    unsigned char uuid[VIR_UUID_BUFLEN];
    char uuidstr[VIR_UUID_STRING_BUFLEN + 2];
    bool job = false;
    int err;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_CREATE_VALIDATE, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotCreateXMLEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (flags & VIR_DOMAIN_SNAPSHOT_CREATE_VALIDATE)
        parse_flags |= VIR_DOMAIN_SNAPSHOT_PARSE_VALIDATE;
    if (!virDomainObjIsActive(dom))
        parse_flags |= VIR_DOMAIN_SNAPSHOT_PARSE_OFFLINE;

    if (!(def = virDomainSnapshotDefParseString(xmlDesc, driver->xmlopt,
                                                NULL, NULL, parse_flags)))
        goto cleanup;

    if (def->memory == VIR_DOMAIN_SNAPSHOT_LOCATION_EXTERNAL) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("external memory snapshots are not supported"));
        goto cleanup;
    }

    if (def->parent.name && !def->parent.name_generated) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("can not set name, it is currently only generated"));
        goto cleanup;
    }

    if (virDomainObjIsActive(dom)) {
        if (def->memory == VIR_DOMAIN_SNAPSHOT_LOCATION_DEFAULT)
            def->memory = VIR_DOMAIN_SNAPSHOT_LOCATION_INTERNAL;

        if (def->memory != VIR_DOMAIN_SNAPSHOT_LOCATION_INTERNAL) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("only internal memory location is supported"
                             " for active domains"));
            goto cleanup;
        }
    } else {
        if (def->memory == VIR_DOMAIN_SNAPSHOT_LOCATION_DEFAULT)
            def->memory = VIR_DOMAIN_SNAPSHOT_LOCATION_NONE;

        if (def->memory != VIR_DOMAIN_SNAPSHOT_LOCATION_NONE) {
            virReportError(VIR_ERR_INVALID_ARG, "%s",
                           _("memory location for inactive domains"
                             " must be none"));
            goto cleanup;
        }
    }

    if (!(def->parent.dom = vzctDomainDefCopy(dom->def, driver->xmlopt, true)))
        goto cleanup;

    if (virDomainObjIsActive(dom) &&
        !(def->parent.inactiveDom = vzctDomainDefCopy(dom->newDef,
                                                      driver->xmlopt, true)))
        goto cleanup;

    def->state = virDomainObjGetState(dom, NULL);

    /* libvzctl requires '{uuid}' form of uuid. */
    virUUIDGenerate(uuid);
    virUUIDFormat(uuid, uuidstr + 1);
    uuidstr[0] = '{';
    uuidstr[VIR_UUID_STRING_BUFLEN] = '}';
    uuidstr[VIR_UUID_STRING_BUFLEN + 1] = '\0';
    VIR_FREE(def->parent.name);
    def->parent.name_generated = false;
    def->parent.name = g_strdup(uuidstr);

    if (!(snap = virDomainSnapshotAssignDef(dom->snapshots, def)))
        goto cleanup;
    def = NULL;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (!virDomainObjIsActive(dom) &&
        vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    param.desc = snap->def->description;
    param.guid = snap->def->name;
    VZCT_CHECK_LOCKED(cleanup, env, "creating snapshot",
                      vzctl2_env_create_snapshot(env, &param));

    if (vzctDomainSnapshotWriteMetadata(driver, dom, snap) < 0)
        goto cleanup;

    ret = virGetDomainSnapshot(domain, snap->def->name);

 cleanup:
    if (!ret && snap)
        virDomainSnapshotObjListRemove(dom->snapshots, snap);

    if (env && !virDomainObjIsActive(dom))
        vzctDomainClearVolumes(dom->def, env);

    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainRevertToSnapshot(virDomainSnapshotPtr snapshot,
                           unsigned int flags)
{
    vzctDriver *driver = snapshot->domain->conn->privateData;
    virDomainMomentObj *snap = NULL;
    virDomainSnapshotDef *snapDef;
    virDomainDef *def;
    g_autoptr(virDomainDef) drvDef = NULL;
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    struct vzctl_switch_snapshot_param param = { 0 };
    g_autoptr(vzctl_env_handle) env = NULL;
    vzctl_env_status_t status;
    virDomainState state;
    bool job = false;
    int ret = -1;
    int err;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainRevertToSnapshotEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (!virDomainObjIsActive(dom) &&
        vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    param.guid = snap->def->name;
    VZCT_CHECK_LOCKED(cleanup,
                      env, "switching snapshot",
                      vzctl2_env_switch_snapshot(env, &param));

    if (virDomainObjIsActive(dom))
        vzctDomainStop(driver, dom);

    snapDef = virDomainSnapshotObjGetDef(snap);
    /*
     * If snapshot is done thru vzctl than we don't know domain state
     * in snapshot.
     */
    if (snapDef->state == VIR_DOMAIN_NOSTATE) {
        VZCT_CHECK(cleanup,
                   vzctl2_get_env_status(dom->def->extraId, &status,
                                         ENV_STATUS_RUNNING));

        if (status.mask & ENV_STATUS_RUNNING)
            state = VIR_DOMAIN_RUNNING;
        else
            state = VIR_DOMAIN_SHUTOFF;
    } else {
        state = snapDef->state;
    }

    if (!(def = vzctLoadDomainConfig(driver, dom->def->extraId, NULL)))
        goto cleanup;
    VIR_FREE(dom->def);
    VIR_FREE(dom->newDef);
    dom->def = g_steal_pointer(&def);
    dom->def->id = -1;

    if (snapDef->parent.inactiveDom)
        drvDef = snapDef->parent.inactiveDom;
    else
        drvDef = snapDef->parent.dom;

    /* Make a copy as merge damages second argument. */
    if (drvDef &&
        !(drvDef = vzctDomainDefCopy(drvDef, driver->xmlopt, false)))
        goto cleanup;
    if (vzctMergeConfigs(dom->def, drvDef) < 0)
        goto cleanup;

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (state == VIR_DOMAIN_RUNNING &&
        vzctDomainStart(driver, dom, snapDef->parent.dom) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    if (env && !virDomainObjIsActive(dom))
        vzctDomainClearVolumes(dom->def, env);

    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSnapshotDelete(virDomainSnapshotPtr snapshot,
                         unsigned int flags)
{
    virDomainMomentObj *snap = NULL;
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    bool job = false;
    int ret = -1;
    int err;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotDeleteEnsureACL(snapshot->domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainReadSnapshots(dom) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!(snap = vzctSnapObjFromSnapshot(priv->snapshots, snapshot)))
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (!virDomainObjIsActive(dom) &&
        vzctDomainResolveVolumes(dom->def, env) < 0)
        goto cleanup;

    VZCT_CHECK_LOCKED(cleanup, env, "deleting snapshot",
                      vzctl2_env_delete_snapshot(env, snap->def->name));

    if ((snap = vzctSnapObjFromSnapshot(dom->snapshots, snapshot))) {
        g_autofree char *path = NULL;

        path = g_strdup_printf("%s/"VZCT_SNAPHOTS_DIR"/%s.xml",
                               priv->bundlePath, snap->def->name);
        virDomainSnapshotObjListRemove(dom->snapshots, snap);
        if (unlink(path) < 0)
            VIR_WARN("Failed to unlink path=%s, errno=%d", path, errno);
    }

    ret = 0;

 cleanup:
    if (env && !virDomainObjIsActive(dom))
        vzctDomainClearVolumes(dom->def, env);
    if (job)
        vzctDomainEndJob(dom);
    vzctDomainFinishSnapshots(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}

/*
 * VIR_MIGRATE_NON_SHARED_DISK and VIR_MIGRATE_PARAM_MIGRATE_DISKS are
 * currently not really taken into account although allowed in API. In libvirt
 * migration API the client is responsible to specify which disks are shared
 * and need NOT to be copied and which are NON shared and need to be copied.
 * vzmigrate utility which is used for migration in turn make this desicions by
 * itself. vzmigrate desicions currently are not visible and cannot be sanity
 * checked against client specifications. However if both vzmigrate and client
 * are sane they should have same desicions in this respect.
 */

#define VZCT_MIGRATION_FLAGS \
    VIR_MIGRATE_LIVE | \
    VIR_MIGRATE_PERSIST_DEST | \
    VIR_MIGRATE_UNDEFINE_SOURCE | \
    VIR_MIGRATE_NON_SHARED_DISK | \
    VIR_MIGRATE_CHANGE_PROTECTION | \
    0

#define VZCT_MIGRATION_PARAMETERS \
    VIR_MIGRATE_PARAM_DEST_XML, VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_URI, VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_DEST_NAME, VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_MIGRATE_DISKS, VIR_TYPED_PARAM_STRING | \
                                     VIR_TYPED_PARAM_MULTIPLE, \
    VIR_MIGRATE_PARAM_DEST_XML, VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_PERSIST_XML, VIR_TYPED_PARAM_STRING, \
    NULL

typedef enum {
    /* channel creation channels for migration is supported */
    VZCT_MIGRATION_CAPS_BIT_CHANNEL = 0,

    VZCT_MIGRATION_CAPS_BIT_LAST
} vzctMigrationCapsBit;

VIR_ENUM_DECL(vzctMigrationCapsBit);
VIR_ENUM_IMPL(vzctMigrationCapsBit,
              VZCT_MIGRATION_CAPS_BIT_LAST,
              "channel",
);

typedef enum {
    /* channel creation channels for migration is supported */
    VZCT_MIGRATION_CAPS_CHANNEL = 1 << VZCT_MIGRATION_CAPS_BIT_CHANNEL,
} vzctMigrationCapsFlag;


typedef struct _vzctMigrationCookie vzctMigrationCookie;
typedef vzctMigrationCookie *vzctMigrationCookiePtr;
struct _vzctMigrationCookie {
    char *ctid;

    GHashTable *pools;

    unsigned int capabilities;
    int state;

    int ports[VZCT_MIGRATION_CHANNELS_NUM];
};

static void
vzctMigrationCookieFree(vzctMigrationCookie *cookie)
{
    if (!cookie)
        return;

    virHashFree(cookie->pools);
    cookie->pools = NULL;
    VIR_FREE(cookie->ctid);
    VIR_FREE(cookie);
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC(vzctMigrationCookie, vzctMigrationCookieFree);


static void
vzctBakeCookie(vzctMigrationCookie *cookie,
               char **cookieout,
               int *cookieoutlen)
{
    g_auto(virBuffer) buf = VIR_BUFFER_INITIALIZER;

    virBufferAddLit(&buf, "<vzct-migration>\n");
    virBufferAdjustIndent(&buf, 2);

    if (cookie->ctid)
        virBufferAsprintf(&buf, "<ctid>%s</ctid>\n", cookie->ctid);

    if (cookie->pools) {
        g_autofree virHashKeyValuePair *pools = NULL;
        size_t i;

        virBufferAddLit(&buf, "<volume-pools>\n");
        virBufferAdjustIndent(&buf, 2);
        pools = virHashGetItems(cookie->pools, NULL, false);
        for (i = 0; i < virHashSize(cookie->pools); i++)
            virBufferAsprintf(&buf, "<pool name='%s' path='%s'/>\n",
                              (const char *)pools[i].key,
                              (const char *)pools[i].value);
        virBufferAdjustIndent(&buf, -2);
        virBufferAddLit(&buf, "</volume-pools>\n");
    }

    if (cookie->state >= 0)
        virBufferAsprintf(&buf, "<state>%s</state>\n",
                          virDomainStateTypeToString(cookie->state));

    if (cookie->capabilities) {
        if (cookie->capabilities & VZCT_MIGRATION_CAPS_CHANNEL)
            virBufferAddLit(&buf, "<capability name='channel'/>\n");
    }

    if (cookie->ports[0]) {
        size_t i;

        for (i = 0; i < VZCT_MIGRATION_CHANNELS_NUM; i++)
            virBufferAsprintf(&buf, "<channel port='%d'/>", cookie->ports[i]);
    }

    virBufferAdjustIndent(&buf, -2);
    virBufferAddLit(&buf, "</vzct-migration>");

    *cookieout = virBufferContentAndReset(&buf);
    *cookieoutlen = strlen(*cookieout) + 1;
}


static vzctMigrationCookie *
vzctEatCookie(const char *cookiein,
              int cookieinlen)
{
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    g_autoptr(xmlDoc) doc = NULL;
    g_autoptr(xmlXPathContext) ctx = NULL;
    g_autofree xmlNodePtr *nodes = NULL;
    g_autofree char *state = NULL;
    size_t i;
    int n;

    cookie = g_new0(vzctMigrationCookie, 1);

    if (!cookiein || cookieinlen == 0)
        return g_steal_pointer(&cookie);

    if (cookiein[cookieinlen - 1] != '\0') {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Invalid migration cookie"));
        return NULL;
    }

    if (!(doc = virXMLParseStringCtxt(cookiein,
                                      _("(_migration_cookie)"), &ctx)))
        return NULL;

    cookie->ctid = virXPathString("string(./ctid[1])", ctx);

    if ((n = virXPathNodeSet("./volume-pools/pool", ctx, &nodes)) < 0)
        return NULL;

    if (n > 0 && !(cookie->pools = virHashNew(g_free)))
        return NULL;

    for (i = 0; i < n; i++) {
        g_autofree char *name = NULL;
        g_autofree char *path = NULL;

        if (!(name = virXMLPropString(nodes[i], "name"))) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("missing pool name"));
            return NULL;
        }

        /*
         * Future proof. Don't fail if path is missing. May be in future
         * we will deal with pools that does not have this attribute.
         */
        path = virXMLPropString(nodes[i], "path");
        if (path && virHashAddEntry(cookie->pools, name, path) < 0)
            return NULL;
        path = NULL;
    }

    /* Do not require state for backcompat reason. */
    if ((state = virXPathString("string(./state[1])", ctx)))
        cookie->state = virDomainStateTypeFromString(state);
    else
        cookie->state = -1;

    VIR_FREE(nodes);
    if ((n = virXPathNodeSet("./capability", ctx, &nodes)) < 0)
        return NULL;

    for (i = 0; i < n; i++) {
        g_autofree char *name = virXMLPropString(nodes[i], "name");
        int val;

        if ((val = vzctMigrationCapsBitTypeFromString(name)) < 0)
            continue;

        cookie->capabilities |= 1 << val;
    }


    VIR_FREE(nodes);
    if ((n = virXPathNodeSet("./channel", ctx, &nodes)) < 0)
        return NULL;

    if (n && n != VZCT_MIGRATION_CHANNELS_NUM) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("unsupported number of migration channels '%d'"), n);
        return NULL;
    }

    for (i = 0; i < n; i++) {
        g_autofree char *port = virXMLPropString(nodes[i], "port");

        if (!port ||
            virStrToLong_i(port, NULL, 10, &cookie->ports[i]) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("cannot parse port for channel '%d'"), (int)i);
            return NULL;
        }
    }

    return g_steal_pointer(&cookie);
}


static void
vzctDomainDefCopyPrivate(virDomainDef *dst,
                         virDomainDef *src)
{
    size_t i;

    for (i = 0; i < src->ndisks; i++) {
        virDomainDiskDef *srcDisk = src->disks[i];
        virDomainDiskDef *dstDisk = dst->disks[i];
        vzctDomainDiskPrivate *srcDiskPrivate;
        vzctDomainDiskPrivate *dstDiskPrivate;

        if (!srcDisk->privateData)
            continue;
        srcDiskPrivate = (vzctDomainDiskPrivate *)srcDisk->privateData;

        if (!dstDisk->privateData)
            dstDisk->privateData = vzctDomainDiskPrivateNew();
        dstDiskPrivate = (vzctDomainDiskPrivate *)dstDisk->privateData;

        memcpy(dstDiskPrivate->uuid, srcDiskPrivate->uuid,
               VZCT_DOMAIN_DISK_UUID_SIZE);
    }
}


static bool
vzctDomainDefCheckABIStabilityFlags(const virDomainDef *src,
                                    const virDomainDef *dst)
{
    size_t i;

    if (memcmp(src->uuid, dst->uuid, VIR_UUID_BUFLEN)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Target domain uuid does not match source"));
        return false;
    }
    if (STRNEQ_NULLABLE(src->name, dst->name)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain name '%s' does not match source '%s'"),
                       dst->name, src->name);
        return false;
    }

    if (src->virtType != dst->virtType) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain virt type %s does not match source %s"),
                       virDomainVirtTypeToString(dst->virtType),
                       virDomainVirtTypeToString(src->virtType));
        return false;
    }
    if (src->os.type != dst->os.type) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain OS type %s does not match source %s"),
                       virDomainOSTypeToString(dst->os.type),
                       virDomainOSTypeToString(src->os.type));
        return false;
    }
    if (src->os.arch != dst->os.arch) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain architecture %s does not match source %s"),
                       virArchToString(dst->os.arch),
                       virArchToString(src->os.arch));
        return false;
    }
    if (STRNEQ_NULLABLE(src->os.init, dst->os.init)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain init binary %s does not match source %s"),
                       dst->os.init, src->os.init);
        return false;
    }

    if (virDomainDefGetMemoryInitial(src) != virDomainDefGetMemoryInitial(dst)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain max memory %lld does not match source %lld"),
                       virDomainDefGetMemoryInitial(dst),
                       virDomainDefGetMemoryInitial(src));
        return false;
    }

    if (src->maxvcpus != dst->maxvcpus) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain vCPU max %zu does not match source %zu"),
                       dst->maxvcpus, src->maxvcpus);
        return false;
    }

    if (src->ndisks != dst->ndisks) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain disk count %zu does not match source %zu"),
                       dst->ndisks, src->ndisks);
        return false;
    }

    if (src->nnets != dst->nnets) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain net card count %zu does not match source %zu"),
                       dst->nnets, src->nnets);
        return false;
    }

    for (i = 0; i < src->nnets; i++) {
        char srcmac[VIR_MAC_STRING_BUFLEN];
        char dstmac[VIR_MAC_STRING_BUFLEN];

        if (virMacAddrCmp(&src->nets[i]->mac, &dst->nets[i]->mac) != 0) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                           _("Target network card mac %s does not match source %s"),
                           virMacAddrFormat(&dst->nets[i]->mac, dstmac),
                           virMacAddrFormat(&src->nets[i]->mac, srcmac));
            return false;
        }
    }

    return true;
}


static bool
vzctDomainCheckMigrateXML(virDomainDef *src,
                          virDomainDef *dst)
{
    size_t i;
    g_autofree char *srcstr = NULL;
    g_autofree char *dststr = NULL;

    /*
     * Don't use generic virDomainDefCheckABIStabilityFlags because
     * we simply ignore some parameters in domain xml on define. Thus
     * we can have false check failures because ignored parameter is not
     * present in src and may be present in dst.
     */
    if (!vzctDomainDefCheckABIStabilityFlags(src, dst))
        return false;

    if (src->mem.min_guarantee_auto != dst->mem.min_guarantee_auto ||
        src->mem.min_guarantee != dst->mem.min_guarantee) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("target memory guarantee does not match source"));
        return false;
    }

    if (src->cputune.shares != dst->cputune.shares ||
        src->cputune.quota != dst->cputune.quota ||
        src->cputune.period != dst->cputune.period) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("target cputune does not match source"));
        return false;
    }

    if (!virBitmapEqual(src->cpumask, dst->cpumask)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("target cpumask does not match source"));
        return false;
    }

    if (!virBitmapEqual(virDomainNumatuneGetNodeset(src->numa, NULL, -1),
                        virDomainNumatuneGetNodeset(dst->numa, NULL, -1))) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("target numatune does not match source"));
        return false;
    }

    for (i = 0; i < src->ndisks; i++) {
        if (STRNEQ_NULLABLE(src->disks[i]->src->path,
                            dst->disks[i]->src->path)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("target disk path does not match source"));
            return false;
        }

        if (src->disks[i]->info.bootIndex != dst->disks[i]->info.bootIndex) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("target bootindex does not match source"));
            return false;
        }

        if (STRNEQ_NULLABLE(src->disks[i]->blkdeviotune.group_name,
                            dst->disks[i]->blkdeviotune.group_name) ||
            !virDomainBlockIoTuneInfoEqual(&src->disks[i]->blkdeviotune,
                                           &dst->disks[i]->blkdeviotune)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("target blockdev iotune does not match source"));
            return false;
        }
    }

    for (i = 0; i < src->nnets; i++) {
        size_t j;

        if (src->nets[i]->type != dst->nets[i]->type ||
            STRNEQ_NULLABLE(src->nets[i]->ifname, dst->nets[i]->ifname) ||
            STRNEQ_NULLABLE(src->nets[i]->ifname_guest,
                            dst->nets[i]->ifname_guest)) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("target net does not match source"));
            return false;
        }

        switch ((int)src->nets[i]->type) {
        case VIR_DOMAIN_NET_TYPE_BRIDGE:
            if (STRNEQ_NULLABLE(src->nets[i]->data.bridge.brname,
                                dst->nets[i]->data.bridge.brname)) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("target net does not match source"));
                return false;
            }
            break;

        case VIR_DOMAIN_NET_TYPE_NETWORK:
            if (STRNEQ_NULLABLE(src->nets[i]->data.network.name,
                                dst->nets[i]->data.network.name)) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("target net does not match source"));
                return false;
            }
            break;

        case VIR_DOMAIN_NET_TYPE_VENET0:
            break;

        default:
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("unexpected net interface type: %s"),
                           virDomainNetTypeToString(src->nets[i]->type));
            return false;
        }

        if (src->nets[i]->virtPortProfile || dst->nets[i]->virtPortProfile) {
            if (src->nets[i]->virtPortProfile && dst->nets[i]->virtPortProfile) {
                const virNetDevVPortProfile *srcVP = src->nets[i]->virtPortProfile;
                const virNetDevVPortProfile *dstVP = dst->nets[i]->virtPortProfile;

                if (srcVP->virtPortType != dstVP->virtPortType ||
                    srcVP->interfaceID_specified != dstVP->interfaceID_specified ||
                    memcmp(srcVP->interfaceID, dstVP->interfaceID, VIR_UUID_BUFLEN) != 0) {
                    virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                                   _("target virtualport does not match source"));
                    return false;
                }

            } else {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("target virtualport does not match source"));
                return false;
            }
        }

        if (src->nets[i]->guestIP.nroutes != dst->nets[i]->guestIP.nroutes) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("target routes do not match source"));
            return false;
        }

        for (j = 0; j < src->nets[i]->guestIP.nroutes; j++) {
            if (!virSocketAddrEqual(&src->nets[i]->guestIP.routes[j]->gateway,
                                    &dst->nets[i]->guestIP.routes[j]->gateway)) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("target routes do not match source"));
                return false;
            }
        }

        if (src->nets[i]->guestIP.nips != dst->nets[i]->guestIP.nips) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("target IPs do not match source"));
            return false;
        }

        for (j = 0; j < src->nets[i]->guestIP.nips; j++) {
            const virNetDevIPAddr *srcIP = src->nets[i]->guestIP.ips[j];
            const virNetDevIPAddr *dstIP = dst->nets[i]->guestIP.ips[j];

            if (!virSocketAddrEqual(&srcIP->address, &dstIP->address) ||
                srcIP->prefix != dstIP->prefix) {
                virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                               _("target IPs do not match source"));
                return false;
            }
        }
    }

    if (src->blkio.weight != dst->blkio.weight) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Target domain blkio weight %u does not match source %u"),
                       dst->blkio.weight, src->blkio.weight);
        return false;
    }

    if (vzctDomainDefGetMetaAttr(src,
                                 VZCT_VHS_URI, VZCT_VHS_XPATH_OSTEMPLATE,
                                 NULL, &srcstr) < 0 ||
        vzctDomainDefGetMetaAttr(dst,
                                 VZCT_VHS_URI, VZCT_VHS_XPATH_OSTEMPLATE,
                                 NULL, &dststr) < 0 ||
        STRNEQ_NULLABLE(srcstr, dststr)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("target ostemplate does not match source"));
        return false;
    }
    VIR_FREE(srcstr);
    VIR_FREE(dststr);

    if (vzctDomainDefGetMetaAttr(src,
                                 VZCT_LIBOSINFO_URI, VZCT_LIBOSINFO_XPATH_OSID,
                                 NULL, &srcstr) < 0 ||
        vzctDomainDefGetMetaAttr(dst,
                                 VZCT_LIBOSINFO_URI, VZCT_LIBOSINFO_XPATH_OSID,
                                 NULL, &dststr) < 0 ||
        STRNEQ_NULLABLE(srcstr, dststr)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("target libosinfo OS ID does not match source"));
        return false;
    }

    return true;
}


static char *
vzctDomainMigrateBeginStep(vzctDriver *driver,
                           virDomainObj *dom,
                           virTypedParameterPtr params,
                           int nparams,
                           unsigned int flags)
{
    int nmigrate_disks = 0;
    const char **migrate_disks = NULL;
    const char *xml_param = NULL;
    const char *pxml_param = NULL;
    g_autoptr(virDomainDef) def_param = NULL;
    g_autoptr(virDomainDef) pdef_param = NULL;
    g_autofree char *miginactive = NULL;
    virDomainDef *def = NULL;
    virDomainDef *pdef = NULL;
    g_autofree char *xml = NULL;
    g_autofree char *pxml = NULL;
    size_t i;

    if ((flags & VIR_MIGRATE_LIVE) &&
        !virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("live migration only makes sense for active domains"));
        return NULL;
    }

    if (!(flags & VIR_MIGRATE_UNDEFINE_SOURCE)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("you need to set flag to undefine source because"
                         " not undefining is not supported"));
        return NULL;
    }

    if (!(flags & VIR_MIGRATE_PERSIST_DEST)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("you need to set flag to persist on destination because"
                         " not persisting is not supported"));
        return NULL;
    }

    /* params can be NULL and virTypedParamsGetStringList error out on NULL */
    if (params) {
        nmigrate_disks = virTypedParamsGetStringList(params, nparams,
                                                     VIR_MIGRATE_PARAM_MIGRATE_DISKS,
                                                     &migrate_disks);
        if (nmigrate_disks < 0)
            return NULL;
    }

    for (i = 0; i < nmigrate_disks; i++) {
        size_t j;

        for (j = 0; j < dom->def->ndisks; j++) {
            virDomainDiskDef *disk = dom->def->disks[j];

            if (STREQ(migrate_disks[i], disk->dst)) {
                if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK) {
                    virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                                   _("cannot migrate non disk devices"));
                    return NULL;
                }

                break;
            }
        }

        if (j == dom->def->ndisks) {
            virReportError(VIR_ERR_INVALID_ARG,
                           _("cannot find disk '%s' specified for copy"),
                           migrate_disks[i]);
            return NULL;
        }
    }


    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_XML,
                                &xml_param) < 0 ||
        virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_PERSIST_XML,
                                &pxml_param) < 0)
        return NULL;

    if ((xml_param &&
         !(def_param = virDomainDefParseString(xml_param,
                                              driver->xmlopt, NULL,
                                              VIR_DOMAIN_DEF_PARSE_INACTIVE))) ||
        (pxml_param &&
         !(pdef_param = virDomainDefParseString(pxml_param,
                                               driver->xmlopt, NULL,
                                               VIR_DOMAIN_DEF_PARSE_INACTIVE))))
        return NULL;

    if (virDomainObjIsActive(dom)) {
        def = dom->def;
        pdef = dom->newDef;
    } else {
        pdef = dom->def;
    }

    if ((virDomainObjIsActive(dom) &&
         def_param &&
         !vzctDomainCheckMigrateXML(def, def_param)) ||
        (pdef_param &&
         !vzctDomainCheckMigrateXML(pdef, pdef_param)))
        return NULL;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *migfile = vzctMigFile(dom);
        g_autofree char *migactive = vzctMigActiveFile(dom);

        /* keep old file for backcompat */
        if (virDomainObjSavePath(dom, driver->xmlopt, migfile) < 0)
            return NULL;

        if (def_param) {
            vzctDomainDefCopyPrivate(def_param, def);
            def = def_param;
        }

        if (!(xml = virDomainDefFormat(def, driver->xmlopt,
                                       VIR_DOMAIN_DEF_FORMAT_MIGRATABLE |
                                       VIR_DOMAIN_DEF_FORMAT_SECURE |
                                       VIR_DOMAIN_DEF_FORMAT_INACTIVE |
                                       VIR_DOMAIN_DEF_FORMAT_DISK_PRIVATE)))
            return NULL;

        if (virDomainDefSaveXML(def, migactive, xml, false) < 0)
            return NULL;
    }

    if (pdef_param) {
        vzctDomainDefCopyPrivate(pdef_param, pdef);
        pdef = pdef_param;
    }

    if (!(pxml = virDomainDefFormat(pdef, driver->xmlopt,
                                    VIR_DOMAIN_DEF_FORMAT_MIGRATABLE |
                                    VIR_DOMAIN_DEF_FORMAT_SECURE |
                                    VIR_DOMAIN_DEF_FORMAT_INACTIVE |
                                    VIR_DOMAIN_DEF_FORMAT_DISK_PRIVATE)))
        return NULL;

    miginactive = vzctMigInactiveFile(dom);
    if (virDomainDefSaveXML(pdef, miginactive, pxml, false) < 0)
        return NULL;

    if (virDomainObjIsActive(dom))
        return g_steal_pointer(&xml);
    else
        return g_steal_pointer(&pxml);
}


static char *
vzctDomainMigrateBegin3Params(virDomainPtr domain,
                              virTypedParameterPtr params,
                              int nparams,
                              char **cookieout,
                              int *cookieoutlen,
                              unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    vzctDomainObjPrivate *priv;
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    virDomainObj *dom;
    char* xml = NULL;
    bool job = false;

    virCheckFlags(VZCT_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return NULL;

    if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("change protection flag must be set"));

        return NULL;
    };

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MIGRATION) < 0)
        goto cleanup;
    job = true;
    priv = dom->privateData;
    priv->mig_conn_id = domain->conn->id;

    if (!(xml = vzctDomainMigrateBeginStep(driver, dom, params, nparams, flags)))
        goto cleanup;

    cookie = g_new0(vzctMigrationCookie, 1);
    cookie->ctid = g_strdup(dom->def->extraId);
    cookie->state = dom->state.state;
    cookie->capabilities = VZCT_MIGRATION_CAPS_CHANNEL;
    vzctBakeCookie(cookie, cookieout, cookieoutlen);

    vzctDomainAllowSideJob(dom);

 cleanup:
    if (job && !xml)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return xml;
}


static char*
vzctMigrationCreateURI(void)
{
    g_autofree char *hostname = NULL;
    char *uri;

    if (!(hostname = virGetHostname()))
        return NULL;

    if (STRPREFIX(hostname, "localhost")) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("hostname on destination resolved to localhost,"
                         " but migration requires an FQDN"));
        return NULL;
    }

    uri = g_strdup_printf("vzctmigr://%s", hostname);
    return uri;
}


static int
vzctDomainDstCookieAddPools(virConnectPtr conn,
                            virDomainDef *def,
                            vzctMigrationCookie *cookie)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        g_autoptr(virStoragePool) pool = NULL;
        g_autoptr(virStoragePoolDef) pooldef = NULL;
        g_autofree char *poolxml = NULL;
        g_autofree char *path = NULL;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        if (!(pool = virStoragePoolLookupByName(conn, disk->src->srcpool->pool)))
            return -1;

        if (!cookie->pools &&
            !(cookie->pools = virHashNew(g_free)))
            return -1;

        if (virHashLookup(cookie->pools, disk->src->srcpool->pool))
            continue;

        if (!(poolxml = virStoragePoolGetXMLDesc(pool, 0)))
            return -1;

        if (!(pooldef = virStoragePoolDefParseString(poolxml)))
            return -1;

        switch (pooldef->type) {
        case VIR_STORAGE_POOL_DIR:
        case VIR_STORAGE_POOL_FS:
        case VIR_STORAGE_POOL_NETFS:
        case VIR_STORAGE_POOL_VSTORAGE:
            path = g_strdup(pooldef->target.path);
            if (virHashAddEntry(cookie->pools, pooldef->name, path) < 0)
                return -1;
            path = NULL;
            break;

        default:
            virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                           _("pool type '%s' is not supported"),
                           virStoragePoolTypeToString(pooldef->type));
            return -1;
        }
    }

    return 0;
}


static int
vzctMigrationDstStartServer(vzctDriver *driver,
                            const vzctMigrationCookie *cookie,
                            vzctDomainMigrationData *data)
{
    g_autoptr(virCommand) serverCmd = NULL;
    size_t i;

    if (!cookie->ctid) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("missing ctid"));
        return -1;
    }

    VIR_DEBUG("Starting vzmdest migration server for ctid='%s'", cookie->ctid);

    serverCmd = virCommandNew("/usr/sbin/vzmdest");
    virCommandAddArg(serverCmd, "-ps");

    for (i = 0; i < VZCT_MIGRATION_CHANNELS_NUM; i++) {
        g_autoptr(virCommand) socatCmd = virCommandNew("socat");
        unsigned short port;
        int fds[2];

        if (virPortAllocatorAcquire(driver->migrationPorts, &port) < 0)
            return -1;
        data->ports[i] = port;

        /*
         * Below we pass own of fds to commands thus no any explicit
         * cleanup of fds.
        */
        if (socketpair(PF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, fds) < 0)
            return -1;

        /* log to syslog */
        virCommandAddArg(socatCmd, "-ly");
        virCommandAddArgFormat(socatCmd, "TCP4-LISTEN:%d", port);
        virCommandAddArgFormat(socatCmd, "FD:%d", fds[0]);
        virCommandPassFD(socatCmd, fds[0], VIR_COMMAND_PASS_FD_CLOSE_PARENT);

        virCommandAddArgFormat(serverCmd, "%d", fds[1]);
        virCommandPassFD(serverCmd, fds[1], VIR_COMMAND_PASS_FD_CLOSE_PARENT);

        if (virCommandRunAsync(socatCmd, &data->socatPids[i]) < 0)
            return -1;
    }

    virCommandAddArg(serverCmd, "0.0.0.0");
    virCommandAddArg(serverCmd, cookie->ctid);

    if (virCommandRunAsync(serverCmd, &data->pid) < 0)
        return -1;

    return 0;
}


static int
vzctDomainMigratePrepare3Params(virConnectPtr conn,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein,
                                int cookieinlen,
                                char **cookieout,
                                int *cookieoutlen,
                                char **uri_out,
                                unsigned int flags)
{
    vzctDriver *driver = conn->privateData;
    unsigned parseFlags = VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                          VIR_DOMAIN_DEF_PARSE_INACTIVE;
    g_autoptr(virDomainDef) def = NULL;
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    g_autofree vzctDomainMigrationData *data = NULL;
    const char *xml = NULL;
    const char *uri = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virCheckFlags(VZCT_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return -1;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_XML,
                                &xml) < 0)
        return -1;

    if (!xml) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("missing destination xml parameter"));
        return -1;
    }

    if (!(flags & VIR_MIGRATE_PERSIST_DEST)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("you need to set flag to persist on destination because"
                         " not persisting is not supported"));
        return -1;
    }

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_URI, &uri) < 0)
        return -1;

    if (!uri && !(*uri_out = vzctMigrationCreateURI()))
        return -1;

    if (!(def = virDomainDefParseString(xml, driver->xmlopt,
                                        NULL, parseFlags)))
        return -1;

    if (!(cookie = vzctEatCookie(cookiein, cookieinlen)))
        return -1;

    data = g_new0(vzctDomainMigrationData, 1);
    if (cookie->state == VIR_DOMAIN_RUNNING && def->ngraphics) {
        virDomainGraphicsDef *gr = def->graphics[0];

        if (gr->data.vnc.autoport) {
            unsigned short port;

            if (virPortAllocatorAcquire(driver->vncPorts, &port) < 0)
                return -1;
            data->vncPort = port;
        } else {
            if (virPortAllocatorSetUsed(gr->data.vnc.port) < 0)
                return -1;
            data->vncPort = gr->data.vnc.port;
        }

        VIR_DEBUG("Allocated port '%d' for domain '%s'", data->vncPort, def->name);
    }

    if (cookie->capabilities & VZCT_MIGRATION_CAPS_CHANNEL &&
        vzctMigrationDstStartServer(driver, cookie, data) < 0)
        goto error;

    vzctMigrationCookieFree(cookie);

    cookie = g_new0(vzctMigrationCookie, 1);
    memcpy(cookie->ports, data->ports, sizeof(data->ports));
    cookie->capabilities |= VZCT_MIGRATION_CAPS_CHANNEL;
    cookie->state = -1;
    if (vzctDomainDstCookieAddPools(conn, def, cookie) < 0)
        goto error;

    vzctBakeCookie(cookie, cookieout, cookieoutlen);

    virUUIDFormat(def->uuid, uuidstr);
    virHashAtomicUpdate(driver->migrating, uuidstr, data);
    data = NULL;

    return 0;

 error:
    vzctMigrationDataCleanup(data);
    data = NULL;
    return -1;
}


#define VZCT_VZMIGRATE_BIN "/usr/sbin/vzmigrate"

static int
vzctMigrationStartBackcompat(virDomainObj *dom,
                             const char *host,
                             const char *dname,
                             unsigned int flags)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    g_autoptr(virCommand) cmd = NULL;

    cmd = virCommandNew(VZCT_VZMIGRATE_BIN);
    virCommandAddArg(cmd, host);
    virCommandAddArg(cmd, dom->def->extraId);

    if (flags & VIR_MIGRATE_LIVE)
        virCommandAddArgList(cmd, "--online", NULL);
    if (dname)
        virCommandAddArgList(cmd, "--new-name", dname, NULL);

    if (virCommandRunAsync(cmd, &priv->migrationPid) < 0)
        return -1;

    return 0;
}


static int
vzctMigrationStartChannel(virDomainObj *dom,
                          const char *host,
                          const char *dname,
                          unsigned int flags,
                          const vzctMigrationCookie *cookie)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    g_autoptr(virCommand) cmd = NULL;
    size_t i;

    VIR_DEBUG("Starting vzmsrc migration client for ctid='%s'", dom->def->extraId);

    cmd = virCommandNew("/usr/sbin/vzmsrc");
    virCommandAddArg(cmd, "-ps");

    for (i = 0; i < VZCT_MIGRATION_CHANNELS_NUM; i++) {
        g_autoptr(virCommand) socatCmd = virCommandNew("socat");
        int fds[2];

        /*
         * Below we pass own of fds to commands thus no any explicit
         * cleanup of fds.
        */
        if (socketpair(PF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, fds) < 0)
            return -1;

        /* log to syslog */
        virCommandAddArg(socatCmd, "-ly");
        virCommandAddArgFormat(socatCmd, "TCP4:%s:%d", host, cookie->ports[i]);
        virCommandAddArgFormat(socatCmd, "FD:%d", fds[0]);
        virCommandPassFD(socatCmd, fds[0], VIR_COMMAND_PASS_FD_CLOSE_PARENT);

        virCommandAddArgFormat(cmd, "%d", fds[1]);
        virCommandPassFD(cmd, fds[1], VIR_COMMAND_PASS_FD_CLOSE_PARENT);

        if (virCommandRunAsync(socatCmd, &priv->socatPids[i]) < 0)
            return -1;
    }

    if (flags & VIR_MIGRATE_LIVE)
        virCommandAddArgList(cmd, "--online", NULL);
    if (dname)
        virCommandAddArgList(cmd, "--new-name", dname, NULL);

    virCommandAddArg(cmd, host);
    virCommandAddArg(cmd, dom->def->extraId);

    if (virCommandRunAsync(cmd, &priv->migrationPid) < 0)
        return -1;

    return 0;
}


static void
vzctMigrationCleanupChannel(virDomainObj *dom)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    size_t i;

    for (i = 0; i < VZCT_MIGRATION_CHANNELS_NUM; i++) {
        virProcessAbort(priv->socatPids[i]);
        priv->socatPids[i] = 0;
    }
}


static int
vzctMigrateVz(virDomainObj *dom,
              const char *uristr,
              const char *dname,
              unsigned int flags,
              const vzctMigrationCookie *cookie)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    bool channel = cookie->capabilities & VZCT_MIGRATION_CAPS_CHANNEL;
    g_autoptr(virURI) uri = NULL;
    pid_t pid;
    int status;
    int ret = -1;
    int rc;

    if (!(uri = virURIParse(uristr)))
        return -1;

    if (channel) {
        if (vzctMigrationStartChannel(dom, uri->server, dname, flags, cookie) < 0)
            goto cleanup;
    } else {
        if (vzctMigrationStartBackcompat(dom, uri->server, dname, flags) < 0)
            return -1;
    }

    pid = priv->migrationPid;
    vzctDomainAllowSideJob(dom);
    virObjectUnlock(dom);
    rc = virProcessWait(pid, &status, true);
    virObjectLock(dom);
    vzctDomainDisableSideJob(dom);
    priv->migrationPid = 0;

    if (rc < 0)
        goto cleanup;

    if (!(WIFEXITED(status) && WEXITSTATUS(status) == 0)) {
        if (priv->abortJob) {
            virReportError(VIR_ERR_OPERATION_ABORTED,
                           _("migration of CT '%s' cancelled by client"),
                           dom->def->extraId);
        } else {
            g_autofree char *st = virProcessTranslateStatus(status);

            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Migration error for CT '%s': %s"),
                           dom->def->extraId, st);
        }
        goto cleanup;
    }

    ret = 0;

 cleanup:
    if (channel)
        vzctMigrationCleanupChannel(dom);

    return ret;
}


static void
vzctMigrateAnyCleanupFiles(virDomainObj *dom)
{
    g_autofree char *miginactive = vzctMigInactiveFile(dom);
    g_autofree char *migactive = vzctMigActiveFile(dom);
    g_autofree char *migfile = vzctMigFile(dom);

    if (unlink(miginactive) < 0 && errno != ENOENT)
        VIR_WARN("Cannot remove migrating inactive config '%s', error '%d'",
                 migfile, errno);
    if (unlink(migactive) < 0 && errno != ENOENT)
        VIR_WARN("Cannot remove migrating active config '%s', error '%d'",
                 migfile, errno);
    if (unlink(migfile) < 0 && errno != ENOENT)
        VIR_WARN("Cannot remove migfile '%s', error '%d'",
                 migfile, errno);
}


/*
 * Container migration can be a success or a failure.
 * Migrated container can also be shared or non-shared.
 *
 * non-shared fail: need to cleanup
 * non-shared success: no need to cleanup, vzmigrate is supposed to
 * shared fail: need to cleanup
 * shared success: no need to cleanup to avoid conflict with dest
 */
static void
vzctMigrateSrcCleanup(virDomainObj *dom)
{
    vzctl_env_status_t status;

    if (vzctl2_get_env_status(dom->def->extraId, &status, ENV_STATUS_EXISTS)) {
        VIR_WARN("Cannot get status for '%s'", dom->def->extraId);
        return;
    }

    if (status.mask & ENV_STATUS_EXISTS)
        vzctMigrateAnyCleanupFiles(dom);
}


static int
vzctDomainSrcCheckPools(virDomainDef *def,
                        vzctMigrationCookie *cookie)
{
    g_autoptr(virConnect) conn = NULL;
    size_t i;

    if (!(conn = virGetConnectStorage()))
        return -1;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        g_autoptr(virStoragePool) pool = NULL;
        g_autoptr(virStoragePoolDef) pooldef = NULL;
        g_autofree char *poolxml = NULL;
        const char *path;

        if (disk->src->type != VIR_STORAGE_TYPE_VOLUME)
            continue;

        if (!cookie->pools) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("missing pools info from destination"));
            return -1;
        }

        if (!(path = virHashLookup(cookie->pools, disk->src->srcpool->pool))) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("missing info for pool '%s' from destination"),
                           disk->src->srcpool->pool);
            return -1;
        }

        if (!(pool = virStoragePoolLookupByName(conn, disk->src->srcpool->pool)))
            return -1;

        if (!(poolxml = virStoragePoolGetXMLDesc(pool, 0)))
            return -1;

        if (!(pooldef = virStoragePoolDefParseString(poolxml)))
            return -1;

        switch (pooldef->type) {
        case VIR_STORAGE_POOL_DIR:
        case VIR_STORAGE_POOL_FS:
        case VIR_STORAGE_POOL_NETFS:
        case VIR_STORAGE_POOL_VSTORAGE:
            if (STRNEQ(path, pooldef->target.path)) {
                virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                               _("pool '%s' have different path on src and dst"),
                               pooldef->name);
                return -1;
            }
            break;

        default:
            virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                           _("pool type '%s' is not supported"),
                           virStoragePoolTypeToString(pooldef->type));
            return -1;
        }
    }

    return 0;
}


static int
vzctDomainMigratePerformStep(vzctDriver *driver,
                             virDomainObj *dom,
                             virTypedParameterPtr params,
                             int nparams,
                             const char *cookiein,
                             int cookieinlen,
                             char **cookieout,
                             int *cookieoutlen,
                             unsigned int flags)
{
    virObjectEvent *event;
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    g_autoptr(vzctMigrationCookie) incookie = NULL;
    const char *uri = NULL;
    const char *dname = NULL;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_URI, &uri) < 0 ||
        virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_NAME, &dname) < 0)
        return -1;

    if (!uri) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("missing migration uri"));
        return -1;
    }

    if (!(incookie = vzctEatCookie(cookiein, cookieinlen)))
        return -1;

    if (vzctDomainSrcCheckPools(dom->def, incookie) < 0)
        return -1;

    if (!virDomainObjIsActive(dom)) {
        g_autoptr(vzctl_env_handle) env = NULL;
        int err;

        VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
        if (vzctDomainResolveVolumes(dom->def, env) < 0)
            return -1;
    }

    if (vzctMigrateVz(dom, uri, dname, flags, incookie) < 0) {
        g_autofree char *migfile = vzctMigFile(dom);

        if (unlink(migfile) < 0 && errno != ENOENT)
            VIR_WARN("Cannot remove migrating status file: %s, error: %d",
                     migfile, errno);

        if (!virDomainObjIsActive(dom)) {
            g_autoptr(vzctl_env_handle) env = NULL;
            int err;

            VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
            vzctDomainClearVolumes(dom->def, env);
        }

        vzctMigrateSrcCleanup(dom);
        return -1;
    }

    if (virDomainObjIsActive(dom)) {
        vzctDomainStopFlags(driver, dom, true);
        event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STOPPED,
                                     VIR_DOMAIN_EVENT_STOPPED_DESTROYED);
        virObjectEventStateQueue(driver->eventsManager, event);
    }

    virDomainObjListRemove(driver->domains, dom);
    cookie = g_new0(vzctMigrationCookie, 1);
    cookie->ctid = g_strdup(dom->def->extraId);
    cookie->state = -1;
    vzctBakeCookie(cookie, cookieout, cookieoutlen);
    event = virDomainEventLifecycleNewFromObj(dom,
                                 VIR_DOMAIN_EVENT_UNDEFINED,
                                 VIR_DOMAIN_EVENT_UNDEFINED_REMOVED);
    virObjectEventStateQueue(driver->eventsManager, event);

    return 0;
}


static int
vzctDomainMigrateConfirmStep(virDomainObj *dom,
                             int cancelled);

/*
 * On success steps order is next with begin step on source side
 * and each next step on the other side:
 * - begin/prepare/perform/finish/confirm
 *
 * On any error if the last step on a side was successful then last
 * step of success order on that side is executed with cancelled
 * flag set to 1. And if both finish and confirm have to be
 * executed then finish is executed first.
 *
 * Cookie is passed from step to step but on error the cleanup step
 * SHOULD not use cookie as its content is unspecified.
 *
 */
static int
vzctDomainMigratePerformP2P(vzctDriver *driver,
                            virDomainObj *dom,
                            const char *dconnuri,
                            virTypedParameterPtr params,
                            int nparams,
                            unsigned int flags)
{
    virErrorPtr orig_err = NULL;
    g_autofree char *uri = NULL;
    g_autofree char *cookiein = NULL;
    g_autofree char *cookieout = NULL;
    g_autofree char *xml = NULL;
    g_autoptr(virConnect) dconn = NULL;
    g_autoptr(virDomain) ddomain = NULL;
    int cookieinlen = 0;
    int cookieoutlen = 0;
    bool confirm = true;
    virTypedParameterPtr tmp;
    int ret = -1;
    int rc;

    virCheckFlags(VZCT_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return -1;

    if (virTypedParamsCopy(&tmp, params, nparams) < 0)
        return -1;
    params = tmp;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MIGRATION) < 0)
        goto cleanup;

    if (!(dconn = virConnectOpen(dconnuri)))
        goto cleanup;

    if (!(xml = vzctDomainMigrateBeginStep(driver, dom, params, nparams, flags)))
        goto cleanup;

    virTypedParamsReplaceString(&params, &nparams,
                                VIR_MIGRATE_PARAM_DEST_XML, xml);

    cookiein = g_steal_pointer(&cookieout);
    cookieinlen = cookieoutlen;
    cookieoutlen = 0;

    vzctDomainAllowSideJob(dom);
    virObjectUnlock(dom);
    rc = dconn->driver->domainMigratePrepare3Params
            (dconn, params, nparams, cookiein, cookieinlen,
             &cookieout, &cookieoutlen, &uri, flags);
    virObjectLock(dom);
    vzctDomainDisableSideJob(dom);
    if (rc < 0)
        goto confirm;

    if (uri)
        virTypedParamsReplaceString(&params, &nparams,
                                    VIR_MIGRATE_PARAM_URI, uri);

    VIR_FREE(cookiein);
    cookiein = g_steal_pointer(&cookieout);
    cookieinlen = cookieoutlen;
    cookieoutlen = 0;

    if (vzctDomainMigratePerformStep(driver, dom, params, nparams,
                                     cookiein, cookieinlen,
                                     &cookieout, &cookieoutlen,
                                     flags) < 0) {
        virErrorPreserveLast(&orig_err);
        confirm = false;
        goto finish;
    }

    ret = 0;

 finish:
    VIR_FREE(cookiein);
    cookiein = g_steal_pointer(&cookieout);
    cookieinlen = cookieoutlen;
    cookieoutlen = 0;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_NAME, NULL) == 0)
        virTypedParamsReplaceString(&params, &nparams,
                                    VIR_MIGRATE_PARAM_DEST_NAME,
                                    dom->def->name);
    vzctDomainAllowSideJob(dom);
    virObjectUnlock(dom);
    ddomain = dconn->driver->domainMigrateFinish3Params(dconn, params, nparams,
                                                        cookiein, cookieinlen,
                                                        &cookieout, &cookieoutlen,
                                                        flags,
                                                        ret == 0 ? 0 : 1);
    virObjectLock(dom);
    vzctDomainDisableSideJob(dom);
    if (ret == 0 && !ddomain)
        ret = -1;

 confirm:
    if (confirm)
        vzctDomainMigrateConfirmStep(dom, ret == 0 ? 0 : 1);

 cleanup:
    vzctDomainEndJob(dom);
    virTypedParamsFree(params, nparams);
    virErrorRestore(&orig_err);

    return ret;
}


static int
vzctDomainMigratePerformManaged(vzctDriver *driver,
                                virDomainObj *dom,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein,
                                int cookieinlen,
                                char **cookieout,
                                int *cookieoutlen,
                                unsigned int flags)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    int ret = -1;

    if (priv->job != VZCT_DOMAIN_JOB_MIGRATION) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("migration perform step without active migration"));
        return -1;
    }

    /*
     * Check even flags and params after job presence is established
     * so in case of bad parameters we can cleanup job.
     */
    if (flags & ~(VZCT_MIGRATION_FLAGS)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("unsupported flags"));
        goto cleanup;
    }

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        goto cleanup;

    if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("change protection flag must be set"));
        goto cleanup;
    };

    vzctDomainDisableSideJob(dom);
    if (vzctDomainMigratePerformStep(driver, dom, params, nparams,
                                     cookiein, cookieinlen,
                                     cookieout, cookieoutlen,
                                     flags) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    /*
     * End job for success path too as we remove domain on successful migration
     * on perform step and confirm step has no chances to find domain in list
     * and end migration job.
     */
    vzctDomainEndJob(dom);

    return ret;
}


static int
vzctDomainMigratePerform3Params(virDomainPtr domain,
                                const char *dconnuri G_GNUC_UNUSED,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein,
                                int cookieinlen,
                                char **cookieout,
                                int *cookieoutlen,
                                unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom;
    int ret;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (flags & VIR_MIGRATE_PEER2PEER)
        ret = vzctDomainMigratePerformP2P(driver, dom, dconnuri,
                                          params, nparams,
                                          flags & (~VIR_MIGRATE_PEER2PEER));
    else
        ret = vzctDomainMigratePerformManaged(driver, dom,
                                              params, nparams,
                                              cookiein, cookieinlen,
                                              cookieout, cookieoutlen,
                                              flags);

    virDomainObjEndAPI(&dom);
    return ret;
}


static virDomainPtr
vzctDomainMigrateFinish3Params(virConnectPtr dconn,
                               virTypedParameterPtr params,
                               int nparams,
                               const char *cookiein,
                               int cookieinlen,
                               char **cookieout G_GNUC_UNUSED,
                               int *cookieoutlen G_GNUC_UNUSED,
                               unsigned int flags,
                               int cancelled)
{
    vzctDriver *driver = dconn->privateData;
    g_autoptr(vzctMigrationCookie) cookie = NULL;
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    virDomainPtr domain = NULL;

    virCheckFlags(VZCT_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return NULL;

    if (cancelled) {
        g_autoptr(virDomainDef) def = NULL;
        vzctDomainMigrationData *migdata;
        char uuidstr[VIR_UUID_STRING_BUFLEN];
        const char *xml;

        /*
         * This can be reached in 2 cases: container migration was a failure or
         * a success. Migrated container can also be shared or non-shared.
         * Container migration success is possible, for example, if
         * a connection error occurred during the perfom phase in non-p2p
         * migration.
         *
         * non-shared fail: no need to cleanup, vzmigrate is supposed to
         * non-shared success: TODO need to clear domain migration flag
         * shared fail: need to cleanup, source will do
         * shared success: TODO same as non-shared
         *
         * Cleanup here is deleting migration configs.
         *
         * The success scenario is rare (connection error), so this can be done later,
         * as it seems to require implementing an extra state.
         */
        if (virTypedParamsGetString(params, nparams,
                                    VIR_MIGRATE_PARAM_DEST_XML, &xml) < 0)
            return NULL;

        if (!(def = virDomainDefParseString(xml, driver->xmlopt, NULL,
                                            VIR_DOMAIN_DEF_PARSE_INACTIVE)))
            return NULL;

        virUUIDFormat(def->uuid, uuidstr);
        migdata = virHashAtomicSteal(driver->migrating, uuidstr);
        if (!migdata &&
            (dom = virDomainObjListFindByUUID(driver->domains, domain->uuid))) {
            priv = dom->privateData;
            migdata = g_steal_pointer(&priv->migdata);
            if (vzctDomainSaveConfig(driver, dom) < 0)
                VIR_WARN("failed to save inactive config");

            virDomainObjEndAPI(&dom);
        }
        vzctMigrationDataCleanup(migdata);
        migdata = NULL;

        /* need to report some error */
        virReportError(VIR_ERR_MIGRATE_FINISH_OK, NULL);
        return NULL;
    }

    if (!(cookie = vzctEatCookie(cookiein, cookieinlen)))
        return NULL;

    if (!cookie->ctid) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("missing ctid"));
        return NULL;
    }

    if (!(dom = vzctAddDomain(driver, cookie->ctid, NULL)))
        return NULL;

    /* Move config to a permanent place */
    priv = dom->privateData;
    vzctMigrationDataCleanup(priv->migdata);
    priv->migdata = NULL;
    if (vzctDomainSaveConfig(driver, dom) < 0)
        VIR_WARN("failed to save inactive config");
    vzctMigrateAnyCleanupFiles(dom);

    if (dom->state.state == VIR_DOMAIN_NOSTATE) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("domain is out of sync on destination"));
        goto cleanup;
    }

    if (!virDomainObjIsActive(dom)) {
        g_autoptr(vzctl_env_handle) env = NULL;
        int err;

        VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));
        vzctDomainClearVolumes(dom->def, env);
    }

    domain = virGetDomain(dconn, dom->def->name, dom->def->uuid, dom->def->id);

 cleanup:
    virDomainObjEndAPI(&dom);
    return domain;
}


static int
vzctDomainMigrateConfirmStep(virDomainObj *dom,
                             int cancelled)
{
    if (cancelled) {
        vzctl_env_status_t status;

        vzctMigrateSrcCleanup(dom);

        VZCT_CHECK_RET(0, vzctl2_get_env_status(dom->def->extraId, &status,
                                                ENV_STATUS_RUNNING));

        /*
         * Check status before deleting the file. The problem is we can
         * get cancelled flag set if perform was successful and finish
         * was failed and private directory is shared one. In this case
         * we don't want to mess with destination which will use this
         * file.
         */
        if (status.mask & ENV_STATUS_RUNNING) {
            g_autofree char *migfile = vzctMigFile(dom);

            if (unlink(migfile) < 0 && errno != ENOENT)
                VIR_WARN("Cannot remove migrating status file: %s, error: %d",
                         migfile, errno);
        }
    }
    return 0;
}


static int
vzctDomainMigrateConfirm3Params(virDomainPtr domain,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein G_GNUC_UNUSED,
                                int cookieinlen G_GNUC_UNUSED,
                                unsigned int flags,
                                int cancelled)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom;
    bool job = false;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return 0;

    priv = dom->privateData;
    if (priv->job != VZCT_DOMAIN_JOB_MIGRATION) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("migration confirm step without active migration"));
        goto cleanup;
    }
    job = true;

    /*
     * Check even flags and params after job presence is established
     * so in case of bad parameters we can cleanup job.
     */
    if (flags & ~(VZCT_MIGRATION_FLAGS)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("unsupported flags"));
        goto cleanup;
    }

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        goto cleanup;

    if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("change protection flag must be set"));
        goto cleanup;
    };

    vzctDomainDisableSideJob(dom);
    vzctDomainMigrateConfirmStep(dom, cancelled);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return 0;
}


static int
vzctConnectDomainEventRegisterAny(virConnectPtr conn,
                                  virDomainPtr dom,
                                  int eventID,
                                  virConnectDomainEventGenericCallback callback,
                                  void *opaque,
                                  virFreeCallback freecb)
{
    vzctDriver *driver = conn->privateData;
    int ret = -1;

    if (virConnectDomainEventRegisterAnyEnsureACL(conn) < 0)
        return -1;

    if (virDomainEventStateRegisterID(conn,
                                      driver->eventsManager,
                                      dom, eventID,
                                      callback, opaque, freecb, &ret) < 0)
        return -1;

    return ret;
}


static int
vzctConnectDomainEventDeregisterAny(virConnectPtr conn,
                                    int callbackID)
{
    vzctDriver *driver = conn->privateData;

    if (virConnectDomainEventDeregisterAnyEnsureACL(conn) < 0)
        return -1;

    if (virObjectEventStateDeregisterID(conn,
                                        driver->eventsManager,
                                        callbackID, true) < 0)
        return -1;

    return 0;
}


static int
vzctDomainGetJobStatsInternal(virDomainObj *dom,
                              vzctDomainJobStats **ret)
{
    vzctDomainObjPrivate *priv = dom->privateData;
    g_autofree vzctDomainJobStats *stats = NULL;
    unsigned long long now;

    if (priv->job != VZCT_DOMAIN_JOB_MIGRATION) {
        *ret = NULL;
        return 0;
    }

    if (virTimeMillisNow(&now) < 0)
        return -1;

    stats = g_new0(vzctDomainJobStats, 1);
    stats->timeElapsed = now - priv->jobStarted;

    *ret = g_steal_pointer(&stats);
    return 0;
}


static int
vzctDomainGetJobInfo(virDomainPtr domain,
                     virDomainJobInfoPtr info)
{
    g_autofree vzctDomainJobStats *stats = NULL;
    virDomainObj *dom;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetJobInfoEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainGetJobStatsInternal(dom, &stats) < 0)
        goto cleanup;

    memset(info, 0, sizeof(*info));
    if (stats) {
        info->type = VIR_DOMAIN_JOB_UNBOUNDED;
        info->timeElapsed = stats->timeElapsed;
    } else {
        info->type = VIR_DOMAIN_JOB_NONE;
    }
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctConvertStatsToParams(vzctDomainJobStats *stats,
                         virTypedParameterPtr *params_out,
                         int *nparams_out)
{
    virTypedParameterPtr params = NULL;
    int maxparams = 0;
    int nparams = 0;
    int ret = -1;

    if (!stats) {
        *params_out = NULL;
        *nparams_out = 0;
        return 0;
    }

    if (virTypedParamsAddInt(&params, &nparams, &maxparams,
                             VIR_DOMAIN_JOB_OPERATION,
                             VIR_DOMAIN_JOB_OPERATION_MIGRATION_OUT) < 0)
        goto cleanup;

    if (virTypedParamsAddULLong(&params, &nparams, &maxparams,
                                VIR_DOMAIN_JOB_TIME_ELAPSED,
                                stats->timeElapsed) < 0)
        goto cleanup;

    *params_out = g_steal_pointer(&params);
    *nparams_out = nparams;
    nparams = 0;
    ret = 0;

 cleanup:
    virTypedParamsFree(params, nparams);
    return ret;
}


static int
vzctDomainGetJobStats(virDomainPtr domain,
                      int *type,
                      virTypedParameterPtr *params,
                      int *nparams,
                      unsigned int flags)
{
    g_autofree vzctDomainJobStats *stats = NULL;
    virDomainObj *dom;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetJobStatsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainGetJobStatsInternal(dom, &stats) < 0)
        goto cleanup;

    if (vzctConvertStatsToParams(stats, params, nparams) < 0)
        goto cleanup;

    *type = stats ? VIR_DOMAIN_JOB_UNBOUNDED : VIR_DOMAIN_JOB_NONE;
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainAbortJob(virDomainPtr domain)
{
    virDomainObj *dom;
    vzctDomainObjPrivate *priv;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainAbortJobEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    priv = dom->privateData;
    if (!priv->migrationPid) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       _("domain %s does not have active migration"),
                       dom->def->name);
        goto cleanup;
    }

    virProcessKill(priv->migrationPid, SIGTERM);
    priv->abortJob = true;

    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainGetAutostart(virDomainPtr domain,
                       int *autostart)
{
    virDomainObj *dom = NULL;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetAutostartEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    *autostart = dom->autostart;
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainSetAutostart(virDomainPtr domain,
                       int autostart)
{
    g_autofree char *autostartFile = NULL;
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    bool job = false;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetAutostartEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    autostart = !!autostart;
    if (dom->autostart == autostart) {
        ret = 0;
        goto cleanup;
    }

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    priv = dom->privateData;
    autostartFile = g_strdup_printf("%s/"VZCT_PRIVATE_DIR"/autostart", priv->bundlePath);

    if (autostart) {
        if (virFileTouch(autostartFile, 0600)) {
            virReportSystemError(errno,
                                 _("Failed to create autostart symlink '%s'"),
                                 autostartFile);
            goto cleanup;
        }
    } else {
        if (unlink(autostartFile) < 0 && errno != ENOENT) {
            virReportSystemError(errno,
                                 _("Failed to delete autostart symlink '%s'"),
                                 autostartFile);
            goto cleanup;
        }
    }

    dom->autostart = autostart;

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCommandMakePipe(virStreamPtr st,
                          int *fd)
{
    if (st) {
        int pair[2];

        /*
         * Use sockets instead of pipes due to the way streams are implemented
         * in rpc. On receiving finish/abort on stdout stream that does not
         * receive EOF yet rpc waits for writable event which will never
         * come if stdout is read end of pipe. Thus ^C on next command hangs:
         *
         * virsh -k0 -c vzct:///system x-exec 100 sleep 600
         */
        if (socketpair(PF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0, pair) < 0) {
            virReportSystemError(errno, "%s", _("socketpair() failed"));
            return -1;
        }

        if (virFDStreamOpen(st, pair[0]) < 0) {
            VIR_FORCE_CLOSE(pair[0]);
            VIR_FORCE_CLOSE(pair[1]);
            return -1;
        }
        *fd = pair[1];
    } else {
        if ((*fd = open("/dev/null", O_RDWR|O_CLOEXEC)) < 0) {
            virReportSystemError(errno, "%s", _("open(/dev/null) failed"));
            return -1;
        }
    }

    return 0;
}


static int
vzctDomainCommandExecInternal(virConnectPtr conn,
                              virDomainObj *dom,
                              const char *cmd,
                              const char **args,
                              unsigned int nargs,
                              const char **envs,
                              unsigned int nenvs,
                              int rows,
                              int columns,
                              virStreamPtr in,
                              virStreamPtr out,
                              virStreamPtr err,
                              unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_auto(GStrv) argv = NULL;
    g_auto(GStrv) envp = NULL;
    vzctDomainObjPrivate *priv;
    vzctDomainExecProcess *exec_proc;
    struct vzctl_exec_handle *exec_handle;
    int fds[3] = { -1, -1, -1 };
    int vz_err;
    size_t i;
    int ret = -1;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        return -1;
    }

    argv = g_new0(char *, nargs + 2);
    argv[0] = g_strdup(cmd);
    for (i = 0; i < nargs; i++)
        argv[i + 1] = g_strdup(args[i]);

    envp = g_new0(char *, nenvs + 1);
    for (i = 0; i < nenvs; i++)
        envp[1] = g_strdup(envs[i]);

    if (vzctDomainCommandMakePipe(in, &fds[0]) < 0 ||
        vzctDomainCommandMakePipe(out, &fds[1]) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                   VZCTL_CONF_SKIP_PARSE, &vz_err));

    if (flags & VIR_DOMAIN_COMMAND_X_EXEC_TERMINAL) {
        struct winsize ws;

        memset(&ws, 0, sizeof(ws));
        ws.ws_row = rows;
        ws.ws_col = columns;

        VZCT_CHECK(cleanup,
                   vzctl2_env_exec_pty(env, argv, envp, fds,
                                       NULL, &ws, &exec_handle));
    } else {
        exec_mode_e mode;

        if (flags & VIR_DOMAIN_COMMAND_X_EXEC_SHELL)
            mode = MODE_BASH;
        else
            mode = MODE_EXEC;

        if (vzctDomainCommandMakePipe(err, &fds[0]) < 0)
            goto cleanup;

        VZCT_CHECK(cleanup,
                   vzctl2_env_execve(env, argv, envp, fds, mode, &exec_handle));
    }

    priv = dom->privateData;
    exec_proc = g_new0(vzctDomainExecProcess, 1);
    exec_proc->handle = exec_handle;
    exec_proc->conn_id = conn->id;
    VIR_RESIZE_N(priv->exec_procs,
                 priv->exec_nprocs_max, priv->exec_nprocs, 1);
    priv->exec_procs[priv->exec_nprocs++] = exec_proc;

    ret = exec_handle->pid;

 cleanup:
    if (ret < 0) {
        virErrorPtr saveerr = virSaveLastError();

        if (fds[0] != -1)
            virStreamAbort(in);
        if (fds[1] != -1)
            virStreamAbort(out);
        if (fds[2] != -1)
            virStreamAbort(err);

        virSetError(saveerr);
        virFreeError(saveerr);
    }

    VIR_FORCE_CLOSE(fds[0]);
    VIR_FORCE_CLOSE(fds[1]);
    VIR_FORCE_CLOSE(fds[2]);

    return ret;
}


static int
vzctDomainCommandXExec(virDomainPtr domain,
                       const char *cmd,
                       const char **args,
                       unsigned int nargs,
                       const char **envs,
                       unsigned int nenvs,
                       virTypedParameterPtr params,
                       int nparams,
                       virStreamPtr in,
                       virStreamPtr out,
                       virStreamPtr err,
                       unsigned int flags)
{
    virDomainObj *dom = NULL;
    int rows = 0;
    int columns = 0;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_COMMAND_X_EXEC_SHELL |
                  VIR_DOMAIN_COMMAND_X_EXEC_TERMINAL, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                               VIR_TYPED_PARAM_UINT,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                               VIR_TYPED_PARAM_UINT,
                               NULL) < 0)
        return -1;

    if (flags & VIR_DOMAIN_COMMAND_X_EXEC_TERMINAL) {
        if (err) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("stderr is not available in terminal mode"));
            return -1;
        }

        if (virTypedParamsGetUInt(params, nparams,
                                  VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                                  (unsigned int *)&rows) < 0)
            return -1;

        if (virTypedParamsGetUInt(params, nparams,
                                  VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                                  (unsigned int *)&columns) < 0)
            return -1;

        if ((rows == -1 && columns >= 0) ||
            (columns == -1 && rows >= 0)) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("rows and columns should be either"
                             " both set or both unset"));
            return -1;
        }

        if (flags & VIR_DOMAIN_COMMAND_X_EXEC_SHELL) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("'shell' flag does not make sense with"
                             " 'terminal' flag"));
            return -1;
        }
    } else {
        if (virTypedParamsGet(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS) ||
            virTypedParamsGet(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS)) {

            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("window size parameters does not make sense "
                             "without terminal flag"));
            return -1;
        }
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXExecEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    ret = vzctDomainCommandExecInternal(domain->conn, dom,
                                        cmd, args, nargs, envs, nenvs,
                                        rows, columns, in, out, err,
                                        flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSearchExecHandle(virDomainObj *dom,
                           int pid_value)
{
    vzctDomainObjPrivate *priv;
    size_t i;

    priv = dom->privateData;
    for (i = 0; i < priv->exec_nprocs; i++) {
        if (priv->exec_procs[i]->handle->pid == pid_value)
            return i;
    }

    virReportError(VIR_ERR_INVALID_ARG, _("invalid pid %d"), pid_value);
    return -1;
}


static int
vzctDomainCommandGetStatusInternal(virDomainObj *dom,
                                   int pid_value,
                                   virDomainCommandXStatusPtr status,
                                   void *opaque G_GNUC_UNUSED)
{
    vzctDomainObjPrivate *priv;
    int wstatus;
    int idx;
    int rc;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        return -1;
    }

    if ((idx = vzctDomainSearchExecHandle(dom, pid_value)) < 0)
        return -1;

    priv = dom->privateData;
    rc = vzctl2_env_waitpid(priv->exec_procs[idx]->handle, 1, &wstatus);
    if (rc < 0) {
        const char *err = vzctl2_get_last_error();
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("vzctl2_env_waitpid failed with error %s"),
                       NULLSTR(err));
        return -1;
    }

    status->exited = 0;
    status->code = -1;
    status->signal = -1;
    if (rc == 0)
        return 0;

    status->exited = 1;
    if (WIFEXITED(wstatus)) {
        status->code = WEXITSTATUS(wstatus);
    } else {
        status->signal = WTERMSIG(wstatus);
    }

    vzctl2_release_exec_handle(priv->exec_procs[idx]->handle);
    VIR_DELETE_ELEMENT_INPLACE(priv->exec_procs, idx, priv->exec_nprocs);

    return 0;
}


static int
vzctDomainCommandXGetStatus(virDomainPtr domain,
                            int pid_value,
                            virDomainCommandXStatusPtr status,
                            unsigned int flags)
{
    virDomainObj *dom = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXGetStatusEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    ret = vzctDomainCommandGetStatusInternal(dom, pid_value, status, NULL);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCommandTerminateInternal(virDomainObj *dom,
                                   int pid_value,
                                   void *opaque G_GNUC_UNUSED)
{
    vzctDomainObjPrivate *priv;
    int idx;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        return -1;
    }

    if ((idx = vzctDomainSearchExecHandle(dom, pid_value)) < 0)
        return -1;

    priv = dom->privateData;
    vzctl2_env_exec_terminate(priv->exec_procs[idx]->handle);
    priv->exec_procs[idx]->terminated = true;
    return 0;
}


static int
vzctDomainCommandXTerminate(virDomainPtr domain,
                            int pid_value,
                            unsigned int flags)
{
    virDomainObj *dom = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXTerminateEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    ret = vzctDomainCommandTerminateInternal(dom, pid_value, NULL);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCommandXSetParams(virDomainPtr domain,
                            int pid_value,
                            virTypedParameterPtr params,
                            int nparams,
                            unsigned int flags)
{
    vzctDomainObjPrivate *priv;
    virDomainObj *dom = NULL;
    struct winsize ws;
    int rows = -1;
    int columns = -1;
    bool job = false;
    int idx;
    int ret = -1;

    virCheckFlags(0, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                               VIR_TYPED_PARAM_UINT,
                               VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                               VIR_TYPED_PARAM_UINT,
                               NULL) < 0)
        return -1;

    if (virTypedParamsGetUInt(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_ROWS,
                              (unsigned int *)&rows) < 0)
        return -1;

    if (virTypedParamsGetUInt(params, nparams,
                              VIR_DOMAIN_COMMAND_X_EXEC_PTY_COLUMNS,
                              (unsigned int *)&columns) < 0)
        return -1;

    if ((rows == -1 && columns >= 0) ||
        (columns == -1 && rows >= 0)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("rows and columns should be either both set or both unset"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainCommandXSetParamsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if ((idx = vzctDomainSearchExecHandle(dom, pid_value)) < 0)
        goto cleanup;
    priv = dom->privateData;

    memset(&ws, 0, sizeof(ws));
    ws.ws_col = columns;
    ws.ws_row = rows;

    VZCT_CHECK(cleanup,
               vzctl2_env_exec_set_winsize(priv->exec_procs[idx]->handle, &ws));
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainCheckModificationFlags(virDomainObj *dom,
                                 unsigned int flags)
{
    if (virDomainObjIsActive(dom)) {
        const unsigned int allflags = VIR_DOMAIN_AFFECT_LIVE |
                                      VIR_DOMAIN_AFFECT_CONFIG;
        flags &= allflags;
        if (!(flags == 0 || flags == allflags)) {
            virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                           _("Cannot modify persistent and active configs independently"));
            return -1;
        }
    } else {
        if (flags & VIR_DOMAIN_AFFECT_LIVE) {
            virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                           _("Cannot modify the active config of an inactive domain"));
            return -1;
        }
    }

    return 0;
}


static int
vzctDomainSetMemoryFlags(virDomainPtr domain,
                         unsigned long memory,
                         unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    virObjectEvent *event;
    bool job = false;
    int err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetMemoryFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    /* Ceiling new memory size from Kb to Mb */
    VZCT_CHECK(cleanup, vzctl2_env_set_ramsize(param,
                                               (memory + 1023) / 1024));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (virDomainObjIsActive(dom)) {
        dom->newDef->mem.cur_balloon = memory;
        virDomainDefSetMemoryTotal(dom->newDef, dom->newDef->mem.cur_balloon);
    }

    dom->def->mem.cur_balloon = memory;
    virDomainDefSetMemoryTotal(dom->def, dom->def->mem.cur_balloon);

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetMemory(virDomainPtr dom,
                    unsigned long newmem)
{
    return vzctDomainSetMemoryFlags(dom, newmem, 0);
}


static int
vzctDomainSetVcpusFlags(virDomainPtr domain,
                        unsigned int nvcpus,
                        unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    virObjectEvent *event;
    bool job = false;
    int err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetVcpusFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    VZCT_CHECK(cleanup, vzctl2_env_set_cpu_count(param, nvcpus));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (virDomainObjIsActive(dom))
        dom->newDef->maxvcpus = nvcpus;
    dom->def->maxvcpus = nvcpus;

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetVcpus(virDomainPtr dom,
                   unsigned int nvcpus)
{
    return vzctDomainSetVcpusFlags(dom, nvcpus, 0);
}


static int
vzctDomainGetVcpusFlags(virDomainPtr domain,
                        unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_DOMAIN_VCPU_MAXIMUM, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get simultaneously persistent "
                         "and active configs"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetVcpusFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get active config of inactive domain"));
        goto cleanup;
    }

    if (flags & VIR_DOMAIN_VCPU_MAXIMUM) {
        /* 0 means unlimited */
        ret = 0;
        goto cleanup;
    }

    ret = virDomainDefGetVcpusMax(dom->def);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetMaxVcpus(virDomainPtr dom)
{
    return vzctDomainGetVcpusFlags(dom, VIR_DOMAIN_VCPU_MAXIMUM);
}


static int
vzctDomainSetMemoryParameters(virDomainPtr domain,
                              virTypedParameterPtr params,
                              int nparams,
                              unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    struct vzctl_mem_guarantee mem_guarantee;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    virObjectEvent *event;
    unsigned long long min_guarantee;
    bool set_min_guarantee = false;
    bool set_min_guarantee_auto = false;
    bool job = false;
    int min_guarantee_auto;
    int rc, err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_MEMORY_MIN_GUARANTEE,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_MEMORY_MIN_GUARANTEE_VZ_AUTO,
                               VIR_TYPED_PARAM_BOOLEAN,
                               NULL) < 0)
        return -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetMemoryParametersEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    if ((rc = virTypedParamsGetULLong(params, nparams,
                                      VIR_DOMAIN_MEMORY_MIN_GUARANTEE,
                                      &min_guarantee)) < 0)
        goto cleanup;
    if (rc == 1)
        set_min_guarantee = true;

    if ((rc = virTypedParamsGetBoolean(params, nparams,
                                       VIR_DOMAIN_MEMORY_MIN_GUARANTEE_VZ_AUTO,
                                       &min_guarantee_auto)) < 0)
        goto cleanup;
    if (rc == 1)
        set_min_guarantee_auto = true;

    if (set_min_guarantee && set_min_guarantee_auto) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("min_guarantee and min_guarantee_vz_auto are "
                         "mutually exclusive"));
        goto cleanup;
    }

    if (set_min_guarantee_auto && !min_guarantee_auto) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("Set a min_guarantee to disable "
                         "automatic min_guarantee"));
        goto cleanup;
    }

    memset(&mem_guarantee, 0, sizeof(mem_guarantee));
    if (set_min_guarantee) {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_BYTES;
        mem_guarantee.value = min_guarantee << 10; /* KiB to bytes */
    } else if (set_min_guarantee_auto) {
        mem_guarantee.type = VZCTL_MEM_GUARANTEE_AUTO;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    VZCT_CHECK(cleanup, vzctl2_env_set_memguarantee(param, &mem_guarantee));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (set_min_guarantee) {
        dom->def->mem.min_guarantee = min_guarantee;
        dom->def->mem.min_guarantee_auto = false;
        if (virDomainObjIsActive(dom)) {
            dom->newDef->mem.min_guarantee = min_guarantee;
            dom->newDef->mem.min_guarantee_auto = false;
        }
    } else if (set_min_guarantee_auto) {
        dom->def->mem.min_guarantee = 0;
        dom->def->mem.min_guarantee_auto = true;
        if (virDomainObjIsActive(dom)) {
            dom->newDef->mem.min_guarantee = 0;
            dom->newDef->mem.min_guarantee_auto = true;
        }
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetMemoryParameters(virDomainPtr domain,
                              virTypedParameterPtr params,
                              int *nparams,
                              unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get simultaneously persistent "
                         "and active configs"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetMemoryParametersEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get active config of inactive domain"));
        goto cleanup;
    }

    if (!*nparams) {
        *nparams = 1;
        ret = 0;
        goto cleanup;
    }

    if (!dom->def->mem.min_guarantee_auto) {
        if (virTypedParameterAssign(&params[0], VIR_DOMAIN_MEMORY_MIN_GUARANTEE,
                                    VIR_TYPED_PARAM_ULLONG,
                                    dom->def->mem.min_guarantee) < 0)
            goto cleanup;
    } else {
        if (virTypedParameterAssign(&params[0],
                                    VIR_DOMAIN_MEMORY_MIN_GUARANTEE_VZ_AUTO,
                                    VIR_TYPED_PARAM_BOOLEAN, true) < 0)
            goto cleanup;
    }

    *nparams = 1;
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static void
vzctSetDefBlockIoTune(virDomainDef *def,
                      unsigned long long io,
                      unsigned long long iops,
                      bool set_io,
                      bool set_iops,
                      const char *group_name)
{
    size_t i;

    for (i = 0; i < def->ndisks; i++) {
        virDomainDiskDef *disk = def->disks[i];
        virDomainBlockIoTuneInfo *iotune = &disk->blkdeviotune;

        if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK)
            continue;

        if (set_io)
            iotune->total_bytes_sec = io;
        if (set_iops)
            iotune->total_iops_sec = iops;

        if (group_name) {
            VIR_FREE(iotune->group_name);
            iotune->group_name = g_strdup(group_name);
        } else if (def->ndisks > 1 && !iotune->group_name) {
            iotune->group_name = g_strdup("virtuozzo");
        }
    }
}


static int
vzctDomainSetBlockIoTune(virDomainPtr domain,
                         const char *disk,
                         virTypedParameterPtr params,
                         int nparams,
                         unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    virObjectEvent *event;
    unsigned long long io, iops;
    bool set_io = false;
    bool set_iops = false;
    bool job = false;
    int rc, err, ret = -1;
    const char *group_name = NULL;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_BLOCK_IOTUNE_GROUP_NAME,
                               VIR_TYPED_PARAM_STRING,
                               NULL) < 0)
        return -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetBlockIoTuneEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    if (!virDomainDiskByName(dom->def, disk, false)) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", disk);
        goto cleanup;
    }

    if ((rc = virTypedParamsGetULLong(params, nparams,
                                      VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                                      &io)) < 0)
        goto cleanup;
    if (rc == 1)
        set_io = true;

    if ((rc = virTypedParamsGetULLong(params, nparams,
                                      VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                                      &iops)) < 0)
        goto cleanup;
    if (rc == 1)
        set_iops = true;

    if (!set_io && !set_iops) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("You must set or change at least one limit"));
        goto cleanup;
    }

    if (virTypedParamsGetString(params, nparams,
                                VIR_DOMAIN_BLOCK_IOTUNE_GROUP_NAME,
                                &group_name) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                     0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    if (set_io)
        VZCT_CHECK(cleanup, vzctl2_env_set_iolimit(param, io));
    if (set_iops)
        VZCT_CHECK(cleanup, vzctl2_env_set_iopslimit(param, iops));
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    vzctSetDefBlockIoTune(dom->def, io, iops, set_io, set_iops, group_name);

    if (virDomainObjIsActive(dom))
        vzctSetDefBlockIoTune(dom->newDef, io, iops, set_io, set_iops,
                              group_name);

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(dom, VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_UPDATED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetBlockIoTune(virDomainPtr domain,
                         const char *disk,
                         virTypedParameterPtr params,
                         int *nparams,
                         unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    virDomainDiskDef *def;
    bool job = false;
    int param = 0;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get simultaneously persistent "
                         "and active configs"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetBlockIoTuneEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get active config of inactive domain"));
        goto cleanup;
    }

    if (!*nparams) {
        *nparams = 3;
        ret = 0;
        goto cleanup;
    }

    if (!(def = virDomainDiskByName(dom->def, disk, false))) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", disk);
        goto cleanup;
    }

    if (virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_BYTES_SEC,
                                VIR_TYPED_PARAM_ULLONG,
                                def->blkdeviotune.total_bytes_sec) < 0)
        goto cleanup;

    if (param < *nparams &&
        virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_BLOCK_IOTUNE_TOTAL_IOPS_SEC,
                                VIR_TYPED_PARAM_ULLONG,
                                def->blkdeviotune.total_iops_sec) < 0)
        goto cleanup;

    if (param < *nparams) {
        g_autofree char *group_name = NULL;

        group_name = g_strdup(def->blkdeviotune.group_name);
        if (virTypedParameterAssign(&params[param++],
                                    VIR_DOMAIN_BLOCK_IOTUNE_GROUP_NAME,
                                    VIR_TYPED_PARAM_STRING, group_name) < 0)
            goto cleanup;

        group_name = NULL;
    }

    *nparams = param;
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainBlockResize(virDomainPtr domain,
                      const char *disk,
                      unsigned long long size,
                      unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    vzctl_env_param_ptr param;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    virDomainDiskDef *def;
    bool job = false;
    int layout, err, ret = -1;

    virCheckFlags(VIR_DOMAIN_BLOCK_RESIZE_BYTES, -1);

    if (flags & VIR_DOMAIN_BLOCK_RESIZE_BYTES)
        size = (size + 1023) / 1024; /* bytes to KiB */

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainBlockResizeEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!(def = virDomainDiskByName(dom->def, disk, false))) {
        virReportError(VIR_ERR_DEVICE_MISSING, "%s", disk);
        goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    param = vzctl2_get_env_param(env);
    VZCT_CHECK(cleanup, vzctl2_env_get_layout(param, &layout));

    if (layout == VZCTL_LAYOUT_5) {
        struct vzctl_disk_param disk_param = { .size = size };
        const char *uuid = ((vzctDomainDiskPrivate *)def->privateData)->uuid;

        memcpy(disk_param.uuid, uuid, sizeof(disk_param.uuid));
        VZCT_CHECK(cleanup, vzctl2_env_set_disk(env, &disk_param));

        /* Also update diskspace for root disk image */
        if (STREQ(uuid, VZCT_DISK_ROOT_UUID)) {
            g_autoptr(vzctl_env_param) cparam = NULL;
            struct vzctl_2UL_res val = {size, size};

            VZCT_CHECK_PTR(cleanup, cparam, vzctl2_alloc_env_param());
            VZCT_CHECK(cleanup, vzctl2_env_set_diskspace(cparam, &val));
            VZCT_CHECK(cleanup, vzctl2_apply_param(env, cparam, VZCTL_SAVE));
        }
    } else {
        g_autoptr(vzctl_env_param) cparam = NULL;
        struct vzctl_2UL_res space = {size, size};
        struct vzctl_2UL_res inodes = {size / 4, size / 4};

        VZCT_CHECK_PTR(cleanup, cparam, vzctl2_alloc_env_param());
        VZCT_CHECK(cleanup, vzctl2_env_set_diskspace(cparam, &space));
        VZCT_CHECK(cleanup, vzctl2_env_set_diskinodes(cparam, &inodes));
        VZCT_CHECK(cleanup, vzctl2_apply_param(env, cparam, VZCTL_SAVE));
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetMetadata(virDomainPtr domain,
                      int type,
                      const char *metadata,
                      const char *key,
                      const char *uri,
                      unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virObjectEvent *event;
    virDomainObj *dom;
    g_autofree char *osid = NULL;
    g_autofree char *ostemplate = NULL;
    bool isOsinfo = STREQ(uri, VZCT_LIBOSINFO_URI);
    bool isVhs = STREQ(uri, VZCT_VHS_URI);
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetMetadataEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    if (isOsinfo || isVhs) {
        /* If we resetting vhs metadata then ostemplate is NULL */
        if (!(isVhs && !metadata) &&
            vzctDomainDefGetMetaAttr(dom->def,
                                     VZCT_VHS_URI, VZCT_VHS_XPATH_OSTEMPLATE,
                                     isVhs ? metadata : NULL,
                                     &ostemplate) < 0)
            goto cleanup;

        /* If we resetting libosinfo metadata then osid is NULL */
        if (!(isOsinfo && !metadata) &&
            vzctDomainDefGetMetaAttr(dom->def,
                                     VZCT_LIBOSINFO_URI, VZCT_LIBOSINFO_XPATH_OSID,
                                     isOsinfo ? metadata : NULL,
                                     &osid) < 0)
            goto cleanup;

        if (ostemplate && osid) {
            virReportError(VIR_ERR_INVALID_ARG, "%s",
                           _("cannot set both VHS ostemplate and libosinfo OS ID at the same time"));
            goto cleanup;
        }
    }

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = vzctStateFile(driver->config, dom->def->uuid);

        if (virDomainDefSetMetadata(dom->newDef, type, metadata, key, uri) < 0)
            goto cleanup;

        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    if (virDomainDefSetMetadata(dom->def, type, metadata, key, uri) < 0)
        goto cleanup;

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (isOsinfo || isVhs) {
        g_autoptr(vzctl_env_param) param = NULL;
        g_autoptr(vzctl_env_handle) env = NULL;
        int err;

        VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));
        VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());

        if (vzctConfigSetOsDistribution(param, osid, ostemplate) < 0)
            goto cleanup;

        VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));
    }

    event = virDomainEventMetadataChangeNewFromObj(dom, type, uri);
    virObjectEventStateQueue(driver->eventsManager, event);
    event = NULL;

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);

    virDomainObjEndAPI(&dom);
    return ret;
}


static char *
vzctDomainGetMetadata(virDomainPtr domain,
                      int type,
                      const char *uri,
                      unsigned int flags)
{
    virDomainObj *dom;
    char *ret = NULL;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetMetadataEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = virDomainObjGetMetadata(dom, type, uri, flags);

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainRenameCallback(virDomainObj *dom,
                         const char *new_name,
                         unsigned int fflags G_GNUC_UNUSED,
                         void *opaque)
{
    vzctDriver *driver = opaque;
    g_autoptr(vzctl_env_handle) env = NULL;
    int err;

    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_RET(-1, vzctl2_set_name(env, new_name));

    VIR_FREE(dom->def->name);
    dom->def->name = g_strdup(new_name);

    if (virDomainObjIsActive(dom)) {
        VIR_FREE(dom->newDef->name);
        dom->newDef->name = g_strdup(new_name);
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        VIR_WARN("Cannot save config with new name: %s", new_name);

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            VIR_WARN("Cannot save status with new name: %s", new_name);
    }

    return 0;
}


static int
vzctDomainRename(virDomainPtr domain,
                 const char *new_name,
                 unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virObjectEvent *event_new;
    virObjectEvent *event_old;
    virDomainObj *dom = NULL;
    g_autofree char *old_name = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainRenameEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    old_name = g_strdup(dom->def->name);
    if (virDomainObjListRename(driver->domains, dom, new_name, flags,
                               vzctDomainRenameCallback, driver) < 0)
        goto cleanup;

    event_old = virDomainEventLifecycleNew(dom->def->id, old_name,
                                           dom->def->uuid,
                                           VIR_DOMAIN_EVENT_UNDEFINED,
                                           VIR_DOMAIN_EVENT_UNDEFINED_RENAMED);
    event_new = virDomainEventLifecycleNewFromObj(dom,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              VIR_DOMAIN_EVENT_DEFINED_RENAMED);
    virObjectEventStateQueue(driver->eventsManager, event_old);
    virObjectEventStateQueue(driver->eventsManager, event_new);
    event_old = NULL;
    event_new = NULL;

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static char *
vzctConnectGetHostname(virConnectPtr conn)
{
    if (virConnectGetHostnameEnsureACL(conn) < 0)
        return NULL;

    return virGetHostname();
}


static int
vzctConnectGetVersion(virConnectPtr conn,
                      unsigned long *version)
{
    const char *ver;

    if (virConnectGetVersionEnsureACL(conn) < 0)
        return -1;

    ver = vzctl2_get_version();
    if (virParseVersionString(ver, version, true) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, _("Unknown release '%s'"), ver);
        return -1;
    }

    return 0;
}


static int
vzctNodeGetInfo(virConnectPtr conn,
                virNodeInfoPtr nodeinfo)
{
    if (virNodeGetInfoEnsureACL(conn) < 0)
        return -1;

    return virCapabilitiesGetNodeInfo(nodeinfo);
}


static int
vzctNodeGetCPUMap(virConnectPtr conn,
                  unsigned char **cpumap,
                  unsigned int *online,
                  unsigned int flags)
{
    if (virNodeGetCPUMapEnsureACL(conn) < 0)
        return -1;

    return virHostCPUGetMap(cpumap, online, flags);
}


static int
vzctDomainManagedSave(virDomainPtr domain,
                      unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autofree char *xmlfile = NULL;
    struct vzctl_cpt_param param;
    virDomainObj *dom = NULL;
    virObjectEvent *event;
    bool job = false;
    bool saved = false;
    int err;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainManagedSaveEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjCheckActive(dom) < 0)
        goto cleanup;

    xmlfile = vzctManagedSaveXMLFile(dom);
    if (virDomainDefSavePath(dom->def, driver->xmlopt, xmlfile,
                             VIR_DOMAIN_DEF_FORMAT_SECURE |
                             VIR_DOMAIN_DEF_FORMAT_MIGRATABLE |
                             VIR_DOMAIN_DEF_FORMAT_INACTIVE) < 0)
        goto cleanup;
    saved = true;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));

    memset(&param, 0, sizeof(param));
    param.cmd = VZCTL_CMD_CHKPNT;
    VZCT_CHECK(cleanup, vzctl2_env_chkpnt(env, VZCTL_CMD_CHKPNT, &param, 0));

    vzctDomainStop(driver, dom);

    event = virDomainEventLifecycleNewFromObj(dom,
                                     VIR_DOMAIN_EVENT_STOPPED,
                                     VIR_DOMAIN_EVENT_STOPPED_SAVED);
    virObjectEventStateQueue(driver->eventsManager, event);

    ret = 0;

 cleanup:
    if (ret < 0 && saved && unlink(xmlfile) < 0)
        VIR_WARN("cannot unlink managed save xml file, name: %s, errno: %d",
                 dom->def->name, errno);

    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainHasManagedSaveImage(virDomainPtr domain,
                              unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    vzctl_env_status_t status;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainHasManagedSaveImageEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_SUSPENDED));

    ret = !!(status.mask & ENV_STATUS_SUSPENDED);

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainManagedSaveRemove(virDomainPtr domain,
                            unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autofree char *xmlfile = NULL;
    virDomainObj *dom = NULL;
    vzctl_env_status_t status;
    bool job = false;
    int err;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainManagedSaveRemoveEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_SUSPENDED));

    if (!(status.mask & ENV_STATUS_SUSPENDED)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("domain does not have managed save image"));
        goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK(cleanup, vzctl2_env_remove_dump(env));

    xmlfile = vzctManagedSaveXMLFile(dom);
    if (unlink(xmlfile) < 0)
        VIR_WARN("cannot unlink managed save, name: %s, errno: %d",
                 dom->def->name, errno);

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static char *
vzctDomainManagedSaveGetXMLDesc(virDomainPtr domain,
                                unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    g_autoptr(virDomainDef) def = NULL;
    g_autofree char *xmlfile = NULL;
    virDomainObj *dom = NULL;
    vzctl_env_status_t status;
    char *ret = NULL;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_SAVE_IMAGE_XML_SECURE, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainManagedSaveGetXMLDescEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_SUSPENDED));

    if (!(status.mask & ENV_STATUS_SUSPENDED)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("domain does not have managed save image"));
        goto cleanup;
    }

    xmlfile = vzctManagedSaveXMLFile(dom);
    if (!(def = virDomainDefParseFile(xmlfile, driver->xmlopt, NULL,
                                      VIR_DOMAIN_DEF_PARSE_INACTIVE |
                                      VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE |
                                      VIR_DOMAIN_DEF_PARSE_ALLOW_POST_PARSE_FAIL)))
        goto cleanup;

    ret = virDomainDefFormat(def, driver->xmlopt, flags);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static int
vzctDomainManagedSaveDefineXML(virDomainPtr domain,
                               const char *dxml,
                               unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    g_autoptr(virDomainDef) def = NULL;
    g_autofree char *xmlfile = NULL;
    virDomainObj *dom = NULL;
    vzctl_env_status_t status;
    bool job = false;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainManagedSaveDefineXMLEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_get_env_status(dom->def->extraId,
                                              &status, ENV_STATUS_SUSPENDED));

    if (!(status.mask & ENV_STATUS_SUSPENDED)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("domain does not have managed save image"));
        goto cleanup;
    }

    if (!(def = virDomainDefParseString(dxml, driver->xmlopt, NULL,
                                        VIR_DOMAIN_DEF_PARSE_INACTIVE)))
        goto cleanup;

    xmlfile = vzctManagedSaveXMLFile(dom);
    if (virDomainDefSavePath(def, driver->xmlopt, xmlfile,
                             VIR_DOMAIN_DEF_FORMAT_SECURE |
                             VIR_DOMAIN_DEF_FORMAT_MIGRATABLE |
                             VIR_DOMAIN_DEF_FORMAT_INACTIVE) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);
    return ret;
}


static char *
vzctDomainGetSchedulerType(virDomainPtr domain,
                           int *nparams)
{
    char *ret = NULL;
    virDomainObj *dom = NULL;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetSchedulerTypeEnsureACL(domain->conn, dom->def) >= 0) {
        ret = g_strdup("posix");
        if (nparams)
            *nparams = 3; /* shares, period, quota */
    }

    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetSchedulerParametersFlags(virDomainPtr domain,
                                      virTypedParameterPtr params,
                                      int *nparams,
                                      unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    bool job = false;
    int param = 0;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get persistent and active configs "
                         "at the same time"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetSchedulerParametersFlagsEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get the active config of an inactive domain"));
        goto cleanup;
    }

    if (!*nparams) {
        *nparams = 3;
        ret = 0;
        goto cleanup;
    }

    if (virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_SCHEDULER_CPU_SHARES,
                                VIR_TYPED_PARAM_ULLONG,
                                dom->def->cputune.shares) < 0)
        goto cleanup;

    if (param < *nparams &&
        virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_SCHEDULER_VCPU_PERIOD,
                                VIR_TYPED_PARAM_ULLONG,
                                dom->def->cputune.period) < 0)
        goto cleanup;

    if (param < *nparams &&
        virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_SCHEDULER_VCPU_QUOTA,
                                VIR_TYPED_PARAM_LLONG,
                                dom->def->cputune.quota) < 0)
        goto cleanup;

    *nparams = param;
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainGetSchedulerParameters(virDomainPtr domain,
                                 virTypedParameterPtr params,
                                 int *nparams)
{
    return vzctDomainGetSchedulerParametersFlags(domain, params, nparams,
                                                 VIR_DOMAIN_AFFECT_CURRENT);
}


static int
vzctDomainSetSchedulerParametersFlags(virDomainPtr domain,
                                      virTypedParameterPtr params,
                                      int nparams,
                                      unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    virTypedParameterPtr eventParams = NULL;
    unsigned long long shares = 0;
    unsigned long long period;
    long long quota;
    bool job = false;
    bool set_shares = false;
    bool set_limit = false;
    int eventNparams = 0;
    int eventMaxNparams = 0;
    int err, ret = -1;
    size_t i;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_SCHEDULER_CPU_SHARES,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_SCHEDULER_VCPU_PERIOD,
                               VIR_TYPED_PARAM_ULLONG,
                               VIR_DOMAIN_SCHEDULER_VCPU_QUOTA,
                               VIR_TYPED_PARAM_LLONG,
                               NULL) < 0)
        return -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetSchedulerParametersFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    quota = dom->def->cputune.quota;
    period = dom->def->cputune.period;

    for (i = 0; i < nparams; i++) {
        if (STREQ(params[i].field, VIR_DOMAIN_SCHEDULER_CPU_SHARES)) {
            shares = params[i].value.ul;

            if (virTypedParamsAddULLong(&eventParams, &eventNparams,
                                        &eventMaxNparams,
                                        VIR_DOMAIN_TUNABLE_CPU_CPU_SHARES,
                                        shares) < 0)
                goto cleanup;

            set_shares = true;
        } else if (STREQ(params[i].field, VIR_DOMAIN_SCHEDULER_VCPU_PERIOD)) {
            period = params[i].value.ul;

            if (period < 1000 || period > 1000000) {
                virReportError(VIR_ERR_INVALID_ARG, "%s",
                               _("The value of 'period' must be in the range "
                                 "[1000, 1000000]"));
                goto cleanup;
            }

            if (virTypedParamsAddULLong(&eventParams, &eventNparams,
                                        &eventMaxNparams,
                                        VIR_DOMAIN_TUNABLE_CPU_VCPU_PERIOD,
                                        period) < 0)
                goto cleanup;

            set_limit = true;
        } else if (STREQ(params[i].field, VIR_DOMAIN_SCHEDULER_VCPU_QUOTA)) {
            quota = params[i].value.l;

            if (quota > 0 && (quota < 1000 || quota > 18446744073709551)) {
                virReportError(VIR_ERR_INVALID_ARG, "%s",
                               _("The value of 'quota' must be in the range "
                                 "[1000, 18446744073709551]"));
                goto cleanup;
            }

            if (virTypedParamsAddLLong(&eventParams, &eventNparams,
                                       &eventMaxNparams,
                                       VIR_DOMAIN_TUNABLE_CPU_VCPU_QUOTA,
                                       quota) < 0)
                goto cleanup;

            set_limit = true;
        }
    }

    if (set_limit && quota > 0 && !period) {
        virReportError(VIR_ERR_INVALID_ARG, "%s",
                       _("If 'quota' is positive, 'period' must be "
                         "set as well"));
        goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    if (set_limit) {
        struct vzctl_cpulimit_param cpu_limit;

        cpu_limit.type = VZCTL_CPULIMIT_PCT;
        cpu_limit.limit = quota > 0 ? ((quota * 100) / period) : 0;
        VZCT_CHECK(cleanup, vzctl2_env_set_cpulimit(param, &cpu_limit));
    }
    if (set_shares) {
        if (!shares)
            shares = VZCTL_CPUUNITS_DEFAULT;
        VZCT_CHECK(cleanup, vzctl2_env_set_cpuunits(param, shares));
    }
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (virDomainObjIsActive(dom)) {
        dom->newDef->cputune.quota = quota;
        dom->newDef->cputune.period = period;
        if (set_shares) {
            dom->newDef->cputune.shares = shares;
            dom->newDef->cputune.sharesSpecified = true;
        }
    }
    dom->def->cputune.quota = quota;
    dom->def->cputune.period = period;
    if (set_shares) {
        dom->def->cputune.shares = shares;
        dom->def->cputune.sharesSpecified = true;
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    if (eventNparams) {
        virObjectEvent *event;

        event = virDomainEventTunableNewFromDom(domain, eventParams,
                                                eventNparams);
        virObjectEventStateQueue(driver->eventsManager, event);
        eventNparams = 0;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    if (eventNparams)
        virTypedParamsFree(eventParams, eventNparams);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetSchedulerParameters(virDomainPtr domain,
                                 virTypedParameterPtr params,
                                 int nparams)
{
    return vzctDomainSetSchedulerParametersFlags(domain, params, nparams,
                                                 VIR_DOMAIN_AFFECT_CURRENT);
}


static int
vzctDomainGetBlkioParameters(virDomainPtr domain,
                             virTypedParameterPtr params,
                             int *nparams,
                             unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    bool job = false;
    int param = 0;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_TYPED_PARAM_STRING_OKAY, -1);

    if ((flags & VIR_DOMAIN_AFFECT_LIVE) &&
        (flags & VIR_DOMAIN_AFFECT_CONFIG)) {
        virReportError(VIR_ERR_ARGUMENT_UNSUPPORTED, "%s",
                       _("Cannot get persistent and active configs "
                         "at the same time"));
        return -1;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetBlkioParametersEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(dom) && (flags & VIR_DOMAIN_AFFECT_LIVE)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("Cannot get the active config of an inactive domain"));
        goto cleanup;
    }

    if (!*nparams) {
        *nparams = 1;
        ret = 0;
        goto cleanup;
    }

    if (virTypedParameterAssign(&params[param++],
                                VIR_DOMAIN_BLKIO_WEIGHT,
                                VIR_TYPED_PARAM_UINT,
                                dom->def->blkio.weight) < 0)
        goto cleanup;

    *nparams = param;
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetBlkioParameters(virDomainPtr domain,
                             virTypedParameterPtr params,
                             int nparams,
                             unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    bool set_weight = false;
    bool job = false;
    unsigned int weight;
    int err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_BLKIO_WEIGHT,
                               VIR_TYPED_PARAM_UINT,
                               NULL) < 0)
        return -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetBlkioParametersEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    if ((err = virTypedParamsGetUInt(params, nparams,
                                     VIR_DOMAIN_BLKIO_WEIGHT,
                                     &weight)) < 0)
        goto cleanup;
    if (err == 1)
        set_weight = true;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    if (set_weight) {
        int ioprio = vzctConvertWeightToIOPrio(weight);
        VZCT_CHECK(cleanup, vzctl2_env_set_ioprio(param, ioprio));
        weight = vzctConvertIOPrioToWeight(ioprio);
    }
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (set_weight) {
        dom->def->blkio.weight = weight;
        if (virDomainObjIsActive(dom))
            dom->newDef->blkio.weight = weight;
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetGuestNetworkParametersVz(virDomainPtr domain,
                                      virTypedParameterPtr params,
                                      int nparams,
                                      unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    const char *hostname = NULL;
    g_autofree const char **servers = NULL;
    g_autofree const char **searchs = NULL;
    g_autofree virSocketAddr *addrs = NULL;
    bool set_servers = false;
    bool set_searchs = false;
    bool job = false;
    int nservers = 0;
    int nsearchs = 0;
    int err, ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (virTypedParamsValidate(params, nparams,
                               VIR_DOMAIN_GUEST_NET_HOSTNAME,
                               VIR_TYPED_PARAM_STRING,
                               VIR_DOMAIN_GUEST_NET_DNS_SERVER,
                               VIR_TYPED_PARAM_STRING |
                               VIR_TYPED_PARAM_MULTIPLE,
                               VIR_DOMAIN_GUEST_NET_SEARCH_DOMAIN,
                               VIR_TYPED_PARAM_STRING |
                               VIR_TYPED_PARAM_MULTIPLE,
                               NULL) < 0)
        return -1;

    if (virTypedParamsGetString(params, nparams,
                                VIR_DOMAIN_GUEST_NET_HOSTNAME,
                                &hostname) < 0)
        return -1;

    nservers = virTypedParamsGetStringList(params, nparams,
                                           VIR_DOMAIN_GUEST_NET_DNS_SERVER,
                                           &servers);
    if (nservers < 0)
        return -1;

    if (servers) {
        size_t i;

        /* List with first empty item have to remove all items */
        if (virStringIsEmpty(servers[0])) {
            VIR_FREE(servers);
            nservers = 0;
        } else {
            addrs = g_new0(virSocketAddr, nservers);

            for (i = 0; i < nservers; i++) {
                if (virSocketAddrParse(&addrs[i], servers[i], AF_UNSPEC) < 0) {
                    virReportError(VIR_ERR_INVALID_ARG,
                                   _("The DNS server address '%s' is invalid"),
                                   servers[i]);
                    return -1;
                }
            }
        }

        set_servers = true;
    }

    nsearchs = virTypedParamsGetStringList(params, nparams,
                                           VIR_DOMAIN_GUEST_NET_SEARCH_DOMAIN,
                                           &searchs);
    if (nsearchs < 0)
        return -1;

    if (searchs) {
        /* List with first empty item have to remove all items */
        if (virStringIsEmpty(searchs[0])) {
            VIR_FREE(searchs);
            nsearchs = 0;
        } else {
            size_t n = nsearchs;

            VIR_EXPAND_N(searchs, n, 1); /* compatibility with string list */
        }

        set_searchs = true;
    }

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetGuestNetworkParametersVzEnsureACL(domain->conn, dom->def,
                                                      flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId,
                                                 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    if (hostname)
        VZCT_CHECK(cleanup, vzctl2_env_set_hostname(param, hostname));
    if (set_servers) {
        if (nservers) {
            size_t i;

            for (i = 0; i < nservers; i++)
                VZCT_CHECK(cleanup, vzctl2_env_add_nameserver(param,
                                                              servers[i]));
        } else {
            VZCT_CHECK(cleanup, vzctl2_env_add_nameserver(param, ""));
        }
    }
    if (set_searchs) {
        if (nsearchs) {
            size_t i;

            for (i = 0; i < nsearchs; i++)
                VZCT_CHECK(cleanup, vzctl2_env_add_searchdomain(param,
                                                                searchs[i]));
        } else {
            VZCT_CHECK(cleanup, vzctl2_env_add_searchdomain(param, ""));
        }
    }
    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (hostname) {
        VIR_FREE(dom->def->dns.hostname);
        dom->def->dns.hostname = g_strdup(hostname);

        if (virDomainObjIsActive(dom)) {
            VIR_FREE(dom->newDef->dns.hostname);
            dom->newDef->dns.hostname = g_strdup(hostname);
        }
    }

    if (set_servers) {
        VIR_FREE(dom->def->dns.servers);
        dom->def->dns.servers = g_new0(virSocketAddr, nservers);
        memcpy(dom->def->dns.servers, addrs, sizeof(*addrs) * nservers);
        dom->def->dns.nservers = nservers;

        if (virDomainObjIsActive(dom)) {
            VIR_FREE(dom->newDef->dns.servers);
            dom->newDef->dns.servers = g_new0(virSocketAddr, nservers);
            memcpy(dom->newDef->dns.servers, addrs, sizeof(*addrs) * nservers);
            dom->newDef->dns.nservers = nservers;
        }
    }

    if (set_searchs) {
        g_strfreev(dom->def->dns.searchs);
        dom->def->dns.searchs = g_strdupv((char **)searchs);

        if (virDomainObjIsActive(dom)) {
            g_strfreev(dom->newDef->dns.searchs);
            dom->newDef->dns.searchs = g_strdupv((char **)searchs);
        }
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static void
vzctCopyDiskUUID(virDomainDiskDef *dst,
                 const virDomainDiskDef *src)
{
    const vzctDomainDiskPrivate* srcPriv =
        (const vzctDomainDiskPrivate*) src->privateData;
    vzctDomainDiskPrivate* dstPriv =
        (vzctDomainDiskPrivate*) dst->privateData;

    memcpy(dstPriv->uuid, srcPriv->uuid, sizeof(dstPriv->uuid));
}


static int
vzctDomainAttachDisk(vzctDriver *driver,
                     virDomainObj *dom,
                     virDomainDeviceDef *dev,
                     virDomainDeviceDef *devCopy)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    virDomainDiskDef *disk = dev->data.disk;
    virDomainDiskDef *diskCopy;
    int diskIdx = -1;
    int diskCopyIdx = -1;
    int err;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("can only attach disks of device type 'disk'"));
        return -1;
    }

    vzctGenerateDiskUUID(disk);

    diskIdx = virDomainDiskInsert(dom->def, disk);
    if (virDomainDefValidate(dom->def, 0, driver->xmlopt, NULL) < 0)
        goto error;

    if (virDomainObjIsActive(dom)) {
        diskCopy = devCopy->data.disk;
        vzctCopyDiskUUID(diskCopy, disk);

        diskCopyIdx = virDomainDiskInsert(dom->newDef, diskCopy);
        if (virDomainDefValidate(dom->newDef, 0, driver->xmlopt, NULL) < 0)
            goto error;
    }

    VZCT_CHECK_PTR(error, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (vzctConfigDiskAdd(env, disk) < 0)
        goto error;

    dev->data.disk = NULL;
    if (virDomainObjIsActive(dom)) {
        virObjectEvent *event;
        char *name;

        if (disk->info.alias)
            name = g_strdup(disk->info.alias);
        else
            name = g_strdup(disk->dst);
        event = virDomainEventDeviceAddedNewFromObj(dom, name);
        virObjectEventStateQueue(driver->eventsManager, event);

        devCopy->data.disk = NULL;
    }

    if (dom->def->ndisks == 1 &&
        virDomainBlockIoTuneInfoHasAny(&disk->blkdeviotune)) {

        g_autoptr(vzctl_env_param) param = NULL;
        const virDomainBlockIoTuneInfo *iotune = &disk->blkdeviotune;

        VZCT_CHECK_PTR_RET(-1, param, vzctl2_alloc_env_param());
        VZCT_CHECK_RET(-1, vzctl2_env_set_iolimit(param, iotune->total_bytes_sec));
        VZCT_CHECK_RET(-1, vzctl2_env_set_iopslimit(param, iotune->total_iops_sec));
        VZCT_CHECK_RET(-1, vzctl2_apply_param(env, param, VZCTL_SAVE));
    }

    return 0;

 error:
    if (diskIdx >= 0)
        virDomainDiskRemove(dom->def, diskIdx);
    if (diskCopyIdx >= 0)
        virDomainDiskRemove(dom->newDef, diskCopyIdx);

    return -1;
}


/*
 * Load autogenerated network fields from libvzctl
 * to libvirt configurations.
 */
static void
vzctDomainFillNetName(virDomainObj *dom,
                      vzctl_env_handle *env,
                      virDomainNetDef *net,
                      virDomainNetDef *netCopy)
{
    vzctl_env_param_ptr p = vzctl2_get_env_param(env);
    vzctl_veth_dev_iterator it = NULL;

    while ((it = vzctl2_env_get_veth(p, it))) {
        struct vzctl_veth_dev_param net_param;

        vzctl2_env_get_veth_param(it, &net_param, sizeof(net_param));
        if (STREQ_NULLABLE(net_param.dev_name_ve, net->ifname_guest)) {
            net->ifname = g_strdup(net_param.dev_name);
            if (virDomainObjIsActive(dom))
                netCopy->ifname = g_strdup(net_param.dev_name);
            break;
        }
    }

    if (!it)
        VIR_WARN("cannot load autogenerated iface values, net='%s'",
                 net->ifname_guest);
}


static int
vzctDomainAttachNet(vzctDriver *driver,
                    virDomainObj *dom,
                    virDomainDeviceDef *dev,
                    virDomainDeviceDef *devCopy)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    virDomainNetDef *net = dev->data.net;
    virDomainNetDef *netCopy = NULL;
    bool added = false;
    bool copyAdded = false;
    int err;

    virDomainNetInsert(dom->def, net);
    added = true;
    if (virDomainDefValidate(dom->def, 0, driver->xmlopt, NULL) < 0)
        goto error;

    if (virDomainObjIsActive(dom)) {
        netCopy = devCopy->data.net;

        virDomainNetInsert(dom->newDef, netCopy);
        copyAdded = true;
        if (virDomainDefValidate(dom->newDef, 0, driver->xmlopt, NULL) < 0)
            goto error;
    }

    VZCT_CHECK_PTR(error, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_PTR(error, param, vzctl2_alloc_env_param());
    if (vzctConfigNetAdd(param, net, NULL) < 0)
        goto error;
    VZCT_CHECK(error, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (!net->ifname)
        vzctDomainFillNetName(dom, env, net, netCopy);

    dev->data.net = NULL;
    if (virDomainObjIsActive(dom)) {
        virObjectEvent *event;
        char *name;

        if (net->info.alias)
            name = g_strdup(net->info.alias);
        else
            name = g_strdup(net->ifname);
        event = virDomainEventDeviceAddedNewFromObj(dom, name);
        virObjectEventStateQueue(driver->eventsManager, event);

        devCopy->data.net = NULL;
    }

    return 0;

 error:
    if (added)
        VIR_DELETE_ELEMENT(dom->def->nets, dom->def->nnets - 1,
                           dom->def->nnets);
    if (copyAdded)
        VIR_DELETE_ELEMENT(dom->newDef->nets, dom->newDef->nnets - 1,
                           dom->newDef->nnets);
    return -1;
}


static int
vzctDomainAttachDeviceFlags(virDomainPtr domain,
                            const char *xml,
                            unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    virDomainObj *dom = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                               VIR_DOMAIN_DEF_PARSE_ABI_UPDATE;
    g_autoptr(virDomainDeviceDef) dev = NULL;
    g_autoptr(virDomainDeviceDef) devCopy = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainAttachDeviceFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    if (!(dev = virDomainDeviceDefParse(xml, dom->def, driver->xmlopt, NULL,
                                        parse_flags)))
        goto cleanup;

    if (virDomainObjIsActive(dom) &&
        !(devCopy = virDomainDeviceDefParse(xml, dom->def, driver->xmlopt, NULL,
                                            parse_flags)))
        goto cleanup;

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        if (vzctDomainAttachDisk(driver, dom, dev, devCopy) < 0)
            goto cleanup;
        break;

    case VIR_DOMAIN_DEVICE_NET:
        if (vzctDomainAttachNet(driver, dom, dev, devCopy) < 0)
            goto cleanup;
        break;

    default:
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("attaching this type of device is not supported"));
        goto cleanup;
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainAttachDevice(virDomainPtr domain,
                       const char *xml)
{
    return vzctDomainAttachDeviceFlags(domain, xml, 0);
}


static int
vzctDomainDetachDisk(vzctDriver *driver,
                     virDomainObj *dom,
                     virDomainDiskDef *disk)
{
    virDomainDiskDef *defDisk;
    vzctDomainDiskPrivate *diskPriv;
    g_autoptr(vzctl_env_handle) env = NULL;
    int idx;
    int newIdx = -1;
    int err;

    if (disk->device != VIR_DOMAIN_DISK_DEVICE_DISK) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("can only detach disks of device type 'disk'"));
        return -1;
    }

    if ((idx = virDomainDiskIndexByName(dom->def, disk->dst, false)) < 0 ||
        (virDomainObjIsActive(dom) &&
         (newIdx = virDomainDiskIndexByName(dom->newDef, disk->dst, false)) < 0)) {
        virReportError(VIR_ERR_DEVICE_MISSING,
                       _("no target device %s"), disk->dst);
        return -1;
    }

    defDisk = dom->def->disks[idx];
    diskPriv = (vzctDomainDiskPrivate *) defDisk->privateData;
    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_RET(-1, vzctl2_env_detach_disk(env, diskPriv->uuid));

    virDomainDiskDefFree(virDomainDiskRemove(dom->def, idx));
    if (virDomainObjIsActive(dom)) {
        virDomainDiskDef *newDefDisk = dom->newDef->disks[newIdx];
        virObjectEvent *event;
        char *name;

        if (newDefDisk->info.alias)
            name = g_strdup(newDefDisk->info.alias);
        else
            name = g_strdup(newDefDisk->dst);

        event = virDomainEventDeviceRemovedNewFromObj(dom, name);
        virObjectEventStateQueue(driver->eventsManager, event);

        virDomainDiskDefFree(virDomainDiskRemove(dom->newDef, newIdx));
    }

    return 0;
}


static int
vzctDomainDetachNet(vzctDriver *driver,
                    virDomainObj *dom,
                    virDomainNetDef *net)
{
    g_autoptr(vzctl_env_param) param = NULL;
    g_autoptr(vzctl_env_handle) env = NULL;
    int idx;
    int newIdx = -1;
    int err;

    if (net->type == VIR_DOMAIN_NET_TYPE_VENET0) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Detaching venet0 is not supported"));
        return -1;
    }

    if ((idx = virDomainNetFindIdx(dom->def, net)) < 0 ||
        (virDomainObjIsActive(dom) &&
         (newIdx = virDomainNetFindIdx(dom->newDef, net)) < 0))
        return -1;

    VZCT_CHECK_PTR_RET(-1, param, vzctl2_alloc_env_param());
    VZCT_CHECK_PTR_RET(-1, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_RET(-1, vzctl2_env_del_veth(param, dom->def->nets[idx]->ifname_guest));
    VZCT_CHECK_RET(-1, vzctl2_apply_param(env, param, VZCTL_SAVE));

    virDomainNetDefFree(virDomainNetRemove(dom->def, idx));
    if (virDomainObjIsActive(dom)) {
        const virDomainNetDef *newDefNet = dom->newDef->nets[newIdx];
        virObjectEvent *event;
        char *name;

        if (newDefNet->info.alias)
            name = g_strdup(newDefNet->info.alias);
        else
            name = g_strdup(newDefNet->ifname);

        event = virDomainEventDeviceRemovedNewFromObj(dom, name);
        virObjectEventStateQueue(driver->eventsManager, event);

        virDomainNetDefFree(virDomainNetRemove(dom->newDef, newIdx));
    }

    return 0;
}


static int
vzctDomainDetachDeviceFlags(virDomainPtr domain,
                            const char *xml,
                            unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    g_autoptr(virDomainDeviceDef) dev = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE;
    virDomainObj *dom = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainDetachDeviceFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    dev = virDomainDeviceDefParse(xml, dom->def,
                                  driver->xmlopt, NULL, parse_flags);

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        if (vzctDomainDetachDisk(driver, dom, dev->data.disk) < 0)
            goto cleanup;
        break;

    case VIR_DOMAIN_DEVICE_NET:
        if (vzctDomainDetachNet(driver, dom, dev->data.net) < 0)
            goto cleanup;
        break;

    default:
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("detaching this type of device is not supported"));
        goto cleanup;
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainDetachDevice(virDomainPtr domain,
                       const char *xml)
{
    return vzctDomainDetachDeviceFlags(domain, xml, 0);
}


static int
vzctDomainUpdateDisk(vzctDriver *driver,
                     virDomainObj *dom,
                     virDomainDeviceDef *dev,
                     virDomainDeviceDef *devCopy)
{
    int ret = -1;
    g_autoptr(vzctl_env_handle) env = NULL;
    virDomainDiskDef *disk = NULL;
    virDomainDiskDef *diskCopy = NULL;
    int idx = -1;
    int newIdx = -1;
    int err;

    if ((idx = virDomainDiskIndexByName(dom->def, dev->data.disk->dst,
                                        false)) < 0 ||
        (virDomainObjIsActive(dom) &&
         (newIdx = virDomainDiskIndexByName(dom->newDef,
                                            devCopy->data.disk->dst,
                                            false)) < 0)) {
        virReportError(VIR_ERR_DEVICE_MISSING,
                       _("Unable to find the disk '%s' to update"),
                       disk->dst);
        return -1;
    }

    disk = dom->def->disks[idx];
    dom->def->disks[idx] = g_steal_pointer(&dev->data.disk);

    if (virDomainDefValidate(dom->def, 0, driver->xmlopt, NULL) < 0)
        goto cleanup;

    if (STRNEQ_NULLABLE(disk->src->path, dom->def->disks[idx]->src->path)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Cannot update the disk path"));
        goto cleanup;
    }

    if (virDomainObjIsActive(dom)) {
        diskCopy = dom->newDef->disks[newIdx];
        dom->newDef->disks[newIdx] = g_steal_pointer(&devCopy->data.disk);

        if (virDomainDefValidate(dom->newDef, 0, driver->xmlopt, NULL) < 0)
            goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));

    if (vzctConfigDiskUpdate(env, disk, dom->def->disks[idx]) < 0)
        goto cleanup;

    if (dom->def->ndisks == 1) {
        g_autoptr(vzctl_env_param) param = NULL;
        const virDomainBlockIoTuneInfo *iotune = &dom->def->disks[idx]->blkdeviotune;

        VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
        VZCT_CHECK(cleanup, vzctl2_env_set_iolimit(param, iotune->total_bytes_sec));
        VZCT_CHECK(cleanup, vzctl2_env_set_iopslimit(param, iotune->total_iops_sec));
        VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));
    }

    virDomainDiskDefFree(disk);
    disk = NULL;

    virDomainDiskDefFree(diskCopy);
    diskCopy = NULL;

    ret = 0;

 cleanup:
    if (disk) {
        dev->data.disk = dom->def->disks[idx];
        dom->def->disks[idx] = disk;
    }

    if (diskCopy) {
        devCopy->data.disk = dom->newDef->disks[newIdx];
        dom->newDef->disks[newIdx] = diskCopy;
    }

    return ret;
}


static int
vzctChangeNet(vzctl_env_param_ptr param,
              vzctl_env_param_ptr oldparam,
              virDomainNetDef *net)
{
    vzctl_veth_dev_iterator it = NULL;
    struct vzctl_veth_dev_param oldnet_param;

    if (net->type == VIR_DOMAIN_NET_TYPE_VENET0) {
        if (vzctConfigUpdateVenet0(param, net) < 0)
            return -1;

        return 0;
    }

    while ((it = vzctl2_env_get_veth(oldparam, it))) {
        vzctl2_env_get_veth_param(it, &oldnet_param, sizeof(oldnet_param));
        if (STREQ_NULLABLE(net->ifname, oldnet_param.dev_name))
            break;
    }

    if (vzctConfigNetAdd(param, net, it ? &oldnet_param : NULL) < 0)
        return -1;

    return 0;
}


static int
vzctDomainUpdateNet(vzctDriver *driver,
                    virDomainObj *dom,
                    virDomainDeviceDef *dev,
                    virDomainDeviceDef *devCopy)
{
    int ret = -1;
    g_autoptr(vzctl_env_handle) env = NULL;
    g_autoptr(vzctl_env_param) param = NULL;
    vzctl_env_param_ptr oldparam;
    virDomainNetDef *net = NULL;
    virDomainNetDef *netCopy = NULL;
    int idx = -1;
    int newIdx = -1;
    int err;

    if ((idx = virDomainNetFindIdx(dom->def, dev->data.net)) < 0 ||
        (virDomainObjIsActive(dom) &&
         (newIdx = virDomainNetFindIdx(dom->newDef, devCopy->data.net)) < 0))
        return -1;

    net = dom->def->nets[idx];
    dom->def->nets[idx] = g_steal_pointer(&dev->data.net);

    if (virDomainDefValidate(dom->def, 0, driver->xmlopt, NULL) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        netCopy = dom->newDef->nets[newIdx];
        dom->newDef->nets[newIdx] = g_steal_pointer(&devCopy->data.net);

        if (virDomainDefValidate(dom->newDef, 0, driver->xmlopt, NULL) < 0)
            goto cleanup;
    }

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK_PTR(cleanup, param, vzctl2_alloc_env_param());
    oldparam = vzctl2_get_env_param(env);

    if (vzctChangeNet(param, oldparam, dom->def->nets[idx]) < 0)
        goto cleanup;

    VZCT_CHECK(cleanup, vzctl2_apply_param(env, param, VZCTL_SAVE));

    if (!dom->def->nets[idx]->ifname)
        vzctDomainFillNetName(dom, env, dom->def->nets[idx],
                              dom->newDef->nets[newIdx]);

    virDomainNetDefFree(net);
    net = NULL;

    virDomainNetDefFree(netCopy);
    netCopy = NULL;

    ret = 0;

 cleanup:
    if (net) {
        dev->data.net = dom->def->nets[idx];
        dom->def->nets[idx] = net;
    }

    if (netCopy) {
        devCopy->data.net = dom->newDef->nets[newIdx];
        dom->newDef->nets[newIdx] = netCopy;
    }

    return ret;
}


static int
vzctDomainUpdateDeviceFlags(virDomainPtr domain,
                            const char *xml,
                            unsigned int flags)
{
    vzctDriver *driver = domain->conn->privateData;
    g_autoptr(virDomainDeviceDef) dev = NULL;
    g_autoptr(virDomainDeviceDef) devCopy = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE;
    virDomainObj *dom = NULL;
    bool job = false;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_DOMAIN_DEVICE_MODIFY_FORCE, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainUpdateDeviceFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (vzctCheckSync(driver, dom) < 0)
        goto cleanup;

    if (vzctDomainCheckModificationFlags(dom, flags) < 0)
        goto cleanup;

    if (!(dev = virDomainDeviceDefParse(xml, dom->def, driver->xmlopt, NULL,
                                        parse_flags)))
        goto cleanup;

    if (virDomainObjIsActive(dom) &&
        !(devCopy = virDomainDeviceDefParse(xml, dom->def, driver->xmlopt, NULL,
                                            parse_flags)))
        goto cleanup;

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        if (vzctDomainUpdateDisk(driver, dom, dev, devCopy) < 0)
            goto cleanup;
        break;

    case VIR_DOMAIN_DEVICE_NET:
        if (vzctDomainUpdateNet(driver, dom, dev, devCopy) < 0)
            goto cleanup;
        break;

    default:
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Cannot update '%s', device type unsupported"),
                       virDomainDeviceTypeToString(dev->type));
        goto cleanup;
    }

    if (vzctDomainSaveConfig(driver, dom) < 0)
        goto cleanup;

    if (virDomainObjIsActive(dom)) {
        g_autofree char *statefile = NULL;

        statefile = vzctStateFile(driver->config, dom->def->uuid);
        if (virDomainObjSavePath(dom, driver->xmlopt, statefile) < 0)
            goto cleanup;
    }

    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static int
vzctDomainSetUserPassword(virDomainPtr domain,
                          const char *user,
                          const char *password,
                          unsigned int flags)
{
    g_autoptr(vzctl_env_handle) env = NULL;
    virDomainObj *dom = NULL;
    bool job = false;
    int ret = -1;
    int vzflags = 0;
    int err;

    virCheckFlags(VIR_DOMAIN_PASSWORD_ENCRYPTED |
                  VIR_DOMAIN_CREATE_USER, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSetUserPasswordEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_MODIFY) < 0)
        goto cleanup;
    job = true;

    if (!virDomainObjIsActive(dom)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if (flags & VIR_DOMAIN_PASSWORD_ENCRYPTED)
        vzflags |= VZCTL_SET_USERPASSWD_CRYPTED;

    VZCT_CHECK_PTR(cleanup, env, vzctl2_env_open(dom->def->extraId, 0, &err));
    VZCT_CHECK(cleanup, vzctl2_env_set_userpasswd(env, user, password, vzflags));
    ret = 0;

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static char *
vzctDomainGetHostname(virDomainPtr domain,
                      unsigned int flags)
{
    virDomainObj *dom = NULL;
    g_autofree char *hostname;
    char *ret = NULL;
    char *p;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_GET_HOSTNAME_AGENT, NULL);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetHostnameEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    if (vzctDomainBeginJob(dom, VZCT_DOMAIN_JOB_QUERY) < 0)
        goto cleanup;
    job = true;

    if (virDomainExecuteCommand(domain->conn, dom,
                                "hostname",
                                NULL, 0, &hostname,
                                &vzctExecuteCallbacks, NULL))
        goto cleanup;

    if ((p = strrchr(hostname, '\n')))
        *p = 0;
    ret = g_steal_pointer(&hostname);

 cleanup:
    if (job)
        vzctDomainEndJob(dom);
    virDomainObjEndAPI(&dom);

    return ret;
}


static virHypervisorDriver vzctHypervisorDriver = {
    .domainCommandXExec = vzctDomainCommandXExec, /* 5.6.0 */
    .domainCommandXGetStatus = vzctDomainCommandXGetStatus, /* 5.6.0 */
    .domainCommandXTerminate = vzctDomainCommandXTerminate, /* 5.6.0 */
    .domainCommandXSetParams = vzctDomainCommandXSetParams, /* 5.6.0 */
    .name = "vzct",
    .connectOpen = vzctConnectOpen, /* 5.6.0 */
    .connectClose = vzctConnectClose, /* 5.6.0 */
    .connectSupportsFeature = vzctConnectSupportsFeature, /* 5.6.0 */
    .connectGetType = vzctConnectGetType,  /* 5.6.0 */
    .connectListAllDomains = vzctConnectListAllDomains, /* 5.6.0 */
    .connectGetCapabilities = vzctConnectGetCapabilities, /* 5.6.0 */
    .connectGetAllDomainStats = vzctConnectGetAllDomainStats, /* 5.6.0 */
    .connectDomainEventRegisterAny = vzctConnectDomainEventRegisterAny, /* 5.6.0 */
    .connectDomainEventDeregisterAny = vzctConnectDomainEventDeregisterAny, /* 5.6.0 */
    .connectGetHostname = vzctConnectGetHostname, /* 5.6.0 */
    .connectGetVersion = vzctConnectGetVersion, /* 5.6.0 */
    .nodeGetInfo = vzctNodeGetInfo, /* 5.6.0 */
    .nodeGetCPUMap = vzctNodeGetCPUMap, /* 5.6.0 */
    .domainGetState = vzctDomainGetState, /* 5.6.0 */
    .domainLookupByUUID = vzctDomainLookupByUUID, /* 5.6.0 */
    .domainLookupByName = vzctDomainLookupByName, /* 5.6.0 */
    .domainGetXMLDesc = vzctDomainGetXMLDesc, /* 5.6.0 */
    .domainDefineXML = vzctDomainDefineXML,      /* 5.6.0 */
    .domainDefineXMLFlags = vzctDomainDefineXMLFlags, /* 5.6.0 */
    .domainUndefine = vzctDomainUndefine, /* 5.6.0 */
    .domainUndefineFlags = vzctDomainUndefineFlags, /* 5.6.0 */
    .domainIsActive = vzctDomainIsActive, /* 5.6.0 */
    .domainCreate = vzctDomainCreate, /* 5.6.0 */
    .domainCreateWithFlags = vzctDomainCreateWithFlags, /* 5.6.0 */
    .domainDestroy = vzctDomainDestroy, /* 5.6.0 */
    .domainDestroyFlags = vzctDomainDestroyFlags, /* 5.6.0 */
    .domainShutdown = vzctDomainShutdown, /* 5.6.0 */
    .domainShutdownFlags = vzctDomainShutdownFlags, /* 5.6.0 */
    .domainGetInfo = vzctDomainGetInfo, /* 5.6.0 */
    .domainIsPersistent = vzctDomainIsPersistent, /* 5.6.0 */
    .domainGetOSType = vzctDomainGetOSType, /* 5.6.0 */
    .domainSnapshotGetXMLDesc = vzctDomainSnapshotGetXMLDesc, /* 5.6.0 */
    .domainSnapshotNum = vzctDomainSnapshotNum, /* 5.6.0 */
    .domainSnapshotListNames = vzctDomainSnapshotListNames, /* 5.6.0 */
    .domainListAllSnapshots = vzctDomainListAllSnapshots, /* 5.6.0 */
    .domainSnapshotNumChildren = vzctDomainSnapshotNumChildren, /* 5.6.0 */
    .domainSnapshotListChildrenNames = vzctDomainSnapshotListChildrenNames, /* 5.6.0 */
    .domainSnapshotListAllChildren = vzctDomainSnapshotListAllChildren, /* 5.6.0 */
    .domainSnapshotLookupByName = vzctDomainSnapshotLookupByName, /* 5.6.0 */
    .domainHasCurrentSnapshot = vzctDomainHasCurrentSnapshot, /* 5.6.0 */
    .domainSnapshotGetParent = vzctDomainSnapshotGetParent, /* 5.6.0 */
    .domainSnapshotCurrent = vzctDomainSnapshotCurrent, /* 5.6.0 */
    .domainSnapshotIsCurrent = vzctDomainSnapshotIsCurrent, /* 5.6.0 */
    .domainSnapshotHasMetadata = vzctDomainSnapshotHasMetadata, /* 5.6.0 */
    .domainRevertToSnapshot = vzctDomainRevertToSnapshot, /* 5.6.0 */
    .domainSnapshotDelete = vzctDomainSnapshotDelete, /* 5.6.0 */
    .domainSnapshotCreateXML = vzctDomainSnapshotCreateXML, /* 5.6.0 */
    .domainMigrateBegin3Params = vzctDomainMigrateBegin3Params, /* 5.6.0 */
    .domainMigratePrepare3Params = vzctDomainMigratePrepare3Params, /* 5.6.0 */
    .domainMigratePerform3Params = vzctDomainMigratePerform3Params, /* 5.6.0 */
    .domainMigrateFinish3Params = vzctDomainMigrateFinish3Params, /* 5.6.0 */
    .domainMigrateConfirm3Params = vzctDomainMigrateConfirm3Params, /* 5.6.0 */
    .domainGetJobInfo = vzctDomainGetJobInfo, /* 5.6.0 */
    .domainGetJobStats = vzctDomainGetJobStats, /* 5.6.0 */
    .domainAbortJob = vzctDomainAbortJob, /* 5.6.0 */
    .domainGetAutostart = vzctDomainGetAutostart, /* 5.6.0 */
    .domainSetAutostart = vzctDomainSetAutostart, /* 5.6.0 */
    .domainSetMemory = vzctDomainSetMemory, /* 5.6.0 */
    .domainSetMemoryFlags = vzctDomainSetMemoryFlags, /* 5.6.0 */
    .domainSetVcpus = vzctDomainSetVcpus, /* 5.6.0 */
    .domainSetVcpusFlags = vzctDomainSetVcpusFlags, /* 5.6.0 */
    .domainGetVcpusFlags = vzctDomainGetVcpusFlags, /* 5.6.0 */
    .domainGetMaxVcpus = vzctDomainGetMaxVcpus, /* 5.6.0 */
    .domainSetMemoryParameters = vzctDomainSetMemoryParameters, /* 5.6.0 */
    .domainGetMemoryParameters = vzctDomainGetMemoryParameters, /* 5.6.0 */
    .domainSetBlockIoTune = vzctDomainSetBlockIoTune, /* 5.6.0 */
    .domainGetBlockIoTune = vzctDomainGetBlockIoTune, /* 5.6.0 */
    .domainBlockResize = vzctDomainBlockResize, /* 5.6.0 */
    .domainSetMetadata = vzctDomainSetMetadata, /* 6.6.0 */
    .domainGetMetadata = vzctDomainGetMetadata, /* 6.6.0 */
    .domainRename = vzctDomainRename, /* 6.6.0 */
    .domainManagedSave = vzctDomainManagedSave, /* 5.6.0 */
    .domainHasManagedSaveImage = vzctDomainHasManagedSaveImage, /* 5.6.0 */
    .domainManagedSaveRemove = vzctDomainManagedSaveRemove, /* 5.6.0 */
    .domainManagedSaveGetXMLDesc = vzctDomainManagedSaveGetXMLDesc, /* 5.6.0 */
    .domainManagedSaveDefineXML = vzctDomainManagedSaveDefineXML, /* 5.6.0 */
    .domainGetSchedulerType = vzctDomainGetSchedulerType, /* 5.6.0 */
    .domainGetSchedulerParameters = vzctDomainGetSchedulerParameters, /* 5.6.0 */
    .domainGetSchedulerParametersFlags = vzctDomainGetSchedulerParametersFlags, /* 5.6.0 */
    .domainSetSchedulerParameters = vzctDomainSetSchedulerParameters, /* 5.6.0 */
    .domainSetSchedulerParametersFlags = vzctDomainSetSchedulerParametersFlags, /* 5.6.0 */
    .domainGetBlkioParameters = vzctDomainGetBlkioParameters, /* 5.6.0 */
    .domainSetBlkioParameters = vzctDomainSetBlkioParameters, /* 5.6.0 */
    .domainSetGuestNetworkParametersVz = vzctDomainSetGuestNetworkParametersVz, /* 6.6.0 */
    .domainAttachDevice = vzctDomainAttachDevice, /* 5.6.0 */
    .domainAttachDeviceFlags = vzctDomainAttachDeviceFlags, /* 5.6.0 */
    .domainDetachDevice = vzctDomainDetachDevice, /* 5.6.0 */
    .domainDetachDeviceFlags = vzctDomainDetachDeviceFlags, /* 5.6.0 */
    .domainUpdateDeviceFlags = vzctDomainUpdateDeviceFlags, /* 5.6.0 */
    .domainSetUserPassword = vzctDomainSetUserPassword, /* 5.6.0 */
    .domainGetHostname = vzctDomainGetHostname, /* 7.6.0 */
};

static virConnectDriver vzctConnectDriver = {
    .localOnly = true,
    .uriSchemes = (const char *[]){ "vzct", NULL },
    .hypervisorDriver = &vzctHypervisorDriver,
};

static virStateDriver vzctStateDriver = {
    .name = "vzct",
    .stateInitialize = vzctStateInitialize,
    .stateCleanup = vzctStateCleanup,
};


int
vzctRegister(void)
{
    if (virRegisterConnectDriver(&vzctConnectDriver, true) < 0)
        return -1;

    if (virRegisterStateDriver(&vzctStateDriver) < 0)
        return -1;

    return 0;
}
